@@ -7,12 +7,14 @@ from __future__ import unicode_literals
 import hashlib
 import mimetypes
 from wsgiref.handlers import format_date_time
+import os
 import py
+import re
 from devpi_common.types import cached_property, parse_hash_spec
-from .keyfs import _nodefault
 from .log import threadlog
 
 log = threadlog
+_nodefault = object()
 
 def get_default_hash_spec(content):
     #return "md5=" + hashlib.md5(content).hexdigest()
@@ -36,9 +38,10 @@ class FileStore:
     def __init__(self, xom):
         self.xom = xom
         self.keyfs = xom.keyfs
-        self.storedir = self.keyfs.basedir.ensure("+files", dir=1)
+        self.rel_storedir = "+files"
+        self.storedir = self.keyfs.basedir.join(self.rel_storedir)
 
-    def maplink(self, link):
+    def maplink(self, link, user, index):
         if link.hash_spec:
             # we can only create 32K entries per directory
             # so let's take the first 3 bytes which gives
@@ -51,10 +54,12 @@ class FileStore:
             parts = link.torelpath().split("/")
             assert parts
             dirname = "_".join(parts[:-1])
-            key = self.keyfs.PYPIFILE_NOMD5(user="root", index="pypi",
-                   dirname=dirname,
-                   basename=parts[-1])
-        entry = FileEntry(self.xom, key)
+            dirname = re.sub('[^a-zA-Z0-9_.-]', '_', dirname)
+            key = self.keyfs.PYPIFILE_NOMD5(
+                user=user, index=index,
+                dirname=dirname,
+                basename=parts[-1])
+        entry = FileEntry(self.xom, key, readonly=False)
         entry.url = link.geturl_nofragment().url
         entry.eggfragment = link.eggfragment
         # verify checksum if the entry is fresh, a file exists
@@ -70,12 +75,12 @@ class FileStore:
         entry.hash_spec = unicode_if_bytes(link.hash_spec)
         return entry
 
-    def get_file_entry(self, relpath):
+    def get_file_entry(self, relpath, readonly=True):
         try:
-            key = self.keyfs.derive_key(relpath)
+            key = self.keyfs.tx.derive_key(relpath)
         except KeyError:
             return None
-        return FileEntry(self.xom, key)
+        return FileEntry(self.xom, key, readonly=readonly)
 
     def get_file_entry_raw(self, key, meta):
         return FileEntry(self.xom, key, meta=meta)
@@ -86,7 +91,7 @@ class FileStore:
         hashdir_a, hashdir_b = make_splitdir(dir_hash_spec)
         key = self.keyfs.STAGEFILE(user=user, index=index,
                    hashdir_a=hashdir_a, hashdir_b=hashdir_b, filename=basename)
-        entry = FileEntry(self.xom, key)
+        entry = FileEntry(self.xom, key, readonly=False)
         entry.file_set_content(file_content)
         return entry
 
@@ -96,28 +101,35 @@ def metaprop(name):
         if self.meta is not None:
             return self.meta.get(name)
     def fset(self, val):
-        self.meta[name] = unicode_if_bytes(val)
-        self.key.set(self.meta)
+        val = unicode_if_bytes(val)
+        if self.meta.get(name) != val:
+            self.meta[name] = val
+            self.key.set(self.meta)
     return property(fget, fset)
 
 
-class FileEntry(object):
-    class BadGateway(Exception):
-        pass
+class BadGateway(Exception):
+    pass
 
+
+class FileEntry(object):
+    BadGateway = BadGateway
     hash_spec = metaprop("hash_spec")  # e.g. "md5=120938012"
     eggfragment = metaprop("eggfragment")
     last_modified = metaprop("last_modified")
     url = metaprop("url")
-    projectname = metaprop("projectname")
+    project = metaprop("project")
     version = metaprop("version")
 
-    def __init__(self, xom, key, meta=_nodefault):
+    def __init__(self, xom, key, meta=_nodefault, readonly=True):
         self.xom = xom
         self.key = key
         self.relpath = key.relpath
         self.basename = self.relpath.split("/")[-1]
-        self._filepath = str(self.xom.filestore.storedir.join(self.relpath))
+        self.readonly = readonly
+        self._storepath = os.path.join(
+            self.xom.filestore.rel_storedir,
+            str(self.relpath))
         if meta is not _nodefault:
             self.meta = meta or {}
 
@@ -147,25 +159,28 @@ class FileEntry(object):
 
     @cached_property
     def meta(self):
-        return self.key.get()
+        return self.key.get(readonly=self.readonly)
 
     def file_exists(self):
-        return self.tx.io_file_exists(self._filepath)
+        return self.tx.conn.io_file_exists(self._storepath)
 
     def file_delete(self):
-        return self.tx.io_file_delete(self._filepath)
+        return self.tx.conn.io_file_delete(self._storepath)
 
     def file_size(self):
-        return self.tx.io_file_size(self._filepath)
+        return self.tx.conn.io_file_size(self._storepath)
 
     def __repr__(self):
         return "<FileEntry %r>" %(self.key)
 
     def file_open_read(self):
-        return open(self._filepath, "rb")
+        return self.tx.conn.io_file_open(self._storepath)
 
     def file_get_content(self):
-        return self.tx.io_file_get(self._filepath)
+        return self.tx.conn.io_file_get(self._storepath)
+
+    def file_os_path(self):
+        return self.tx.conn.io_file_os_path(self._storepath)
 
     def file_set_content(self, content, last_modified=None, hash_spec=None):
         assert isinstance(content, bytes)
@@ -181,7 +196,12 @@ class FileEntry(object):
         else:
             hash_spec = get_default_hash_spec(content)
         self.hash_spec = hash_spec
-        self.tx.io_file_set(self._filepath, content)
+        self.tx.conn.io_file_set(self._storepath, content)
+        # we make sure we always refresh the meta information
+        # when we set the file content. Otherwise we might
+        # end up only committing file content without any keys
+        # changed which will not replay correctly at a replica.
+        self.key.set(self.meta)
 
     def gethttpheaders(self):
         assert self.file_exists()
@@ -209,7 +229,18 @@ class FileEntry(object):
         self.meta = {}
         self.file_delete()
 
-    def cache_remote_file(self):
+    def _headers_from_response(self, r):
+        headers = {
+            str("X-Accel-Buffering"): str("no"),  # disable buffering in nginx
+            str("content-type"): r.headers.get(
+                "content-type", (str("application/octet-stream"), None))[0]}
+        if "last-modified" in r.headers:
+            headers[str("last-modified")] = r.headers["last-modified"]
+        if "content-length" in r.headers:
+            headers[str("content-length")] = str(r.headers["content-length"])
+        return headers
+
+    def iter_cache_remote_file(self):
         # we get and cache the file and some http headers from remote
         r = self.xom.httpget(self.url, allow_redirects=True)
         if r.status_code != 200:
@@ -217,11 +248,20 @@ class FileEntry(object):
             threadlog.error(msg)
             raise self.BadGateway(msg)
         log.info("reading remote: %s, target %s", r.url, self.relpath)
-        content = r.raw.read()
-        filesize = len(content)
         content_size = r.headers.get("content-length")
         err = None
 
+        yield self._headers_from_response(r)
+
+        content = b''
+        while 1:
+            data = r.raw.read(10240)
+            if not data:
+                break
+            content = content + data
+            yield data
+
+        filesize = len(content)
         if content_size and int(content_size) != filesize:
             err = ValueError(
                       "%s: got %s bytes of %r from remote, expected %s" % (
@@ -233,31 +273,69 @@ class FileEntry(object):
             log.error(str(err))
             raise err
 
-        self.file_set_content(content, r.headers.get("last-modified", None))
+        try:
+            # when pushing from a mirror to an index, we are still in a
+            # transaction
+            tx = self.tx
+        except AttributeError:
+            # when streaming we won't be in a transaction anymore, so we need
+            # to open a new one below
+            tx = None
+        if tx is not None:
+            self.file_set_content(content, r.headers.get("last-modified", None))
+        else:
+            with self.key.keyfs.transaction(write=True):
+                self.file_set_content(content, r.headers.get("last-modified", None))
 
-    def cache_remote_file_replica(self):
+    def iter_remote_file_replica(self):
+        from .replica import H_REPLICA_FILEREPL, ReplicationErrors
+        replication_errors = ReplicationErrors(self.xom.config.serverdir)
         # construct master URL with param
-        assert self.url, "should have private files already: %s" % self.relpath
-        threadlog.info("replica doesn't have file: %s", self.relpath)
         url = self.xom.config.master_url.joinpath(self.relpath).url
-
-        # we do a head request to master and then wait for the file
-        # to arrive through the replication machinery
-        r = self.xom._httpsession.head(url)
+        if not self.url:
+            threadlog.warn("missing private file: %s" % self.relpath)
+        else:
+            threadlog.info("replica doesn't have file: %s", self.relpath)
+        r = self.xom.httpget(
+            url, allow_redirects=True,
+            extra_headers={H_REPLICA_FILEREPL: str("YES")})
         if r.status_code != 200:
-            msg = "%s: received %s from master" %(url, r.status_code)
+            msg = "%s: received %s from master" % (url, r.status_code)
             threadlog.error(msg)
             raise self.BadGateway(msg)
-        serial = int(r.headers["X-DEVPI-SERIAL"])
-        keyfs = self.key.keyfs
-        keyfs.notifier.wait_tx_serial(serial)
-        keyfs.restart_read_transaction()  # use latest serial
-        entry = self.xom.filestore.get_file_entry(self.relpath)
-        if not entry.file_exists():
-            msg = "%s: did not get file after waiting" % url
-            threadlog.error(msg)
-            raise self.BadGateway(msg)
-        return entry
+
+        yield self._headers_from_response(r)
+
+        content = b''
+        while 1:
+            data = r.raw.read(10240)
+            if not data:
+                break
+            content = content + data
+            yield data
+
+        err = self.check_checksum(content)
+        if err:
+            # the file we got is different, so we fail
+            raise self.BadGateway(str(err))
+
+        try:
+            # there is no code path that still has a transaction at this point,
+            # but we handle that case just to be safe
+            tx = self.tx
+        except AttributeError:
+            # when streaming we won't be in a transaction anymore, so we need
+            # to open a new one below
+            tx = None
+        if tx is not None and tx.write:
+            self.tx.conn.io_file_set(self._storepath, content)
+        else:
+            # we need a direct write connection to use the io_file_* methods
+            with self.key.keyfs._storage.get_connection(write=True) as conn:
+                conn.io_file_set(self._storepath, content)
+                conn.commit()
+        # in case there were errors before, we can now remove them
+        replication_errors.remove(self)
 
 
 def get_checksum_error(content, hash_spec):
@@ -266,4 +344,3 @@ def get_checksum_error(content, hash_spec):
     digest = hash_algo(content).hexdigest()
     if digest != hash_value:
        return "%s mismatch, got %s, expected %s" % (hash_type, digest, hash_value)
-
