@@ -9,8 +9,10 @@ import shutil
 from devpi_common.validation import normalize_name
 from devpi_common.metadata import BasenameMeta
 from devpi_common.types import parse_hash_spec
-from devpi_server.main import fatal
-import devpi_server
+from devpi_server import __version__ as server_version
+from devpi_server.model import is_valid_name
+from .main import fatal
+from .readonly import get_mutable_deepcopy, ReadonlyView
 
 
 def do_export(path, xom):
@@ -62,19 +64,24 @@ class Exporter:
         self.export_users = self.export["users"] = {}
         self.export_indexes = self.export["indexes"] = {}
 
-    def copy_file(self, src, dest):
+    def copy_file(self, entry, dest):
         dest.dirpath().ensure(dir=1)
         relpath = dest.relto(self.basepath)
-        if self.config.args.hard_links:
+        src = entry.file_os_path()
+        if self.config.args.hard_links and src is not None:
             self.tw.line("link file at %s" % relpath)
             os.link(src, dest.strpath)
+        elif src is not None:
+            self.tw.line("copy file at %s" % relpath)
+            shutil.copyfile(src, dest.strpath)
         else:
             self.tw.line("write file at %s" % relpath)
-            shutil.copyfile(src, dest.strpath)
+            with open(dest.strpath, 'wb') as f:
+                f.write(entry.file_get_content())
         return relpath
 
     def warn(self, msg):
-        self.tw.line(msg, red=True)
+        self.tw.line(msg, yellow=True)
 
     def completed(self, msg):
         self.tw.line("dumped %s" % msg, bold=True)
@@ -83,7 +90,7 @@ class Exporter:
         self.basepath = path
         self.export["dumpversion"] = self.DUMPVERSION
         self.export["pythonversion"] = list(sys.version_info)
-        self.export["devpi_server"] = devpi_server.__version__
+        self.export["devpi_server"] = server_version
         self.export["secret"] = self.config.secret
         self.export["uuid"] = self.xom.config.get_master_uuid()
         for user in self.xom.model.get_userlist():
@@ -94,14 +101,17 @@ class Exporter:
             self.completed("user %r" % user.name)
             for indexname, indexconfig in indexes.items():
                 stage = self.xom.model.getstage(user.name, indexname)
-                if stage.ixconfig["type"] == "mirror":
-                    continue
-                indexdir = userdir.ensure(indexname, dir=1)
-                IndexDump(self, stage, indexdir).dump()
+                IndexDump(self, stage, userdir.join(indexname)).dump()
         self._write_json(path.join("dataindex.json"), self.export)
 
     def _write_json(self, path, data):
-        writedata = json.dumps(data, indent=2)
+        # use a special handler for serializing ReadonlyViews
+        def handle_readonly(val):
+            if isinstance(val, ReadonlyView):
+                return val._data
+            raise TypeError(type(val))
+
+        writedata = json.dumps(data, indent=2, default=handle_readonly)
         path.dirpath().ensure(dir=1)
         self.tw.line("writing %s, length %s" %(path.relto(self.basepath),
                                                len(writedata)))
@@ -114,18 +124,21 @@ class IndexDump:
         self.stage = stage
         self.basedir = basedir
         self.indexmeta = exporter.export_indexes[stage.name] = {}
-        self.indexmeta["projects"] = {}
         self.indexmeta["indexconfig"] = stage.ixconfig
-        self.indexmeta["files"] = []
 
     def dump(self):
-        import copy
-        for name in self.stage.list_projectnames_perstage():
+        if self.stage.ixconfig["type"] == "mirror":
+            projects = []
+        else:
+            self.indexmeta["projects"] = {}
+            self.indexmeta["files"] = []
+            projects = self.stage.list_projects_perstage()
+        for name in projects:
             data = {}
             versions = self.stage.list_versions_perstage(name)
             for version in versions:
-                data[version] = copy.deepcopy(
-                    self.stage.get_versiondata_perstage(name, version))
+                v = self.stage.get_versiondata_perstage(name, version)
+                data[version] = get_mutable_deepcopy(v)
             for val in data.values():
                 val.pop("+elinks", None)
             norm_name = normalize_name(name)
@@ -135,6 +148,7 @@ class IndexDump:
             for version in data:
                 vername = data[version]["name"]
                 linkstore = self.stage.get_linkstore_perstage(vername, version)
+                self.basedir.ensure(dir=1)
                 self.dump_releasefiles(linkstore)
                 self.dump_toxresults(linkstore)
                 entry = self.stage.get_doczip_entry(vername, version)
@@ -147,42 +161,42 @@ class IndexDump:
             entry = self.exporter.filestore.get_file_entry(link.entrypath)
             assert entry.file_exists(), entry.relpath
             relpath = self.exporter.copy_file(
-                entry._filepath,
-                self.basedir.join(linkstore.projectname, entry.basename))
-            self.add_filedesc("releasefile", linkstore.projectname, relpath,
+                entry,
+                self.basedir.join(linkstore.project, link.version, entry.basename))
+            self.add_filedesc("releasefile", linkstore.project, relpath,
                                version=linkstore.version,
-                               entrymapping=entry.meta.copy(),
+                               entrymapping=entry.meta,
                                log=link.get_logs())
 
     def dump_toxresults(self, linkstore):
         for tox_link in linkstore.get_links(rel="toxresult"):
             reflink = linkstore.stage.get_link_from_entrypath(tox_link.for_entrypath)
             relpath = self.exporter.copy_file(
-                tox_link.entry._filepath,
-                self.basedir.join(linkstore.projectname, reflink.hash_spec,
+                tox_link.entry,
+                self.basedir.join(linkstore.project, reflink.hash_spec,
                                   tox_link.basename)
             )
             self.add_filedesc(type="toxresult",
-                              projectname=linkstore.projectname,
+                              project=linkstore.project,
                               relpath=relpath,
                               version=linkstore.version,
                               for_entrypath=reflink.entrypath,
                               log=tox_link.get_logs())
 
-    def add_filedesc(self, type, projectname, relpath, **kw):
+    def add_filedesc(self, type, project, relpath, **kw):
         assert self.exporter.basepath.join(relpath).check()
         d = kw.copy()
         d["type"] = type
-        d["projectname"] = projectname
+        d["projectname"] = project
         d["relpath"] = relpath
         self.indexmeta["files"].append(d)
         self.exporter.completed("%s: %s " %(type, relpath))
 
-    def dump_docfile(self, projectname, version, entry):
+    def dump_docfile(self, project, version, entry):
         relpath = self.exporter.copy_file(
-            entry._filepath,
-            self.basedir.join("%s-%s.doc.zip" % (projectname, version)))
-        self.add_filedesc("doczip", projectname, relpath, version=version)
+            entry,
+            self.basedir.join("%s-%s.doc.zip" % (project, version)))
+        self.add_filedesc("doczip", project, relpath, version=version)
 
 
 class Importer:
@@ -197,11 +211,63 @@ class Importer:
         return json.loads(path.read())
 
     def warn(self, msg):
-        self.tw.line(msg, red=True)
+        self.tw.line(msg, yellow=True)
+
+    def display_import_header(self, path):
+        self.tw.line("******** Importing packages from %s **********" % path)
+        self.tw.line('Number of users: %d' % len(self.import_users))
+        self.tw.line('Number of indexes: %d' % len(self.import_indexes))
+        total_num_projects = 0
+        total_num_files = 0
+        for idx_name, idx in self.import_indexes.items():
+            if idx['indexconfig']['type'] == 'mirror':
+                continue
+            num_projects = len(idx['projects'])
+            num_files = len(idx['files'])
+            self.tw.line(
+                'Index %s has %d projects and %d files'
+                % (idx_name, num_projects, num_files))
+            total_num_projects += num_projects
+            total_num_files += num_files
+        self.tw.line('Total number of projects: %d' % total_num_projects)
+        self.tw.line('Total number of files: %d' % total_num_files)
+
+    def check_names(self, json_path):
+        errors = False
+        for name in self.import_users:
+            if is_valid_name(name):
+                continue
+            self.xom.log.error(
+                "username '%s' contains characters that aren't allowed. "
+                "Any ascii symbol besides -.@_ is blocked." % name)
+            errors = True
+        for index in self.import_indexes:
+            user, name = index.split('/')
+            if is_valid_name(name):
+                continue
+            self.xom.log.error(
+                "indexname '%s' contains characters that aren't allowed. "
+                "Any ascii symbol besides -.@_ is blocked." % index)
+            errors = True
+        if errors:
+            self.xom.log.warn(
+                "You should edit %s manually to fix the above errors." % json_path)
+            raise SystemExit(1)
+
+    def iter_projects_normalized(self, projects):
+        project_name_map = {}
+        for project in projects:
+            project_name_map.setdefault(normalize_name(project), set()).add(project)
+        for project, names in project_name_map.items():
+            versions = {}
+            for name in names:
+                versions.update(projects[name])
+            yield (project, versions)
 
     def import_all(self, path):
         self.import_rootdir = path
-        self.import_data = self.read_json(path.join("dataindex.json"))
+        json_path = path.join("dataindex.json")
+        self.import_data = self.read_json(json_path)
         self.dumpversion = self.import_data["dumpversion"]
         if self.dumpversion not in ("1", "2"):
             fatal("incompatible dumpversion: %r" %(self.dumpversion,))
@@ -212,10 +278,12 @@ class Importer:
         self.import_indexes = self.import_data["indexes"]
         self.xom.config.secret = secret = self.import_data["secret"]
         self.xom.config.secretfile.write(secret)
+        self.display_import_header(path)
+        self.check_names(json_path)
 
         # first create all users
-        for username, userconfig in self.import_users.items():
-            with self.xom.keyfs.transaction(write=True):
+        with self.xom.keyfs.transaction(write=True):
+            for username, userconfig in self.import_users.items():
                 if username == "root":
                     user = self.xom.model.get_user(username)
                 else:
@@ -224,18 +292,22 @@ class Importer:
 
         # memorize index inheritance structure
         tree = IndexTree()
-        tree.add("root/pypi")  # a root index
+        with self.xom.keyfs.transaction(write=False):
+            stage = self.xom.model.getstage("root/pypi")
+        if stage is not None:
+            tree.add("root/pypi")
         for stagename, import_index in self.import_indexes.items():
             bases = import_index["indexconfig"].get("bases")
             tree.add(stagename, bases)
 
         # create stages in inheritance/root-first order
         stages = []
-        for stagename in tree.iternames():
-            with self.xom.keyfs.transaction(write=True):
+        with self.xom.keyfs.transaction(write=True):
+            for stagename in tree.iternames():
                 if stagename == "root/pypi":
-                    assert self.xom.model.getstage(stagename)
-                    continue
+                    stage = self.xom.model.getstage(stagename)
+                    if stage is not None:
+                        continue
                 import_index = self.import_indexes[stagename]
                 indexconfig = import_index["indexconfig"]
                 if 'uploadtrigger_jenkins' in indexconfig:
@@ -243,21 +315,28 @@ class Importer:
                         # remove if not set, so if the trigger was never
                         # used, you don't need to install the plugin
                         del indexconfig['uploadtrigger_jenkins']
+                if 'pypi_whitelist' in indexconfig:
+                    # this was renamed in 3.0.0
+                    whitelist = indexconfig.pop('pypi_whitelist')
+                    if 'mirror_whitelist' not in indexconfig:
+                        indexconfig['mirror_whitelist'] = whitelist
                 user, index = stagename.split("/")
                 user = self.xom.model.get_user(user)
                 stage = user.create_stage(index, **indexconfig)
-            stages.append(stage)
+                stages.append(stage)
         del tree
 
         # create projects and releasefiles for each index
         for stage in stages:
-            assert stage.name != "root/pypi"
+            if stage.ixconfig["type"] == "mirror":
+                continue
+            imported_files = set()
             import_index = self.import_indexes[stage.name]
             projects = import_index["projects"]
-            #normalized = self.normalize_index_projects(projects)
-            for project, versions in projects.items():
-                for version, versiondata in versions.items():
-                    with self.xom.keyfs.transaction(write=True):
+            files = import_index["files"]
+            for project, versions in self.iter_projects_normalized(projects):
+                with self.xom.keyfs.transaction(write=True):
+                    for version, versiondata in versions.items():
                         assert "+elinks" not in versiondata
                         versiondata.pop('+doczip', None)
                         versiondata.pop(':action', None)
@@ -270,22 +349,36 @@ class Importer:
                             versiondata["version"] = version
                         stage.set_versiondata(versiondata)
 
-            # import release files
-            for filedesc in import_index["files"]:
-                with self.xom.keyfs.transaction(write=True):
-                    self.import_filedesc(stage, filedesc)
+                    # import release files
+                    for filedesc in files:
+                        if normalize_name(filedesc["projectname"]) == normalize_name(project):
+                            imported_files.add(filedesc["relpath"])
+                            self.import_filedesc(stage, filedesc)
+            missing = set(x["relpath"] for x in files) - imported_files
+            if missing:
+                fatal(
+                    "Some files weren't imported: %s" % ", ".join(
+                        sorted(missing)))
+
+        self.tw.line("********* import_all: importing finished ***********")
 
     def wait_for_events(self):
-        latest_serial = self.xom.keyfs.get_next_serial() - 1
-        self.tw.line("waiting for events until latest_serial %s"
-                     % latest_serial)
-        self.xom.keyfs.notifier.wait_event_serial(latest_serial)
-        self.tw.line("importing finished")
+        keyfs = self.xom.keyfs
+        while True:
+            event_serial = keyfs.notifier.read_event_serial()
+            latest_serial = keyfs.get_current_serial()
+            if event_serial == latest_serial:
+                break
+            self.tw.line(
+                "waiting for events until latest_serial %s" % latest_serial)
+            keyfs.notifier.wait_event_serial(latest_serial)
+        self.tw.line("wait_for_events: importing finished"
+                     "; latest_serial = %s" % latest_serial)
 
     def import_filedesc(self, stage, filedesc):
         assert stage.ixconfig["type"] != "mirror"
         rel = filedesc["relpath"]
-        projectname = filedesc["projectname"]
+        project = filedesc["projectname"]
         p = self.import_rootdir.join(rel)
         assert p.check(), p
         if filedesc["type"] == "releasefile":
@@ -296,7 +389,7 @@ class Importer:
             else:
                 version = filedesc["version"]
 
-            link = stage.store_releasefile(projectname, version,
+            link = stage.store_releasefile(project, version,
                                            p.basename, p.read("rb"),
                                            last_modified=mapping["last_modified"])
             # devpi-server-2.1 exported with md5 checksums
@@ -305,12 +398,14 @@ class Importer:
                 mapping["hash_spec"] = "md5=" + mapping["md5"]
             hash_algo, hash_value = parse_hash_spec(mapping["hash_spec"])
             digest = hash_algo(link.entry.file_get_content()).hexdigest()
-            assert digest == hash_value
+            if digest != hash_value:
+                fatal("File %s has bad checksum %s, expected %s" % (
+                      p, digest, hash_value))
             # note that the actual hash_type used within devpi-server is not
             # determined here but in store_releasefile/store_doczip/store_toxresult etc
         elif filedesc["type"] == "doczip":
             version = filedesc["version"]
-            link = stage.store_doczip(projectname, version, p.read("rb"))
+            link = stage.store_doczip(project, version, p.read("rb"))
         elif filedesc["type"] == "toxresult":
             linkstore = stage.get_linkstore_perstage(filedesc["projectname"],
                                            filedesc["version"])
@@ -345,7 +440,17 @@ class IndexTree:
                 children = self.name2children.setdefault(base, [])
                 children.append(name)
 
+    def validate(self):
+        all_bases = set(sum(self.name2bases.values(), []))
+        all_indexes = set(self.name2bases)
+        missing = all_bases - all_indexes
+        if missing:
+            fatal(
+                "The following indexes don't have information in the import "
+                "data: %s" % ", ".join(sorted(missing)))
+
     def iternames(self):
+        self.validate()
         pending = [None]
         created = set()
         while pending:
@@ -362,4 +467,8 @@ class IndexTree:
                     for child in self.name2children.get(name, []):
                         if child not in created:
                             pending.append(child)
-
+        missed = set(self.name2bases) - created
+        if missed:
+            fatal(
+                "The following stages couldn't be reached by the dependency "
+                "tree built from the bases: %s" % ", ".join(sorted(missed)))
