@@ -8,17 +8,17 @@ from __future__ import unicode_literals
 import os, sys
 import py
 
+from requests import Response
+from devpi_common.metadata import Version
 from devpi_common.types import cached_property
 from devpi_common.request import new_requests_session
 from .config import parseoptions, get_pluginmanager
 from .log import configure_logging, threadlog
 from .model import BaseStage
-from . import extpypi, replica, mythread
+from . import mythread
 from . import __version__ as server_version
 
 
-PYPIURL_XMLRPC = "https://pypi.python.org/pypi/"
-
 class Fatal(Exception):
     pass
 
@@ -31,9 +31,12 @@ def check_compatible_version(xom):
         return
     state_version = xom.get_state_version()
     if server_version != state_version:
-        state_ver = state_version.split(".")
-        server_ver = server_version.split(".")
-        if state_ver[:2] != server_ver[:2]:
+        state_ver = tuple(state_version.split("."))
+        server_ver = tuple(server_version.split("."))
+        incompatible = (
+            state_ver[:1] != server_ver[:1] or
+            Version(state_version) < Version("2.2.dev0"))
+        if incompatible:
             fatal("Incompatible state: server %s cannot run serverdir "
                   "%s created by %s.\n"
                   "Use --export from older version, then --import with newer "
@@ -77,11 +80,14 @@ def _main(pluginmanager, argv=None):
 
     configure_logging(config.args)
 
+    if args.export and not config.path_nodeinfo.exists():
+        fatal("The path '%s' contains no devpi-server data." % config.serverdir)
+
     # read/create node UUID and role of this server
     config.init_nodeinfo()
 
     xom = XOM(config)
-    if not xom.is_replica():
+    if not xom.is_replica() and xom.keyfs.get_current_serial() == -1:
         with xom.keyfs.transaction(write=True):
             set_default_indexes(xom.model)
     check_compatible_version(xom)
@@ -140,14 +146,13 @@ def wsgi_run(xom, app):
         pass
     return 0
 
+
 class XOM:
     class Exiting(SystemExit):
         pass
 
-    def __init__(self, config, proxy=None, httpget=None):
+    def __init__(self, config, httpget=None):
         self.config = config
-        if proxy is not None:
-            self.proxy = proxy
         self.thread_pool = mythread.ThreadPool()
         if httpget is not None:
             self.httpget = httpget
@@ -156,13 +161,15 @@ class XOM:
             self.set_state_version(server_version)
         self.log = threadlog
         self.polling_replicas = {}
+        self._stagecache = {}
 
     def get_state_version(self):
         versionfile = self.config.serverdir.join(".serverversion")
         if not versionfile.exists():
-            fatal("serverdir %s is unversioned, you may try use "
-              "depvi-server-1.1 with the --upgrade-state option to "
-              "upgrade from versions prior to 1.1\n" % self.config.serverdir)
+            fatal(
+            "serverdir %s is non-empty and misses devpi-server meta information. "
+            "You need to specify an empty directory or a directory that was "
+            "previously managed by devpi-server>=1.2" % self.config.serverdir)
         return versionfile.read()
 
     def set_state_version(self, version):
@@ -170,7 +177,22 @@ class XOM:
         versionfile.dirpath().ensure(dir=1)
         versionfile.write(version)
 
-    @property
+    def get_singleton(self, indexpath, key):
+        """ return a per-xom singleton for the given indexpath and key
+        or raise KeyError if no such singleton was set yet.
+        """
+        return self._stagecache[indexpath][key]
+
+    def set_singleton(self, indexpath, key, obj):
+        """ set the singleton for indexpath/key to obj. """
+        s = self._stagecache.setdefault(indexpath, {})
+        s[key] = obj
+
+    def del_singletons(self, indexpath):
+        """ delete all singletones for the given indexpath """
+        self._stagecache.pop(indexpath, None)
+
+    @cached_property
     def model(self):
         """ root model object. """
         from devpi_server.model import RootModel
@@ -179,13 +201,6 @@ class XOM:
     def main(self):
         xom = self
         args = xom.config.args
-        # need to initialize the pypi mirror state before importing
-        # because importing may need pypi mirroring state
-        if xom.is_replica():
-            proxy = replica.PyPIProxy(xom._httpsession, xom.config.master_url)
-        else:
-            proxy = self.proxy
-        xom.pypimirror.init_pypi_mirror(proxy)
         if args.export:
             from devpi_server.importexport import do_export
             #xom.thread_pool.start_one(xom.keyfs.notifier)
@@ -214,6 +229,7 @@ class XOM:
             return wsgi_run(xom, app)
 
     def fatal(self, msg):
+        self.keyfs.release_all_wait_tx()
         self.thread_pool.shutdown()
         fatal(msg)
 
@@ -226,20 +242,16 @@ class XOM:
     def keyfs(self):
         from devpi_server.keyfs import KeyFS
         from devpi_server.model import add_keys
-        keyfs = KeyFS(self.config.serverdir, readonly=self.is_replica())
+        keyfs = KeyFS(
+            self.config.serverdir,
+            self.config.storage,
+            readonly=self.is_replica(),
+            cache_size=self.config.args.keyfs_cache_size)
         add_keys(self, keyfs)
-        self.thread_pool.register(keyfs.notifier)
+        if not self.config.args.requests_only:
+            self.thread_pool.register(keyfs.notifier)
         return keyfs
 
-    @cached_property
-    def pypimirror(self):
-        from devpi_server.extpypi import PyPIMirror
-        return PyPIMirror(self)
-
-    @cached_property
-    def proxy(self):
-        return extpypi.XMLProxy(PYPIURL_XMLRPC)
-
     def new_http_session(self, component_name):
         session = new_requests_session(agent=(component_name, server_version))
         session.cert = self.config.args.replica_cert
@@ -250,6 +262,10 @@ class XOM:
         return self.new_http_session("server")
 
     def httpget(self, url, allow_redirects, timeout=30, extra_headers=None):
+        if self.config.args.offline_mode:
+            resp = Response()
+            resp.status_code = 503  # service unavailable
+            return resp
         headers = {}
         if extra_headers:
             headers.update(extra_headers)
@@ -274,8 +290,9 @@ class XOM:
 
     def create_app(self):
         from devpi_server.view_auth import DevpiAuthenticationPolicy
+        from devpi_server.views import ContentTypePredicate
         from devpi_server.views import OutsideURLMiddleware
-        from devpi_server.views import route_url
+        from devpi_server.views import route_url, INSTALLER_USER_AGENT
         from pkg_resources import get_distribution
         from pyramid.authorization import ACLAuthorizationPolicy
         from pyramid.config import Configurator
@@ -287,37 +304,42 @@ class XOM:
 
         version_info = [
             ("devpi-server", get_distribution("devpi_server").version)]
-        # as of 2015-04-24 this won't work with PluginManager from _pytest
         for plug, distinfo in self.config.pluginmanager.list_plugin_distinfo():
             threadlog.info("Found plugin %s-%s (%s)." % (
                 distinfo.project_name, distinfo.version, distinfo.location))
-            version_info.append((distinfo.project_name, distinfo.version))
+            key = (distinfo.project_name, distinfo.version)
+            if key not in version_info:
+                version_info.append(key)
         version_info.sort()
         pyramid_config.registry['devpi_version_info'] = version_info
         self.config.hook.devpiserver_pyramid_configure(
                 config=self.config,
                 pyramid_config=pyramid_config)
 
+        pyramid_config.add_view_predicate('content_type', ContentTypePredicate)
+
         pyramid_config.add_route("/+changelog/{serial}",
                                  "/+changelog/{serial}")
         pyramid_config.add_route("/+status", "/+status")
-        pyramid_config.add_route("/root/pypi/+name2serials",
-                                 "/root/pypi/+name2serials")
         pyramid_config.add_route("/+api", "/+api", accept="application/json")
         pyramid_config.add_route("{path:.*}/+api", "{path:.*}/+api", accept="application/json")
         pyramid_config.add_route("/+login", "/+login", accept="application/json")
         pyramid_config.add_route("/{user}/{index}/+e/{relpath:.*}", "/{user}/{index}/+e/{relpath:.*}")
         pyramid_config.add_route("/{user}/{index}/+f/{relpath:.*}", "/{user}/{index}/+f/{relpath:.*}")
         pyramid_config.add_route("/{user}/{index}/+simple/", "/{user}/{index}/+simple/")
-        pyramid_config.add_route("/{user}/{index}/+simple/{name}", "/{user}/{index}/+simple/{name:[^/]+/?}")
-        pyramid_config.add_route("/{user}/{index}/+simple/{name}/refresh", "/{user}/{index}/+simple/{name}/refresh")
-        pyramid_config.add_route("/{user}/{index}/{name}/{version}", "/{user}/{index}/{name}/{version:[^/]+/?}")
+        pyramid_config.add_route("/{user}/{index}/+simple/{project}",
+                                 "/{user}/{index}/+simple/{project:[^/]+/?}")
+        pyramid_config.add_route("/{user}/{index}/+simple/{project}/refresh",
+                                 "/{user}/{index}/+simple/{project}/refresh")
+        pyramid_config.add_route("/{user}/{index}/{project}/{version}",
+                                 "/{user}/{index}/{project}/{version:[^/]+/?}")
         pyramid_config.add_route(
-            "simple_redirect", "/{user}/{index}/{name:[^/]+/?}",
-            header="User-Agent:(distribute|setuptools|pip)/.*",
+            "simple_redirect", "/{user}/{index}/{project:[^/]+/?}",
+            header="User-Agent:" + INSTALLER_USER_AGENT,
             accept="text/html",
         )
-        pyramid_config.add_route("/{user}/{index}/{name}", "/{user}/{index}/{name:[^/]+/?}")
+        pyramid_config.add_route("/{user}/{index}/{project}",
+                                 "/{user}/{index}/{project:[^/]+/?}")
         pyramid_config.add_route("/{user}/{index}/", "/{user}/{index}/")
         pyramid_config.add_route("/{user}/{index}", "/{user}/{index}")
         pyramid_config.add_route("/{user}", "/{user}")
@@ -334,6 +356,8 @@ class XOM:
         pyramid_config.add_tween("devpi_server.views.tween_keyfs_transaction",
             under="devpi_server.views.tween_request_logging"
         )
+        if self.config.args.profile_requests:
+            pyramid_config.add_tween("devpi_server.main.tween_request_profiling")
         pyramid_config.add_request_method(get_remote_ip)
         pyramid_config.add_request_method(stage_url)
         pyramid_config.add_request_method(simpleindex_url)
@@ -345,22 +369,20 @@ class XOM:
         pyramid_config.registry['xom'] = self
         app = pyramid_config.make_wsgi_app()
         if self.is_replica():
-            from devpi_server.replica import ReplicaThread
-            replica_thread = ReplicaThread(self)
+            from devpi_server.replica import ReplicaThread, register_key_subscribers
+            register_key_subscribers(self)
+            self.replica_thread = ReplicaThread(self)
             # the replica thread replays keyfs changes
-            # and pypimirror.name2serials changes are discovered
+            # and project-specific changes are discovered
             # and replayed through the PypiProjectChange event
-            self.thread_pool.register(replica_thread)
-        else:
-            # the master thread directly syncs using the
-            # pypi changelog protocol
-            self.thread_pool.register(self.pypimirror,
-                                      dict(proxy=self.proxy))
+            if not self.config.args.requests_only:
+                self.thread_pool.register(self.replica_thread)
         return OutsideURLMiddleware(app, self)
 
     def is_replica(self):
         return bool(self.config.args.master_url)
 
+
 class FatalResponse:
     status_code = -1
 
@@ -401,9 +423,37 @@ def set_default_indexes(model):
     if not root_user:
         root_user = model.create_user("root", "")
         threadlog.info("created root user")
-    userconfig = root_user.key.get()
+    userconfig = root_user.key.get(readonly=False)
     indexes = userconfig["indexes"]
-    if "pypi" not in indexes:
-        indexes["pypi"] = {"bases": (), "type": "mirror", "volatile": False}
+    if "pypi" not in indexes and not model.xom.config.args.no_root_pypi:
+        indexes["pypi"] = _pypi_ixconfig_default.copy()
         root_user.key.set(userconfig)
         threadlog.info("created root/pypi index")
+
+_pypi_ixconfig_default = {
+    "type": "mirror", "volatile": False,
+    "title": "PyPI",
+    "mirror_url": "https://pypi.python.org/simple/",
+    "mirror_web_url_fmt": "https://pypi.python.org/pypi/{name}"}
+
+
+def tween_request_profiling(handler, registry):
+    from cProfile import Profile
+    req = [0]
+    num_profile = registry["xom"].config.args.profile_requests
+    # we need to use a list, so we can create a new Profile instance without
+    # getting variable scope issues
+    profile = [Profile()]
+
+    def request_profiling_handler(request):
+        profile[0].enable()
+        try:
+            return handler(request)
+        finally:
+            profile[0].disable()
+            req[0] += 1
+            if req[0] >= num_profile:
+                profile[0].print_stats("cumulative")
+                req[0] = 0
+                profile[:] = [Profile()]
+    return request_profiling_handler
