@@ -1,5 +1,4 @@
 import os
-import py
 import json
 import contextlib
 import time
@@ -9,7 +8,8 @@ from pyramid.response import Response
 from devpi_common.validation import normalize_name
 from webob.headers import EnvironHeaders, ResponseHeaders
 
-from .keyfs import load, loads, dump, get_write_file_ensure_dir, rename
+from . import mythread
+from .fileutil import loads, rename
 from .log import thread_push_log, threadlog
 from .views import is_mutating_http_method, H_MASTER_UUID, make_uuid_headers
 from .model import UpstreamError
@@ -19,9 +19,11 @@ H_REPLICA_OUTSIDE_URL = str("X-DEVPI-REPLICA-OUTSIDE-URL")
 H_REPLICA_FILEREPL = str("X-DEVPI-REPLICA-FILEREPL")
 H_EXPECTED_MASTER_ID = str("X-DEVPI-EXPECTED-MASTER-ID")
 
+MAX_REPLICA_BLOCK_TIME = 30.0
+
+
 class MasterChangelogRequest:
-    MAX_REPLICA_BLOCK_TIME = 30.0
-    WAKEUP_INTERVAL = 2.0
+    MAX_REPLICA_BLOCK_TIME = MAX_REPLICA_BLOCK_TIME
 
     def __init__(self, request):
         self.request = request
@@ -89,9 +91,7 @@ class MasterChangelogRequest:
                     serial = int(serial)
                 except ValueError:
                     raise HTTPNotFound("serial needs to be int")
-                raw_entry = keyfs._fs.get_raw_changelog_entry(serial)
-                if not raw_entry:
-                    raw_entry = self._wait_for_entry(serial)
+                raw_entry = self._wait_for_entry(serial)
 
             devpi_serial = keyfs.get_current_serial()
             r = Response(body=raw_entry, status=200, headers={
@@ -101,157 +101,176 @@ class MasterChangelogRequest:
             return r
 
     def _wait_for_entry(self, serial):
-        max_wakeups = self.MAX_REPLICA_BLOCK_TIME / self.WAKEUP_INTERVAL
         keyfs = self.xom.keyfs
-        with keyfs.notifier.cv_new_transaction:
-            next_serial = keyfs.get_next_serial()
-            if serial > next_serial:
-                raise HTTPNotFound("can only wait for next serial")
-            with threadlog.around("debug",
-                                  "waiting for tx%s", serial):
-                num_wakeups = 0
-                while serial >= keyfs.get_next_serial():
-                    if num_wakeups >= max_wakeups:
-                        raise HTTPAccepted("no new transaction yet",
-                            headers={str("X-DEVPI-SERIAL"):
-                                     str(keyfs.get_current_serial())})
-                    # we loop because we want control-c to get through
-                    keyfs.notifier.cv_new_transaction.wait(
-                        self.WAKEUP_INTERVAL)
-                    num_wakeups += 1
-        return keyfs._fs.get_raw_changelog_entry(serial)
-
-
-    @view_config(route_name="/root/pypi/+name2serials")
-    def get_name2serials(self):
-        io = py.io.BytesIO()
-        dump(self.xom.pypimirror.name2serials, io)
-        headers = {str("Content-Type"): str("application/octet-stream")}
-        return Response(body=io.getvalue(), status=200, headers=headers)
+        next_serial = keyfs.get_next_serial()
+        if serial > next_serial:
+            raise HTTPNotFound("can only wait for next serial")
+        elif serial == next_serial:
+            arrived = keyfs.wait_tx_serial(serial, timeout=self.MAX_REPLICA_BLOCK_TIME)
+            if not arrived:
+                raise HTTPAccepted("no new transaction yet",
+                    headers={str("X-DEVPI-SERIAL"):
+                             str(keyfs.get_current_serial())})
+        return keyfs.tx.conn.get_raw_changelog_entry(serial)
 
 
 class ReplicaThread:
+    REPLICA_REQUEST_TIMEOUT = MAX_REPLICA_BLOCK_TIME * 1.25
     ERROR_SLEEP = 50
 
     def __init__(self, xom):
         self.xom = xom
         self.master_url = xom.config.master_url
-        if xom.is_replica():
-            xom.keyfs.notifier.on_key_change("PYPILINKS",
-                                             PypiProjectChanged(xom))
+        self._master_serial = None
+        self._master_serial_timestamp = None
+        self.started_at = None
+        # updated whenever we try to connect to the master
+        self.master_contacted_at = None
+        # updated on valid reply or 202 from master
+        self.update_from_master_at = None
+        # set whenever the master serial and current replication serial match
+        self.replica_in_sync_at = None
+
+    def get_master_serial(self):
+        return self._master_serial
+
+    def get_master_serial_timestamp(self):
+        return self._master_serial_timestamp
+
+    def update_master_serial(self, serial):
+        now = time.time()
+        # record that we got a reply from the master, so we can produce status
+        # information about the connection to master
+        self.update_from_master_at = now
+        if self.xom.keyfs.get_current_serial() == serial:
+            self.replica_in_sync_at = now
+        if self._master_serial is not None and serial <= self._master_serial:
+            if serial < self._master_serial:
+                self.log.error(
+                    "Got serial %s from master which is smaller than last "
+                    "recorded serial %s." % (serial, self._master_serial))
+            return
+        self._master_serial = serial
+        self._master_serial_timestamp = now
+
+    def tick(self):
+        log = self.log
+        keyfs = self.xom.keyfs
+        config = self.xom.config
+        self.thread.exit_if_shutdown()
+        serial = keyfs.get_next_serial()
+        url = self.master_url.joinpath("+changelog", str(serial)).url
+        log.info("fetching %s", url)
+        uuid, master_uuid = make_uuid_headers(config.nodeinfo)
+        assert uuid != master_uuid
+        try:
+            self.master_contacted_at = time.time()
+            r = self.session.get(url, headers={
+                H_REPLICA_UUID: uuid,
+                H_EXPECTED_MASTER_ID: master_uuid,
+                H_REPLICA_OUTSIDE_URL: config.args.outside_url,
+            }, timeout=self.REPLICA_REQUEST_TIMEOUT)
+            remote_serial = int(r.headers["X-DEVPI-SERIAL"])
+        except Exception as e:
+            log.error("error fetching %s: %s", url, str(e))
+        else:
+            # we check that the remote instance
+            # has the same UUID we saw last time
+            master_uuid = config.get_master_uuid()
+            remote_master_uuid = r.headers.get(H_MASTER_UUID)
+            if not remote_master_uuid:
+                # we don't fatally leave the process because
+                # it might just be a temporary misconfiguration
+                # for example of a nginx frontend
+                log.error("remote provides no %r header, running "
+                          "<devpi-server-2.1?"
+                          " headers were: %s", H_MASTER_UUID, r.headers)
+                self.thread.sleep(self.ERROR_SLEEP)
+                return
+            if master_uuid and remote_master_uuid != master_uuid:
+                # we got a master_uuid and it is not the one we
+                # expect, we are replicating for -- it's unlikely this heals
+                # itself.  It's thus better to die and signal we can't operate.
+                log.error("FATAL: master UUID %r does not match "
+                          "expected master UUID %r. EXITTING.",
+                          remote_master_uuid, master_uuid)
+                # force exit of the process
+                os._exit(3)
+
+            if r.status_code == 200:
+                try:
+                    changes, rel_renames = loads(r.content)
+                    keyfs.import_changes(serial, changes)
+                except Exception:
+                    log.exception("could not process: %s", r.url)
+                else:
+                    # we successfully received data so let's
+                    # record the master_uuid for future consistency checks
+                    if not master_uuid:
+                        self.xom.config.set_master_uuid(remote_master_uuid)
+                    # also record the current master serial for status info
+                    self.update_master_serial(remote_serial)
+                    return
+            elif r.status_code == 202:
+                log.debug("%s: trying again %s\n", r.status_code, url)
+                # also record the current master serial for status info
+                self.update_master_serial(remote_serial)
+                return
+            else:
+                log.error("%s: failed fetching %s", r.status_code, url)
+        # we got an error, let's wait a bit
+        self.thread.sleep(5.0)
 
     def thread_run(self):
         # within a devpi replica server this thread is the only writer
-        log = thread_push_log("[REP]")
-        session = self.xom.new_http_session("replica")
+        self.started_at = time.time()
+        self.log = thread_push_log("[REP]")
+        self.session = self.xom.new_http_session("replica")
         keyfs = self.xom.keyfs
         errors = ReplicationErrors(self.xom.config.serverdir)
         for key in (keyfs.STAGEFILE, keyfs.PYPIFILE_NOMD5):
             keyfs.subscribe_on_import(key, ImportFileReplica(self.xom, errors))
-        config = self.xom.config
         while 1:
-            self.thread.exit_if_shutdown()
-            serial = keyfs.get_next_serial()
-            url = self.master_url.joinpath("+changelog", str(serial)).url
-            log.info("fetching %s", url)
-            uuid, master_uuid = make_uuid_headers(config.nodeinfo)
-            assert uuid != master_uuid
             try:
-                r = session.get(url, headers={
-                    H_REPLICA_UUID: uuid,
-                    H_EXPECTED_MASTER_ID: master_uuid,
-                    H_REPLICA_OUTSIDE_URL: config.args.outside_url,
-                })
-            except session.Errors as e:
-                log.error("error fetching %s: %s", url, str(e))
-            else:
-                # we check that the remote instance
-                # has the same UUID we saw last time
-                master_uuid = config.get_master_uuid()
-                remote_master_uuid = r.headers.get(H_MASTER_UUID)
-                if not remote_master_uuid:
-                    # we don't fatally leave the process because
-                    # it might just be a temporary misconfiguration
-                    # for example of a nginx frontend
-                    log.error("remote provides no %r header, running "
-                              "<devpi-server-2.1?"
-                              " headers were: %s", H_MASTER_UUID, r.headers)
-                    self.thread.sleep(self.ERROR_SLEEP)
-                    continue
-                if master_uuid and remote_master_uuid != master_uuid:
-                    # we got a master_uuid and it is not the one we
-                    # expect, we are replicating for -- it's unlikely this heals
-                    # itself.  It's thus better to die and signal we can't operate.
-                    log.error("FATAL: master UUID %r does not match "
-                              "expected master UUID %r. EXITTING.",
-                              remote_master_uuid, master_uuid)
-                    # force exit of the process
-                    os._exit(3)
-
-                if r.status_code == 200:
-                    try:
-                        changes, rel_renames = loads(r.content)
-                        keyfs.import_changes(serial, changes)
-                    except Exception:
-                        log.exception("could not process: %s", r.url)
-                    else:
-                        serial += 1
-                        # we successfully received data so let's
-                        # record the master_uuid for future consistency checks
-                        if not master_uuid:
-                            self.xom.config.set_master_uuid(remote_master_uuid)
-                        continue
-                elif r.status_code == 202:
-                    log.debug("%s: trying again %s\n", r.status_code, url)
-                    continue
-                else:
-                    log.error("%s: failed fetching %s", r.status_code, url)
-            # we got an error, let's wait a bit
-            self.thread.sleep(5.0)
+                self.tick()
+            except mythread.Shutdown:
+                raise
+            except:
+                self.log.exception(
+                    "Unhandled exception in replica thread.")
+                self.thread.sleep(1.0)
 
 
-class PyPIProxy(object):
-    def __init__(self, http, master_url):
-        self._url = master_url.joinpath("root/pypi/+name2serials").url
-        self._http = http
-
-    def list_packages_with_serial(self):
-        try:
-            r = self._http.get(self._url, stream=True)
-        except self._http.Errors:
-            threadlog.exception("proxy request failed, no connection?")
-        else:
-            if r.status_code == 200:
-                return load(r.raw)
-        from devpi_server.main import fatal
-        fatal("replica: could not get serials from remote")
+def register_key_subscribers(xom):
+    xom.keyfs.PROJSIMPLELINKS.on_key_change(SimpleLinksChanged(xom))
 
 
-class PypiProjectChanged:
-    """ Event executed in notification thread based on a pypi link change. """
+class SimpleLinksChanged:
+    """ Event executed in notification thread based on a pypi link change.
+    It allows a replica to sync up the local full projectnames list."""
     def __init__(self, xom):
         self.xom = xom
 
     def __call__(self, ev):
-        threadlog.info("PypiProjectChanged %s", ev.typedkey)
-        pypimirror = self.xom.pypimirror
-        name2serials = pypimirror.name2serials
+        threadlog.info("SimpleLinksChanged %s", ev.typedkey)
         cache = ev.value
-
-        # get the normalized projectname (PYPILINKS uses it)
-        nname = ev.typedkey.params["name"]
-        if not nname:
-            threadlog.error("project %r missing", nname)
+        # get the normalized project (PYPILINKS uses it)
+        username = ev.typedkey.params["user"]
+        index = ev.typedkey.params["index"]
+        project = ev.typedkey.params["project"]
+        if not project:
+            threadlog.error("project %r missing", project)
             return
-        assert normalize_name(nname) == nname
+        assert normalize_name(project) == project
 
-        projectname = ev.value["projectname"] if cache else nname
-        if cache is None:  # deleted
-            pypimirror.set_project_serial(projectname, None)
-        else:
-            cur_serial = name2serials.get(projectname, -1)
-            if cache and cache["serial"] > cur_serial:
-                pypimirror.set_project_serial(projectname, cache["serial"])
+        with self.xom.keyfs.transaction(write=False):
+            mirror_stage = self.xom.model.getstage(username, index)
+            if mirror_stage.ixconfig["type"] == "mirror":
+                cache_projectnames = mirror_stage.cache_projectnames.get_inplace()
+                if cache is None:  # deleted
+                    cache_projectnames.discard(project)
+                else:
+                    cache_projectnames.add(project)
 
 
 def tween_replica_proxy(handler, registry):
@@ -299,15 +318,17 @@ def clean_response_headers(response):
     return headers
 
 
-def proxy_request_to_master(xom, request):
+def proxy_request_to_master(xom, request, stream=False):
     master_url = xom.config.master_url
     url = master_url.joinpath(request.path).url
+    assert url.startswith(master_url.url)
     http = xom._httpsession
     with threadlog.around("info", "relaying: %s %s", request.method, url):
         try:
             return http.request(request.method, url,
                                 data=request.body,
                                 headers=clean_request_headers(request),
+                                stream=stream,
                                 allow_redirects=False)
         except http.Errors as e:
             raise UpstreamError("proxy-write-to-master %s: %s" % (url, e))
@@ -317,12 +338,13 @@ def proxy_write_to_master(xom, request):
     """ relay modifying http requests to master and wait until
     the change is replicated back.
     """
-    r = proxy_request_to_master(xom, request)
+    r = proxy_request_to_master(xom, request, stream=True)
+    body = r.raw.read()
     #threadlog.debug("relay status_code: %s", r.status_code)
     #threadlog.debug("relay headers: %s", r.headers)
     if r.status_code < 400:
         commit_serial = int(r.headers["X-DEVPI-SERIAL"])
-        xom.keyfs.notifier.wait_tx_serial(commit_serial)
+        xom.keyfs.wait_tx_serial(commit_serial)
     headers = clean_response_headers(r)
     headers[str("X-DEVPI-PROXY")] = str("replica")
     if r.status_code == 302:  # REDIRECT
@@ -332,7 +354,7 @@ def proxy_write_to_master(xom, request):
         headers[str("location")] = str(
             master_location.replace(xom.config.master_url.url, outside_url))
     return Response(status="%s %s" %(r.status_code, r.reason),
-                    body=r.content,
+                    body=body,
                     headers=headers)
 
 
@@ -375,13 +397,13 @@ class ImportFileReplica:
         threadlog.debug("ImportFileReplica for %s, %s", key, val)
         relpath = key.relpath
         entry = self.xom.filestore.get_file_entry_raw(key, val)
-        file_exists = os.path.exists(entry._filepath)
+        file_exists = fswriter.conn.io_file_exists(entry._storepath)
         if val is None:
             if back_serial >= 0:
                 # file was deleted, still might never have been replicated
                 if file_exists:
-                    threadlog.debug("mark for deletion: %s", entry._filepath)
-                    fswriter.record_rename_file(None, entry._filepath)
+                    threadlog.debug("mark for deletion: %s", entry._storepath)
+                    fswriter.conn.io_file_delete(entry._storepath)
             return
         if file_exists or entry.last_modified is None:
             # we have a file or there is no remote file
@@ -400,7 +422,7 @@ class ImportFileReplica:
             return
 
         if r.status_code != 200:
-            raise FileReplicationError(r)
+            raise FileReplicationError(r, relpath)
         err = entry.check_checksum(r.content)
         if err:
             # the file we got is different, it may have changed later.
@@ -412,17 +434,15 @@ class ImportFileReplica:
             return
         # in case there were errors before, we can now remove them
         self.errors.remove(entry)
-        tmppath = entry._filepath + "-tmp"
-        with get_write_file_ensure_dir(tmppath) as f:
-            f.write(r.content)
-        fswriter.record_rename_file(tmppath, entry._filepath)
+        fswriter.conn.io_file_set(entry._storepath, r.content)
 
 
 class FileReplicationError(Exception):
     """ raised when replicating a file from the master failed. """
-    def __init__(self, response, message=None):
+    def __init__(self, response, relpath, message=None):
         self.url = response.url
         self.status_code = response.status_code
+        self.relpath = relpath
         self.message = message or "failed"
 
     def __str__(self):
