@@ -1,13 +1,16 @@
 from __future__ import unicode_literals
 
+import collections
 import os
 import py
-from py.xml import html
+import re
+from time import time
 from devpi_common.types import ensure_unicode
 from devpi_common.url import URL
 from devpi_common.metadata import get_pyversion_filetype
 import devpi_server
 from pyramid.compat import urlparse
+from pyramid.interfaces import IAuthenticationPolicy
 from pyramid.httpexceptions import HTTPException, HTTPFound, HTTPSuccessful
 from pyramid.httpexceptions import HTTPUnauthorized
 from pyramid.httpexceptions import exception_response
@@ -19,8 +22,10 @@ import json
 from devpi_common.request import new_requests_session
 from devpi_common.validation import normalize_name, is_valid_archive_name
 
-from .model import InvalidIndexconfig, InvalidUser, get_ixconfigattrs
-from .keyfs import copy_if_mutable
+from .filestore import BadGateway
+from .model import InvalidIndex, InvalidIndexconfig, InvalidUser
+from .model import get_ixconfigattrs
+from .readonly import get_mutable_deepcopy
 from .log import thread_push_log, thread_pop_log, threadlog
 
 from .auth import Auth
@@ -40,6 +45,9 @@ meta_headers = {str("X-DEVPI-API-VERSION"): str(API_VERSION),
                 str("X-DEVPI-SERVER-VERSION"): server_version}
 
 
+INSTALLER_USER_AGENT = r"([^ ]* )*(distribute|setuptools|pip|pex)/.*"
+
+
 def abort(request, code, body):
     # if no Accept header is set, then force */*, otherwise the exception
     # will be returned as text/plain, which causes easy_install/setuptools
@@ -47,16 +55,22 @@ def abort(request, code, body):
     request.headers.setdefault("Accept", "*/*")
     if "application/json" in request.headers.get("Accept", ""):
         apireturn(code, body)
-    threadlog.error(body)
+    threadlog.error("while handling %s:\n%s" % (request.url, body))
     raise exception_response(code, explanation=body, headers=meta_headers)
 
-def abort_submit(code, msg):
+
+def abort_submit(request, code, msg, level="error"):
     # we construct our own type because we need to set the title
     # so that setup.py upload/register use it to explain the failure
     error = type(
         str('HTTPError'), (HTTPException,), dict(
             code=code, title=msg))
-    threadlog.error(msg)
+    if level == "info":
+        threadlog.info("while handling %s:\n%s" % (request.url, msg))
+    elif level == "warn":
+        threadlog.warn("while handling %s:\n%s" % (request.url, msg))
+    else:
+        threadlog.error("while handling %s:\n%s" % (request.url, msg))
     raise error(headers=meta_headers)
 
 
@@ -100,6 +114,18 @@ def json_preferred(request):
     return "application/json" in request.headers.get("Accept", "")
 
 
+class ContentTypePredicate(object):
+    def __init__(self, val, config):
+        self.val = val
+
+    def text(self):
+        return 'content type = %s' % self.val
+    phash = text
+
+    def __call__(self, context, request):
+        return request.content_type == self.val
+
+
 class OutsideURLMiddleware(object):
     def __init__(self, app, xom):
         self.app = app
@@ -131,7 +157,6 @@ def route_url(self, *args, **kw):
 
 def tween_request_logging(handler, registry):
     req_count = itertools.count()
-    from time import time
 
     nodeinfo = registry["xom"].config.nodeinfo
 
@@ -177,14 +202,19 @@ def tween_keyfs_transaction(handler, registry):
         with keyfs.transaction(write=write) as tx:
             threadlog.debug("in-transaction %s", tx.at_serial)
             response = handler(request)
-        serial = tx.commit_serial if tx.commit_serial is not None \
-                                  else tx.at_serial
-        set_header_devpi_serial(response.headers, serial)
+        set_header_devpi_serial(response, tx)
         return response
     return request_tx_handler
 
-def set_header_devpi_serial(headers, serial):
-    headers[str("X-DEVPI-SERIAL")] = str(serial)
+
+def set_header_devpi_serial(response, tx):
+    if isinstance(response._app_iter, collections.Iterator):
+        return
+    if tx.commit_serial is not None:
+        serial = tx.commit_serial
+    else:
+        serial = tx.at_serial
+    response.headers[str("X-DEVPI-SERIAL")] = str(serial)
 
 
 def is_mutating_http_method(method):
@@ -195,8 +225,7 @@ class StatusView:
         self.request = request
         self.xom = request.registry["xom"]
 
-    @view_config(route_name="/+status")
-    def status(self):
+    def _status(self):
         config = self.xom.config
 
         status = {
@@ -209,7 +238,12 @@ class StatusView:
             "port": config.args.port,
             "outside-url": config.args.outside_url,
             "serial": self.xom.keyfs.get_current_serial(),
+            "last-commit-timestamp": self.xom.keyfs.get_last_commit_timestamp(),
             "event-serial": self.xom.keyfs.notifier.read_event_serial(),
+            "event-serial-timestamp":
+                self.xom.keyfs.notifier.get_event_serial_timestamp(),
+            "event-serial-in-sync-at":
+                self.xom.keyfs.notifier.event_serial_in_sync_at,
         }
         master_url = config.args.master_url
         if master_url:
@@ -217,12 +251,71 @@ class StatusView:
             status["role"] = "REPLICA"
             status["master-url"] = master_url
             status["master-uuid"] = config.nodeinfo.get("master-uuid")
+            status["master-serial"] = self.xom.replica_thread.get_master_serial()
+            status["master-serial-timestamp"] = self.xom.replica_thread.get_master_serial_timestamp()
+            status["replica-started-at"] = self.xom.replica_thread.started_at
+            status["master-contacted-at"] = self.xom.replica_thread.master_contacted_at
+            status["update-from-master-at"] = self.xom.replica_thread.update_from_master_at
+            status["replica-in-sync-at"] = self.xom.replica_thread.replica_in_sync_at
             replication_errors = ReplicationErrors(self.xom.config.serverdir)
             status["replication-errors"] = replication_errors.errors
         else:
             status["role"] = "MASTER"
         status["polling_replicas"] = self.xom.polling_replicas
-        apireturn(200, type="status", result=status)
+        return status
+
+    @view_config(route_name="/+status", accept="application/json")
+    def status(self):
+        apireturn(200, type="status", result=self._status())
+
+
+def devpiweb_get_status_info(request):
+    msgs = []
+    status = StatusView(request)._status()
+    now = time()
+    if status["role"] == "REPLICA":
+        master_serial = status["master-serial"]
+        if master_serial is not None and master_serial > status["serial"]:
+            if status["replica-in-sync-at"] is None or (now - status["replica-in-sync-at"]) > 300:
+                msgs.append(dict(status="fatal", msg="Replica is behind master for more than 5 minutes"))
+            elif (now - status["replica-in-sync-at"]) > 60:
+                msgs.append(dict(status="warn", msg="Replica is behind master for more than 1 minute"))
+        else:
+            if len(status["replication-errors"]):
+                msgs.append(dict(status="fatal", msg="Unhandled replication errors"))
+        if status["replica-started-at"] is not None:
+            last_update = status["update-from-master-at"]
+            if last_update is None:
+                if (now - status["replica-started-at"]) > 300:
+                    msgs.append(dict(status="fatal", msg="No contact to master for more than 5 minutes"))
+                elif (now - status["replica-started-at"]) > 60:
+                    msgs.append(dict(status="warn", msg="No contact to master for more than 1 minute"))
+            elif (now - last_update) > 300:
+                msgs.append(dict(status="fatal", msg="No update from master for more than 5 minutes"))
+            elif (now - last_update) > 60:
+                msgs.append(dict(status="warn", msg="No update from master for more than 1 minute"))
+    if status["serial"] > status["event-serial"]:
+        if status["event-serial-in-sync-at"] is None:
+            sync_at = None
+        else:
+            sync_at = now - status["event-serial-in-sync-at"]
+        if status["event-serial-timestamp"] is None:
+            last_processed = None
+        else:
+            last_processed = now - status["event-serial-timestamp"]
+        if sync_at is None and last_processed is None and (now - status["last-commit-timestamp"]) <= 300:
+            pass
+        elif sync_at is None and last_processed is None and (now - status["last-commit-timestamp"]) > 300:
+            msgs.append(dict(status="fatal", msg="The event processing doesn't seem to start"))
+        elif sync_at is None or (sync_at > 21600):
+            msgs.append(dict(status="fatal", msg="The event processing hasn't been in sync for more than 6 hours"))
+        elif sync_at > 3600:
+            msgs.append(dict(status="warn", msg="The event processing hasn't been in sync for more than 1 hour"))
+        if sync_at is not None and (last_processed is None or (last_processed > 1800)):
+            msgs.append(dict(status="fatal", msg="No changes processed by plugins for more than 30 minutes"))
+        elif sync_at is not None and (last_processed > 300):
+            msgs.append(dict(status="warn", msg="No changes processed by plugins for more than 5 minutes"))
+    return msgs
 
 
 class PyPIView:
@@ -299,115 +392,142 @@ class PyPIView:
     # index serving and upload
     #
 
-    @view_config(route_name="/{user}/{index}/+simple/{name}")
+    @view_config(route_name="/{user}/{index}/+simple/{project}")
     def simple_list_project(self):
         request = self.request
-        name = self.context.name
+        abort_if_invalid_project(request, request.matchdict["project"])
+        project = self.context.project
         # we only serve absolute links so we don't care about the route's slash
-        abort_if_invalid_projectname(request, name)
         stage = self.context.stage
-        if stage.get_projectname(name) is None:
-            # we return 200 instead of !=200 so that pip/easy_install don't
-            # ask for the full simple page although we know it doesn't exist
-            # XXX change that when pip-6.0 is released?
-            abort(request, 200, "no such project %r" % name)
-
-        projectname = self.context.projectname
+        requested_by_pip = re.match(INSTALLER_USER_AGENT,
+            request.user_agent or "")
         try:
-            result = stage.get_releaselinks(projectname)
+            result = stage.get_simplelinks(project, sorted_links=not requested_by_pip)
         except stage.UpstreamError as e:
             threadlog.error(e.msg)
             abort(request, 502, e.msg)
-        links = []
-        for link in result:
-            relpath = link.entrypath
-            href = "/" + relpath
-            href = URL(request.path_info).relpath(href)
-            if link.eggfragment:
-                href += "#egg=%s" % link.eggfragment
-            elif link.hash_spec:
-                href += "#" + link.hash_spec
-            links.extend([
-                 "/".join(relpath.split("/", 2)[:2]) + " ",
-                 html.a(link.basename, href=href),
-                 html.br(), "\n",
-            ])
-        title = "%s: links for %s" % (stage.name, projectname)
-        if stage.has_pypi_base(projectname):
-            refresh_title = "Refresh" if stage.ixconfig["type"] == "mirror" else \
-                            "Refresh PyPI links"
-            refresh_url = request.route_url(
-                "/{user}/{index}/+simple/{name}/refresh",
-                user=self.context.username, index=self.context.index,
-                name=projectname)
-            refresh_form = [
-                html.form(
-                    html.input(
-                        type="submit", value=refresh_title, name="refresh"),
-                    action=refresh_url,
-                    method="post"),
-                "\n"]
+
+        if not result:
+            self.request.context.verified_project  # access will trigger 404 if not found
+
+        if requested_by_pip:
+            # we don't need the extra stuff on the simple page for pip
+            embed_form = False
+            blocked_index = None
         else:
-            refresh_form = []
-        return Response(html.html(
-            html.head(
-                html.title(title)),
-            html.body(
-                html.h1(title), "\n",
-                refresh_form,
-                links)).unicode(indent=2))
+            # only mere humans need to know and do more
+            whitelist_info = stage.get_mirror_whitelist_info(project)
+            embed_form = whitelist_info['has_mirror_base']
+            blocked_index = whitelist_info['blocked_by_mirror_whitelist']
+        response = Response(body=b"".join(self._simple_list_project(
+            stage, project, result, embed_form, blocked_index)))
+        if stage.ixconfig['type'] == 'mirror':
+            serial = stage.key_projsimplelinks(project).get().get("serial")
+            if serial > 0:
+                response.headers[str("X-PYPI-LAST-SERIAL")] = str(serial)
+        return response
+
+    def _simple_list_project(self, stage, project, result, embed_form, blocked_index):
+        response = self.request.response
+        response.content_type = "text/html ; charset=utf-8"
+
+        title = "%s: links for %s" % (stage.name, project)
+        yield ("<html><head><title>%s</title></head><body><h1>%s</h1>\n" %
+               (title, title)).encode("utf-8")
+
+        if embed_form:
+            yield self._index_refresh_form(stage, project).encode("utf-8")
+
+        if blocked_index:
+            yield ("<p><strong>INFO:</strong> Because this project isn't in "
+                   "the <code>mirror_whitelist</code>, no releases from "
+                   "<strong>%s</strong> are included.</p>"
+                   % blocked_index).encode('utf-8')
+
+        url = URL(self.request.path_info)
+        for key, href in result:
+            yield ('%s <a href="%s">%s</a><br/>\n' %
+                   ("/".join(href.split("/", 2)[:2]),
+                    url.relpath("/" + href),
+                    key)).encode("utf-8")
+
+        yield "</body></html>".encode("utf-8")
+
+    def _index_refresh_form(self, stage, project):
+        url = self.request.route_url(
+            "/{user}/{index}/+simple/{project}/refresh",
+            user=self.context.username, index=self.context.index,
+            project=project)
+        title = "Refresh" if stage.ixconfig["type"] == "mirror" else "Refresh PyPI links"
+        submit = '<input name="refresh" type="submit" value="%s"/>' % title
+        return '<form action="%s" method="post">%s</form>' % (url, submit)
 
     @view_config(route_name="/{user}/{index}/+simple/")
     def simple_list_all(self):
         self.log.info("starting +simple")
         stage = self.context.stage
         try:
-            stage_results = list(stage.op_sro("list_projectnames_perstage"))
+            stage_results = list(stage.op_sro("list_projects_perstage"))
         except stage.UpstreamError as e:
             threadlog.error(e.msg)
             abort(self.request, 502, e.msg)
         # at this point we are sure we can produce the data without
         # depending on remote networks
-        return Response(app_iter=self._simple_list_all(stage, stage_results))
+        return Response(body=b"".join(self._simple_list_all(stage, stage_results)))
 
     def _simple_list_all(self, stage, stage_results):
-        encoding = "utf-8"
         response = self.request.response
-        response.content_type = "text/html ; charset=%s" % encoding
+        response.content_type = "text/html ; charset=utf-8"
         title =  "%s: simple list (including inherited indices)" %(
                  stage.name)
         yield ("<html><head><title>%s</title></head><body><h1>%s</h1>" %(
-              title, title)).encode(encoding)
+              title, title)).encode("utf-8")
         all_names = set()
         for stage, names in stage_results:
             h2 = stage.name
             bases = getattr(stage, "ixconfig", {}).get("bases")
             if bases:
                 h2 += " (bases: %s)" % ",".join(bases)
-            yield ("<h2>" + h2 + "</h2>").encode(encoding)
+            yield ("<h2>" + h2 + "</h2>").encode("utf-8")
             for name in sorted(names):
                 if name not in all_names:
                     anchor = '<a href="%s">%s</a><br/>\n' % (name, name)
-                    yield anchor.encode(encoding)
+                    yield anchor.encode("utf-8")
                     all_names.add(name)
-        yield "</body>".encode(encoding)
+        yield "</body>".encode("utf-8")
 
     @view_config(
-        route_name="/{user}/{index}/+simple/{name}/refresh", request_method="POST")
+        route_name="/{user}/{index}/+simple/{project}/refresh", request_method="POST")
     def simple_refresh(self):
         context = self.context
-        # XXX we might want to check if user/index has root/pypi as a base
-        stage = context.model.getstage('root', 'pypi')
-        assert stage.ixconfig["type"] == "mirror", stage.ixconfig
-        stage.clear_cache(context.name)
-        stage.get_releaselinks_perstage(context.name)
+        for stage in context.stage.sro():
+            if stage.ixconfig["type"] != "mirror":
+                continue
+            stage.clear_simplelinks_cache(context.project)
+            stage.get_simplelinks_perstage(context.project)
         redirect(self.request.route_url(
-            "/{user}/{index}/+simple/{name}",
-            user=context.username, index=context.index, name=context.name))
+            "/{user}/{index}/+simple/{project}",
+            user=context.username, index=context.index, project=context.project))
 
     @view_config(
         route_name="/{user}/{index}", request_method="PUT")
     def index_create(self):
+        username = self.context.username
+        user = self.model.get_user(username)
+        if user is None:
+            # If the currently authenticated user tries to create an index,
+            # we create a user object automatically. The user object may
+            # not exist if the user was authenticated by a plugin.
+            if self.request.authenticated_userid == username:
+                registry = self.request.registry
+                auth_policy = registry.queryUtility(IAuthenticationPolicy)
+                # we verify the credentials explicitly here, because the
+                # provided token may belong to a deleted user
+                if auth_policy.verify_credentials(self.request):
+                    try:
+                        user = self.model.create_user(username, password=None)
+                    except InvalidUser as e:
+                        apireturn(400, "%s" % e)
         stage = self.context.user.getstage(self.context.index)
         if stage is not None:
             apireturn(409, "index %r exists" % stage.name)
@@ -418,6 +538,8 @@ class PyPIView:
         try:
             stage = self.context.user.create_stage(self.context.index, **kvdict)
             ixconfig = stage.ixconfig
+        except InvalidIndex as e:
+            apireturn(400, "%s" % e)
         except InvalidIndexconfig as e:
             apireturn(400, message=", ".join(e.messages))
         apireturn(200, type="indexconfig", result=ixconfig)
@@ -427,8 +549,6 @@ class PyPIView:
         permission="index_modify")
     def index_modify(self):
         stage = self.context.stage
-        if stage.name == "root/pypi":
-            apireturn(403, "root/pypi index config can not be modified")
         kvdict = getkvdict_index(self.xom.config.hook, getjson(self.request))
         try:
             ixconfig = stage.modify(**kvdict)
@@ -495,6 +615,8 @@ class PyPIView:
             except target_stage.NonVolatile as e:
                 apireturn(409, "%s already exists in non-volatile index" % (
                           e.link.basename,))
+            except BadGateway as e:
+                return apireturn(502, e.args[0])
             apireturn(200, result=results, type="actionlog")
         else:
             posturl = pushdata["posturl"]
@@ -544,6 +666,9 @@ class PyPIView:
 
     def _push_links(self, links, target_stage, name, version):
         for link in links["releasefile"]:
+            if should_fetch_remote_file(link.entry, self.request.headers):
+                for part in iter_fetch_remote_file(self.xom, link.entry):
+                    pass
             new_link = target_stage.store_releasefile(
                 name, version, link.basename, link.entry.file_get_content(),
                 last_modified=link.entry.last_modified)
@@ -593,7 +718,7 @@ class PyPIView:
         request = self.request
         context = self.context
         if context.username == "root" and context.index == "pypi":
-            abort_submit(404, "cannot submit to pypi mirror")
+            abort_submit(request, 404, "cannot submit to pypi mirror")
         stage = self.context.stage
         if not request.has_permission("pypi_submit"):
             # if there is no authenticated user, then issue a basic auth challenge
@@ -601,11 +726,11 @@ class PyPIView:
                 response = HTTPUnauthorized()
                 response.headers.update(forget(request))
                 return response
-            abort_submit(403, "no permission to submit")
+            abort_submit(request, 403, "no permission to submit")
         try:
             action = request.POST[":action"]
         except KeyError:
-            abort_submit(400, ":action field not found")
+            abort_submit(request, 400, ":action field not found")
         if action == "submit":
             self._set_versiondata_form(stage, request.POST)
             return Response("")
@@ -613,58 +738,84 @@ class PyPIView:
             try:
                 content = request.POST["content"]
             except KeyError:
-                abort_submit(400, "content file field not found")
+                abort_submit(request, 400, "content file field not found")
             name = ensure_unicode(request.POST.get("name"))
-            # version may be empty on plain uploads
+            # version may be empty on plain doczip uploads
             version = ensure_unicode(request.POST.get("version") or "")
-            projectname = stage.get_projectname(name)
-            if projectname is None:
-                abort_submit(400, "no project named %r was ever registered" % (name))
+            project = normalize_name(name)
+            if not stage.has_project(name):
+                abort_submit(
+                    request, 400,
+                    "no project named %r was ever registered" % (name))
+
             if action == "file_upload":
                 self.log.debug("metadata in form: %s",
                                list(request.POST.items()))
-                abort_if_invalid_filename(name, content.filename)
-                metadata = stage.get_versiondata_perstage(projectname, version)
+
+                # we only check for release files if version is
+                # contained in the filename because for doczip files
+                # we construct the filename ourselves anyway.
+                if version and version not in content.filename:
+                    abort_submit(
+                        request, 400,
+                        "filename %r does not contain version %r" % (
+                            content.filename, version))
+
+                abort_if_invalid_filename(request, name, content.filename)
+                metadata = stage.get_versiondata_perstage(project, version)
                 if not metadata:
                     self._set_versiondata_form(stage, request.POST)
-                    metadata = stage.get_versiondata(projectname, version)
+                    metadata = stage.get_versiondata(project, version)
                     if not metadata:
-                        abort_submit(400, "could not process form metadata")
+                        abort_submit(
+                            request, 400, "could not process form metadata")
                 file_content = content.file.read()
                 try:
                     link = stage.store_releasefile(
-                        projectname, version,
+                        project, version,
                         content.filename, file_content)
                 except stage.NonVolatile as e:
                     if e.link.matches_checksum(file_content):
-                        abort_submit(200,
-                            "Upload of identical file to non volatile index.")
-                    abort_submit(409, "%s already exists in non-volatile index" % (
-                         content.filename,))
+                        abort_submit(
+                            request, 200,
+                            "Upload of identical file to non volatile index.",
+                            level="info")
+                    abort_submit(
+                        request, 409,
+                        "%s already exists in non-volatile index" % (
+                            content.filename,))
                 link.add_log(
                     'upload', request.authenticated_userid, dst=stage.name)
                 try:
                     self.xom.config.hook.devpiserver_on_upload_sync(
                         log=request.log, application_url=request.application_url,
-                        stage=stage, projectname=projectname, version=version)
+                        stage=stage, project=project, version=version)
                 except Exception as e:
-                    abort_submit(200, "OK, but a trigger plugin failed: %s" % e)
+                    abort_submit(
+                        request, 200,
+                        "OK, but a trigger plugin failed: %s" % e, level="warn")
             else:
                 doczip = content.file.read()
                 try:
-                    link = stage.store_doczip(projectname, version, doczip)
+                    link = stage.store_doczip(project, version, doczip)
                 except stage.MissesRegistration:
-                    apireturn(400, "%s-%s is not registered" %(name, version))
+                    abort_submit(
+                        request, 400,
+                        "%s-%s is not registered" % (name, version))
                 except stage.NonVolatile as e:
                     if e.link.matches_checksum(doczip):
-                        abort_submit(200,
-                            "Upload of identical file to non volatile index.")
-                    abort_submit(409, "%s already exists in non-volatile index" % (
-                         content.filename,))
+                        abort_submit(
+                            request, 200,
+                            "Upload of identical file to non volatile index.",
+                            level="info")
+                    abort_submit(
+                        request, 409,
+                        "%s already exists in non-volatile index" % (
+                            content.filename,))
                 link.add_log(
                     'upload', request.authenticated_userid, dst=stage.name)
         else:
-            abort_submit(400, "action %r not supported" % action)
+            abort_submit(request, 400, "action %r not supported" % action)
         return Response("")
 
     def _set_versiondata_form(self, stage, form):
@@ -686,7 +837,7 @@ class PyPIView:
         try:
             stage.set_versiondata(metadata)
         except ValueError as e:
-            abort_submit(400, "invalid metadata: %s" % (e,))
+            abort_submit(self.request, 400, "invalid metadata: %s" % (e,))
         self.log.info("%s: got submit release info %r",
                  stage.name, metadata["name"])
 
@@ -696,12 +847,10 @@ class PyPIView:
 
     @view_config(route_name="simple_redirect")
     def simple_redirect(self):
-        stage, name = self.context.stage, self.context.name
-        projectname = stage.get_projectname(name)
-        real_name = projectname if projectname else name
-        redirect("/%s/+simple/%s" % (stage.name, real_name))
+        stage, name = self.context.stage, self.context.project
+        redirect("/%s/+simple/%s" % (stage.name, name))
 
-    @view_config(route_name="/{user}/{index}/{name}",
+    @view_config(route_name="/{user}/{index}/{project}",
                  accept="application/json", request_method="GET")
     def project_get(self):
         if not json_preferred(self.request):
@@ -715,33 +864,35 @@ class PyPIView:
         apireturn(200, type="projectconfig", result=view_metadata)
 
     @view_config(
-        route_name="/{user}/{index}/{name}", request_method="DELETE",
+        route_name="/{user}/{index}/{project}", request_method="DELETE",
         permission="del_project")
     def del_project(self):
         stage = self.context.stage
-        if stage.name == "root/pypi":
-            abort(self.request, 405, "cannot delete root/pypi index")
-        projectname = self.context.projectname
+        if stage.ixconfig["type"] == "mirror":
+            abort(self.request, 405, "cannot delete on mirror index")
+        project = self.context.project
         if not stage.ixconfig["volatile"]:
             apireturn(403, "project %r is on non-volatile index %s" %(
-                      projectname, stage.name))
-        stage.del_project(projectname)
-        apireturn(200, "project {name} deleted from stage {sname}".format(
-                  name=projectname, sname=stage.name))
+                      project, stage.name))
+        try:
+            stage.del_project(project)
+        except KeyError:
+            apireturn(404, "project not found")
+        apireturn(200, "project {project} deleted from stage {sname}".format(
+                  project=project, sname=stage.name))
 
-    @view_config(route_name="/{user}/{index}/{name}/{version}", accept="application/json", request_method="GET")
+    @view_config(route_name="/{user}/{index}/{project}/{version}", accept="application/json", request_method="GET")
     def version_get(self):
         verdata = self.context.get_versiondata(perstage=False)
         view_verdata = self._make_view_verdata(verdata)
         apireturn(200, type="versiondata", result=view_verdata)
 
     def _make_view_verdata(self, verdata):
-        view_verdata = copy_if_mutable(verdata)
+        view_verdata = get_mutable_deepcopy(verdata)
         elinks = view_verdata.pop("+elinks", None)
         if elinks is not None:
             view_verdata["+links"] = links = []
-            for elinkdict in elinks:
-                linkdict = copy_if_mutable(elinkdict)
+            for linkdict in elinks:
                 entrypath = linkdict.pop("entrypath")
                 linkdict["href"] = url_for_entrypath(self.request, entrypath)
                 for_entrypath = linkdict.pop("for_entrypath", None)
@@ -757,14 +908,14 @@ class PyPIView:
                     [self._make_view_verdata(x) for x in shadowing]
         return view_verdata
 
-    @view_config(route_name="/{user}/{index}/{name}/{version}",
+    @view_config(route_name="/{user}/{index}/{project}/{version}",
                  permission="del_verdata",
                  request_method="DELETE")
     def del_versiondata(self):
         stage = self.context.stage
-        name, version = self.context.name, self.context.version
-        if stage.name == "root/pypi":
-            abort(self.request, 405, "cannot delete on root/pypi index")
+        name, version = self.context.project, self.context.version
+        if stage.ixconfig["type"] == "mirror":
+            abort(self.request, 405, "cannot delete on mirror index")
         if not stage.ixconfig["volatile"]:
             abort(self.request, 403, "cannot delete version on non-volatile index")
         try:
@@ -785,24 +936,21 @@ class PyPIView:
         if json_preferred(request):
             if not entry or not entry.meta:
                 apireturn(404, "no such release file")
-            apireturn(200, type="releasefilemeta", result=entry.meta)
+            entry_data = get_mutable_deepcopy(entry.meta)
+            apireturn(200, type="releasefilemeta", result=entry_data)
         if not entry or not entry.meta:
             if entry is None:
                 abort(request, 404, "no such file")
             else:
                 abort(request, 410, "file existed, deleted in later serial")
 
-        if should_fetch_remote_file(entry, request.headers):
-            keyfs = self.xom.keyfs
-            try:
-                if not self.xom.is_replica():
-                    keyfs.restart_as_write_transaction()
-                    entry = filestore.get_file_entry(relpath)
-                    entry.cache_remote_file()
-                else:
-                    entry = entry.cache_remote_file_replica()
-            except entry.BadGateway as e:
-                return apireturn(502, e.args[0])
+        try:
+            if should_fetch_remote_file(entry, request.headers):
+                app_iter = iter_fetch_remote_file(self.xom, entry)
+                headers = next(app_iter)
+                return Response(app_iter=app_iter, headers=headers)
+        except entry.BadGateway as e:
+            return apireturn(502, e.args[0])
 
         headers = entry.gethttpheaders()
         if self.request.method == "HEAD":
@@ -815,7 +963,7 @@ class PyPIView:
     def index_get(self):
         stage = self.context.stage
         result = dict(stage.ixconfig)
-        result['projects'] = sorted(stage.list_projectnames_perstage())
+        result['projects'] = sorted(stage.list_projects_perstage())
         apireturn(200, type="indexconfig", result=result)
 
     #
@@ -841,11 +989,18 @@ class PyPIView:
         permission="user_modify")
     def user_patch(self):
         request = self.request
-        dict = getjson(request, allowed_keys=["email", "password"])
-        email = dict.get("email")
-        password = dict.get("password")
+        ignored_keys = set(('indexes', 'username'))
+        allowed_keys = set((
+            "email", "password", "title", "description", "custom_data"))
+        result = getjson(request, allowed_keys=allowed_keys.union(ignored_keys))
+        kvdict = dict()
+        for key in allowed_keys:
+            if key not in result:
+                continue
+            kvdict[key] = result[key]
         user = self.context.user
-        user.modify(password=password, email=email)
+        password = kvdict.get("password")
+        user.modify(**kvdict)
         if password is not None:
             apireturn(200, "user updated, new proxy auth",
                       type="userpassword",
@@ -916,6 +1071,19 @@ def should_fetch_remote_file(entry, headers):
     return should_fetch
 
 
+def iter_fetch_remote_file(xom, entry):
+    filestore = xom.filestore
+    keyfs = xom.keyfs
+    if not xom.is_replica():
+        keyfs.restart_as_write_transaction()
+        entry = filestore.get_file_entry(entry.relpath, readonly=False)
+        for part in entry.iter_cache_remote_file():
+            yield part
+    else:
+        for part in entry.iter_remote_file_replica():
+            yield part
+
+
 def url_for_entrypath(request, entrypath):
     parts = entrypath.split("/")
     user, index = parts[:2]
@@ -938,39 +1106,28 @@ def getjson(request, allowed_keys=None):
     return dict
 
 
-def abort_if_invalid_filename(name, filename):
+def abort_if_invalid_filename(request, name, filename):
     if not is_valid_archive_name(filename):
-        abort_submit(400, "%r is not a valid archive name" %(filename))
+        abort_submit(request, 400, "%r is not a valid archive name" %(filename))
     if normalize_name(filename).startswith(normalize_name(name)):
         return
-    abort_submit(400, "filename %r does not match project name %r"
+    abort_submit(request, 400, "filename %r does not match project name %r"
                       %(filename, name))
 
-def abort_if_invalid_projectname(request, projectname):
+def abort_if_invalid_project(request, project):
     try:
-        if isinstance(projectname, bytes):
-            projectname.decode("ascii")
+        if isinstance(project, bytes):
+            project.decode("ascii")
         else:
-            projectname.encode("ascii")
+            project.encode("ascii")
     except (UnicodeEncodeError, UnicodeDecodeError):
         abort(request, 400, "unicode project names not allowed")
 
 
 def getkvdict_index(hook, req):
-    req_volatile = req.get("volatile")
-    kvdict = {"volatile": True, "type": "stage", "bases": ["root/pypi"]}
-    if req_volatile is not None:
-        if req_volatile == False or (req_volatile != True and
-            req_volatile.lower() in ["false", "no"]):
-            kvdict["volatile"] = False
-    bases = req.get("bases")
-    if bases is not None:
-        if not isinstance(bases, list):
-            kvdict["bases"] = bases.split(",")
-        else:
-            kvdict["bases"] = bases
-    additional_keys = get_ixconfigattrs(hook, "stage") - set(('volatile', 'bases'))
-    for key in additional_keys:
+    kvdict = {}
+    ixconfigattrs = get_ixconfigattrs(hook, req.get("type", "stage"))
+    for key in ixconfigattrs:
         if key in req:
             kvdict[key] = req[key]
     return kvdict
