@@ -1,19 +1,24 @@
+from __future__ import print_function
 import re
-import logging
 from webtest.forms import Upload
 import webtest
 import mimetypes
+import subprocess
 
 import pytest
 import py
+import socket
+import sys
 from bs4 import BeautifulSoup
+from contextlib import closing
+from devpi_server import extpypi
 from devpi_server.config import get_pluginmanager
 from devpi_server.main import XOM, parseoptions
 from devpi_common.url import URL
-from devpi_server.extpypi import XMLProxy
 from devpi_server.extpypi import PyPIStage
 from devpi_server.log import threadlog, thread_clear_log
 from pyramid.authentication import b64encode
+from time import sleep
 
 import hashlib
 try:
@@ -21,6 +26,12 @@ try:
 except ImportError:
     from Queue import Queue as BaseQueue
 
+try:
+    import http.server as httpserver
+except ImportError:
+    import BaseHTTPServer as httpserver
+
+
 def make_file_url(basename, content, stagename=None, baseurl="http://localhost/"):
     from devpi_server.filestore import get_default_hash_spec, make_splitdir
     hash_spec = get_default_hash_spec(content)
@@ -39,6 +50,8 @@ log = threadlog
 def pytest_addoption(parser):
     parser.addoption("--slow", action="store_true", default=False,
         help="run slow tests involving remote services (pypi.python.org)")
+    parser.addoption("--backend", action="store",
+        help="run tests with specified dotted name backend")
 
 @pytest.fixture(autouse=True)
 def _clear():
@@ -62,15 +75,14 @@ def Queue():
 
 @pytest.fixture()
 def caplog(caplog):
-    """ enrich the pytest-capturelog funcarg. """
-    caplog.setLevel(logging.DEBUG)
+    """ enrich the pytest-catchlog funcarg. """
     def getrecords(msgrex=None, minlevel="DEBUG"):
         if msgrex is not None:
             msgrex = re.compile(msgrex)
         minlevelno = {"DEBUG": 10, "INFO": 20, "WARNING": 30,
                       "ERROR": 40, "FATAL": 50}.get(minlevel)
         recs = []
-        for rec in caplog.records():
+        for rec in caplog.records:
             if rec.levelno < minlevelno:
                 continue
             if msgrex is not None and not msgrex.search(rec.getMessage()):
@@ -97,9 +109,6 @@ def gentmp(request):
                 keep=0, rootdir=basedir, lock_timeout=None)
     return gentmp
 
-@pytest.fixture
-def xom_notmocked(request, makexom):
-    return makexom(mocking=False)
 
 @pytest.yield_fixture(autouse=True)
 def auto_transact(request):
@@ -122,19 +131,37 @@ def auto_transact(request):
 @pytest.fixture
 def xom(request, makexom):
     xom = makexom([])
+    request.addfinalizer(xom.keyfs.release_all_wait_tx)
     return xom
 
+
 @pytest.yield_fixture(autouse=True, scope="session")
-def speed_up_sql():
-    from devpi_server.keyfs import Filesystem
-    old = Filesystem.get_sqlconn
+def speed_up_sqlite():
+    from devpi_server.keyfs_sqlite import Storage
+    old = Storage.ensure_tables_exist
     def make_unsynchronous(self, old=old):
         conn = old(self)
-        conn.execute("PRAGMA synchronous=OFF")
-        return conn
-    Filesystem.get_sqlconn = make_unsynchronous
+        with self.get_connection() as conn:
+            conn._sqlconn.execute("PRAGMA synchronous=OFF")
+        return
+    Storage.ensure_tables_exist = make_unsynchronous
     yield
-    Filesystem.get_sqlconn = old
+    Storage.ensure_tables_exist = old
+
+
+@pytest.yield_fixture(autouse=True, scope="session")
+def speed_up_sqlite_fs():
+    from devpi_server.keyfs_sqlite_fs import Storage
+    old = Storage.ensure_tables_exist
+    def make_unsynchronous(self, old=old):
+        conn = old(self)
+        with self.get_connection() as conn:
+            conn._sqlconn.execute("PRAGMA synchronous=OFF")
+        return
+    Storage.ensure_tables_exist = make_unsynchronous
+    yield
+    Storage.ensure_tables_exist = old
+
 
 @pytest.fixture(scope="session")
 def mock():
@@ -144,35 +171,75 @@ def mock():
         import mock
     return mock
 
+
+@pytest.fixture(scope="session")
+def storage_info(request):
+    from pydoc import locate
+    backend = getattr(request.config.option, 'backend', None)
+    if backend is None:
+        backend = 'devpi_server.keyfs_sqlite_fs'
+    plugin = locate(backend)
+    result = plugin.devpiserver_storage_backend(settings=None)
+    result["_test_plugin"] = plugin
+    return result
+
+
+@pytest.fixture(scope="session")
+def storage(storage_info):
+    return storage_info['storage']
+
+
 @pytest.fixture
-def makexom(request, gentmp, httpget, monkeypatch, mock):
-    def makexom(opts=(), httpget=httpget, proxy=None, mocking=True, plugins=()):
+def makexom(request, gentmp, httpget, monkeypatch, storage_info):
+    def makexom(opts=(), httpget=httpget, plugins=()):
         from devpi_server import auth_basic
         from devpi_server import auth_devpi
-        plugins = list(plugins) + [(auth_basic, None), (auth_devpi, None)]
-        pm = get_pluginmanager()
+        plugins = [
+            plugin[0] if isinstance(plugin, tuple) else plugin
+            for plugin in plugins]
+        for plugin in [auth_basic, auth_devpi, storage_info["_test_plugin"]]:
+            if plugin not in plugins:
+                plugins.append(plugin)
+        pm = get_pluginmanager(load_entrypoints=False)
         for plugin in plugins:
             pm.register(plugin)
         serverdir = gentmp()
         fullopts = ["devpi-server", "--serverdir", serverdir] + list(opts)
+        if request.node.get_marker("with_replica_thread"):
+            fullopts.append("--master=http://localhost")
+        if not request.node.get_marker("no_storage_option"):
+            if storage_info["name"] != "sqlite":
+                fullopts.append("--storage=%s" % storage_info["name"])
         fullopts = [str(x) for x in fullopts]
         config = parseoptions(pm, fullopts)
         config.init_nodeinfo()
-        if mocking:
-            if proxy is None:
-                proxy = mock.create_autospec(XMLProxy)
-                proxy.list_packages_with_serial.return_value = {}
-            xom = XOM(config, proxy=proxy, httpget=httpget)
-            add_pypistage_mocks(monkeypatch, httpget)
-        else:
+        for marker in ("storage_with_filesystem",):
+            if request.node.get_marker(marker):
+                info = config._storage_info()
+                markers = info.get("_test_markers", [])
+                if marker not in markers:
+                    pytest.skip("The storage doesn't have marker '%s'." % marker)
+        if not request.node.get_marker("no_storage_option"):
+            assert storage_info["storage"] is config.storage
+        if request.node.get_marker("nomocking"):
             xom = XOM(config)
+        else:
+            xom = XOM(config, httpget=httpget)
+            if not request.node.get_marker("nomockprojectsremote"):
+                monkeypatch.setattr(extpypi.PyPIStage, "_get_remote_projects",
+                    lambda self: set())
+            add_pypistage_mocks(monkeypatch, httpget)
         # initialize default indexes
         from devpi_server.main import set_default_indexes
         if not xom.config.args.master_url:
             with xom.keyfs.transaction(write=True):
                 set_default_indexes(xom.model)
-        if mocking:
-            xom.pypimirror.init_pypi_mirror(proxy)
+        if request.node.get_marker("with_replica_thread"):
+            from devpi_server.replica import ReplicaThread
+            rt = ReplicaThread(xom)
+            xom.replica_thread = rt
+            xom.thread_pool.register(rt)
+            xom.thread_pool.start_one(rt)
         if request.node.get_marker("start_threads"):
             xom.thread_pool.start()
         elif request.node.get_marker("with_notifier"):
@@ -184,10 +251,10 @@ def makexom(request, gentmp, httpget, monkeypatch, mock):
 
 @pytest.fixture
 def replica_xom(request, makexom):
-    from devpi_server.replica import PyPIProxy
+    from devpi_server.replica import register_key_subscribers
     master_url = "http://localhost:3111"
     xom = makexom(["--master", master_url])
-    xom.proxy = PyPIProxy(xom._httpsession, xom.config.master_url)
+    register_key_subscribers(xom)
     return xom
 
 
@@ -210,6 +277,26 @@ def makemapp(request, maketestapp, makexom):
         return m
     return makemapp
 
+
+def make_simple_pkg_info(name, text="", pkgver=None, hash_type=None,
+                         pypiserial=None):
+    class ret:
+        hash_spec = ""
+    if pkgver is not None:
+        assert not text
+        if hash_type and "#" not in pkgver:
+            hv = (pkgver + str(pypiserial)).encode("ascii")
+            hash_value = getattr(hashlib, hash_type)(hv).hexdigest()
+            ret.hash_spec = "%s=%s" %(hash_type, hash_value)
+            pkgver += "#" + ret.hash_spec
+        text = '<a href="../../{name}/{pkgver}" />'.format(name=name, pkgver=pkgver)
+    elif text and "{md5}" in text:
+        text = text.format(md5=getmd5(text))
+    elif text and "{sha256}" in text:
+        text = text.format(sha256=getsha256(text))
+    return ret, text
+
+
 @pytest.fixture
 def httpget(pypiurls):
     class MockHTTPGet:
@@ -227,6 +314,8 @@ def httpget(pypiurls):
                     if "url" not in fakeresponse:
                         xself.url = url
                     xself.allow_redirects = allow_redirects
+                    if "content" in fakeresponse:
+                        xself.raw = py.io.BytesIO(fakeresponse["content"])
                 def __repr__(xself):
                     return "<mockresponse %s url=%s>" % (xself.status_code,
                                                          xself.url)
@@ -240,44 +329,23 @@ def httpget(pypiurls):
             log.debug("set mocking response %s %s", mockurl, kw)
             self.url2response[mockurl] = kw
 
-        def mock_simple(self, name, text=None, pkgver=None, hash_type=None,
-            pypiserial=10000, **kw):
-            class ret:
-                hash_spec = ""
-            if pkgver is not None:
-                assert not text
-                if hash_type and "#" not in pkgver:
-                    hv = (pkgver + str(pypiserial)).encode("ascii")
-                    hash_value = getattr(hashlib, hash_type)(hv).hexdigest()
-                    ret.hash_spec = "%s=%s" %(hash_type, hash_value)
-                    pkgver += "#" + ret.hash_spec
-                text = '<a href="../../pkg/{pkgver}" />'.format(pkgver=pkgver)
-            elif text and "{md5}" in text:
-                text = text.format(md5=getmd5(text))
-            elif text and "{sha256}" in text:
-                text = text.format(sha256=getsha256(text))
+        def mock_simple(self, name, text="", pkgver=None, hash_type=None,
+                        pypiserial=10000, remoteurl=None, **kw):
+            ret, text = make_simple_pkg_info(
+                name, text=text, pkgver=pkgver, hash_type=hash_type,
+                pypiserial=pypiserial)
+            if remoteurl is None:
+                remoteurl = pypiurls.simple
             headers = kw.setdefault("headers", {})
-            headers["X-PYPI-LAST-SERIAL"] = pypiserial
-            self.mockresponse(pypiurls.simple + name + "/",
-                                      text=text, **kw)
+            if pypiserial is not None:
+                headers["X-PYPI-LAST-SERIAL"] = str(pypiserial)
+            self.mockresponse(remoteurl + name + "/", text=text, **kw)
             return ret
 
         def _getmd5digest(self, s):
             self._md5.update(s.encode("utf8"))
             return self._md5.hexdigest()
 
-        def setextfile(self, path, content, **kw):
-            headers = {"content-length": len(content),
-                       "content-type": mimetypes.guess_type(path),
-                       "last-modified": "today",}
-            if path.startswith("/") and pypiurls.base.endswith("/"):
-                path = path.lstrip("/")
-            return self.mockresponse(pypiurls.base + path,
-                                     raw=py.io.BytesIO(content),
-                                     headers=headers,
-                                     **kw)
-
-
     return MockHTTPGet()
 
 @pytest.fixture
@@ -294,30 +362,46 @@ def model(xom):
 
 @pytest.fixture
 def pypistage(xom):
-    return PyPIStage(xom)
+    from devpi_server.main import _pypi_ixconfig_default
+    return PyPIStage(xom, username="root", index="pypi",
+                     ixconfig=_pypi_ixconfig_default)
+
+@pytest.fixture
+def replica_pypistage(replica_xom):
+    return pypistage(replica_xom)
 
 def add_pypistage_mocks(monkeypatch, httpget):
     # add some mocking helpers
     PyPIStage.url2response = httpget.url2response
     def mock_simple(self, name, text=None, pypiserial=10000, **kw):
-        call = lambda: \
-                 self.pypimirror.process_changelog([(name, 0,0,0, pypiserial)])
-        if not hasattr(self.keyfs, "tx"):
-            with self.keyfs.transaction(write=True):
-                call()
-        else:
-            call()
+        self.cache_link_updates.expire(name)
         return self.httpget.mock_simple(name,
-                text=text, pypiserial=pypiserial, **kw)
+                 text=text, pypiserial=pypiserial, **kw)
     monkeypatch.setattr(PyPIStage, "mock_simple", mock_simple, raising=False)
 
+    def mock_simple_projects(self, projectlist):
+        t = "".join('<a href="%s">%s</a>\n' % (name, name) for name in projectlist)
+        threadlog.debug("patching simple page with: %s" %(t))
+        self.httpget.mockresponse(self.mirror_url, code=200, text=t)
+
+    monkeypatch.setattr(PyPIStage, "mock_simple_projects",
+                        mock_simple_projects, raising=False)
+
+    def mock_extfile(self, path, content, **kw):
+        headers = {"content-length": len(content),
+                   "content-type": mimetypes.guess_type(path),
+                   "last-modified": "today",}
+        url = URL(self.mirror_url).joinpath(path)
+        return self.httpget.mockresponse(url.url, raw=py.io.BytesIO(content),
+                                         headers=headers, **kw)
+    monkeypatch.setattr(PyPIStage, "mock_extfile", mock_extfile, raising=False)
+
 @pytest.fixture
 def pypiurls():
-    from devpi_server.extpypi import PYPIURL, PYPIURL_SIMPLE
+    from devpi_server.main import _pypi_ixconfig_default
     class PyPIURL:
         def __init__(self):
-            self.base = PYPIURL
-            self.simple = PYPIURL_SIMPLE
+            self.simple = _pypi_ixconfig_default['mirror_url']
     return PyPIURL()
 
 
@@ -376,7 +460,7 @@ class Mapp(MappMixin):
     def getindexlist(self, user=None):
         if user is None:
             user = self.testapp.auth[0]
-        r = self.testapp.get("/%s" % user, {"Accept": "*/json"})
+        r = self.testapp.get("/%s" % user, accept="application/json")
         assert r.status_code == 200
         name = r.json["result"]["username"]
         result = {}
@@ -384,6 +468,33 @@ class Mapp(MappMixin):
             result["%s/%s" % (name, index)] = data
         return result
 
+    def getpkglist(self, user=None, indexname=None):
+        indexname = self._getindexname(indexname)
+        if user is None:
+            user = self.testapp.auth[0]
+        r = self.testapp.get("/%s" % indexname, accept="application/json")
+        assert r.status_code == 200
+        return r.json["result"]["projects"]
+
+    def getreleaseslist(self, name, code=200, user=None, indexname=None):
+        indexname = self._getindexname(indexname)
+        if user is None:
+            user = self.testapp.auth[0]
+        r = self.testapp.get("/%s/%s" % (indexname, name), accept="application/json")
+        assert r.status_code == 200
+        result = r.json["result"]
+        links = set()
+        for version in result.values():
+            for link in version["+links"]:
+                links.add(link["href"])
+        return sorted(links)
+
+    def downloadrelease(self, code, url):
+        r = self.testapp.xget(code, url)
+        if code < 300:
+            return r.body
+        return r.json
+
     def change_password(self, user, password):
         r = self.testapp.patch_json("/%s" % user, dict(password=password))
         assert r.status_code == 200
@@ -401,12 +512,12 @@ class Mapp(MappMixin):
             assert res["username"] == user
             assert res.get("email") == email
 
-    def modify_user(self, user, code=200, password=None, email=None):
+    def modify_user(self, user, code=200, password=None, **kwargs):
         reqdict = {}
         if password:
             reqdict["password"] = password
-        if email:
-            reqdict["email"] = email
+        for key, value in kwargs.items():
+            reqdict[key] = value
         r = self.testapp.patch_json("/%s" % user, reqdict, expect_errors=True)
         assert r.status_code == code
         if code == 200:
@@ -448,7 +559,7 @@ class Mapp(MappMixin):
                                   expect_errors=True)
         assert r.status_code == code
         if code in (200,201):
-            assert r.json["result"]["type"] == "stage"
+            assert r.json["result"]["type"] == indexconfig.get("type", "stage")
             return self.use("%s/%s" % (user, index))
         if code in (400,):
             return r.json["message"]
@@ -463,7 +574,7 @@ class Mapp(MappMixin):
                                   expect_errors=True)
         assert r.status_code == code
         if code in (200,201):
-            assert r.json["result"]["type"] == "stage"
+            assert r.json["result"]["type"] == indexconfig.get("type", "stage")
             return r.json["result"]
         if code in (400,):
             return r.json["message"]
@@ -481,11 +592,14 @@ class Mapp(MappMixin):
         assert r.status_code == code
 
     def set_custom_data(self, data, indexname=None):
+        return self.set_key_value("custom_data", data, indexname=indexname)
+
+    def set_key_value(self, key, value, indexname=None):
         indexname = self._getindexname(indexname)
         indexurl = "/" + indexname
         r = self.testapp.get_json(indexurl)
         result = r.json["result"]
-        result["custom_data"] = data
+        result[key] = value
         r = self.testapp.patch_json(indexurl, result)
         assert r.status_code == 200
 
@@ -499,14 +613,11 @@ class Mapp(MappMixin):
         assert r.status_code == 200
         assert r.json["result"][key] == value
 
-    def set_pypi_whitelist(self, whitelist, indexname=None):
+    def set_mirror_whitelist(self, whitelist, indexname=None):
         indexname = self._getindexname(indexname)
         r = self.testapp.get_json("/%s" % indexname)
         result = r.json["result"]
-        if not isinstance(whitelist, list):
-            whitelist = whitelist.split(",")
-        assert isinstance(whitelist, list)
-        result["pypi_whitelist"] = whitelist
+        result["mirror_whitelist"] = whitelist
         r = self.testapp.patch_json("/%s" % (indexname,), result)
         assert r.status_code == 200
 
@@ -526,16 +637,16 @@ class Mapp(MappMixin):
         r = self.testapp.get_json("/%s" % indexname)
         return r.json["result"].get("acl_" + acltype, None)
 
-    def get_pypi_whitelist(self, indexname=None):
+    def get_mirror_whitelist(self, indexname=None):
         indexname = self._getindexname(indexname)
         r = self.testapp.get_json("/%s" % indexname)
-        return r.json["result"]["pypi_whitelist"]
+        return r.json["result"]["mirror_whitelist"]
 
-    def delete_project(self, projectname, code=200, indexname=None,
+    def delete_project(self, project, code=200, indexname=None,
                        waithooks=False):
         indexname = self._getindexname(indexname)
         r = self.testapp.delete_json("/%s/%s" % (indexname,
-                projectname), {}, expect_errors=True)
+                project), {}, expect_errors=True)
         assert r.status_code == code
         if waithooks:
             self._wait_for_serial_in_result(r)
@@ -550,9 +661,9 @@ class Mapp(MappMixin):
                               expect_errors=True)
         assert r.status_code == code
         if r.status_code == 200 and set_whitelist:
-            whitelist = set(self.get_pypi_whitelist(indexname=indexname))
+            whitelist = set(self.get_mirror_whitelist(indexname=indexname))
             whitelist.add(metadata["name"])
-            self.set_pypi_whitelist(sorted(whitelist), indexname=indexname)
+            self.set_mirror_whitelist(sorted(whitelist), indexname=indexname)
         if waithooks:
             self._wait_for_serial_in_result(r)
         return r
@@ -591,11 +702,11 @@ class Mapp(MappMixin):
         assert r.status_code == code
         return r
 
-    def get_release_paths(self, projectname):
-        r = self.get_simple(projectname)
+    def get_release_paths(self, project):
+        r = self.get_simple(project)
         pkg_url = URL(r.request.url)
         paths = [pkg_url.joinpath(link["href"]).path
-                 for link in BeautifulSoup(r.body).findAll("a")]
+                 for link in BeautifulSoup(r.body, "html.parser").findAll("a")]
         return paths
 
     def upload_doc(self, basename, content, name, version, indexname=None,
@@ -618,8 +729,8 @@ class Mapp(MappMixin):
             self._wait_for_serial_in_result(r)
         return r
 
-    def get_simple(self, projectname, code=200):
-        r = self.testapp.get(self.api.simpleindex + projectname,
+    def get_simple(self, project, code=200):
+        r = self.testapp.get(self.api.simpleindex + project,
                              expect_errors=True)
         assert r.status_code == code
         return r
@@ -695,9 +806,16 @@ class MyTestApp(TApp):
         if accept is not None:
             headers = kwargs.setdefault("headers", {})
             headers[str("Accept")] = str(accept)
-        return self._gen_request("GET", *args, **kwargs)
+        follow = kwargs.pop("follow", True)
+        response = self._gen_request("GET", *args, **kwargs)
+        if follow and response.status_code == 302:
+            assert response.location != args[0]
+            return self.get(response.location, *args[1:], **kwargs)
+        return response
 
     def xget(self, code, *args, **kwargs):
+        if code == 302:
+            kwargs["follow"] = False
         r = self.get(*args, **kwargs)
         assert r.status_code == code
         return r
@@ -713,7 +831,7 @@ class MyTestApp(TApp):
         headers = kwargs.setdefault("headers", {})
         headers["Accept"] = "application/json"
         self.x = 1
-        return super(MyTestApp, self).get(*args, **kwargs)
+        return self.get(*args, **kwargs)
 
 
 
@@ -722,6 +840,309 @@ class MyTestApp(TApp):
 def testapp(request, maketestapp, xom):
     return maketestapp(xom)
 
+
+class SimPyPIRequestHandler(httpserver.BaseHTTPRequestHandler):
+    def do_GET(self):
+        def start_response(status, headers):
+            self.send_response(status)
+            for key, value in headers.items():
+                self.send_header(key, value)
+            self.end_headers()
+
+        simpypi = self.server.simpypi
+        headers = {
+            'X-Simpypi-Method': 'GET'}
+        p = self.path.split('/')
+        if len(p) == 4 and p[0] == '' and p[1] == 'simple' and p[3] == '':
+            # project listing
+            project = simpypi.projects.get(p[2])
+            if project is not None:
+                releases = project['releases']
+                simpypi.add_log(
+                    "do_GET", self.path, "found",
+                    project['title'], "with", list(releases))
+                start_response(200, headers)
+                self.wfile.write(b'\n'.join(releases))
+                return
+        elif p == ['', 'simple', '']:
+            # root listing
+            projects = [
+                '<a href="%s">%s</a>' % (k, v['title'])
+                for k, v in simpypi.projects.items()]
+            simpypi.add_log("do_GET", self.path, "found", list(simpypi.projects))
+            start_response(200, headers)
+            self.wfile.write(b'\n'.join(x.encode('utf-8') for x in projects))
+            return
+        elif self.path in simpypi.files:
+            # file serving
+            f = simpypi.files[self.path]
+            content = f['content']
+            if 'length' in f:
+                headers['Content-Length'] = f['length']
+                content = content[:f['length']]
+            start_response(200, headers)
+            simpypi.add_log("do_GET", self.path, "sending")
+            if not f['stream']:
+                self.wfile.write(content)
+                simpypi.add_log("do_GET", self.path, "sent")
+                return
+            else:
+                chunksize = f['chunksize']
+                callback = f.get('callback')
+                for i in range(len(content) // chunksize):
+                    data = content[i * chunksize:(i + 1) * chunksize]
+                    if not data:
+                        break
+                    self.wfile.write(data)
+                    if callback:
+                        callback(i * chunksize)
+                    simpypi.add_log(
+                        "do_GET", self.path,
+                        "streamed %i bytes" % (i * chunksize))
+                return
+        simpypi.add_log("do_GET", self.path, "not found")
+        start_response(404, headers)
+
+
+def get_open_port(host):
+    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
+        s.bind((host, 0))
+        s.listen(1)
+        port = s.getsockname()[1]
+    return port
+
+
+def wait_for_port(host, port, timeout=60):
+    while timeout > 0:
+        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
+            s.settimeout(1)
+            if s.connect_ex((host, port)) == 0:
+                return
+        sleep(1)
+        timeout -= 1
+    raise RuntimeError(
+        "The port %s on host %s didn't become accessible" % (port, host))
+
+
+@pytest.yield_fixture(scope="module")
+def server_directory():
+    import tempfile
+    srvdir = py.path.local(
+        tempfile.mkdtemp(prefix='test-', suffix='-server-directory'))
+    yield srvdir
+    srvdir.remove(ignore_errors=True)
+
+
+@pytest.fixture(scope="module")
+def call_devpi_in_dir():
+    # let xproc find the correct executable instead of py.test
+    devpiserver = str(py.path.local.sysfind("devpi-server"))
+
+    def devpi(server_dir, args):
+        from devpi_server.main import main
+        from _pytest.monkeypatch import MonkeyPatch
+        m = MonkeyPatch()
+        m.setenv("DEVPI_SERVERDIR", server_dir)
+        m.setattr("sys.argv", [devpiserver])
+        try:
+            main(args)
+        finally:
+            m.undo()
+
+    return devpi
+
+
+@pytest.yield_fixture(scope="module")
+def master_host_port(request, call_devpi_in_dir, server_directory):
+    host = 'localhost'
+    port = get_open_port(host)
+    master_dir = server_directory.join("master").strpath
+    call_devpi_in_dir(
+        master_dir,
+        ["devpi-server", "--start", "--host", host, "--port", str(port)])
+    try:
+        wait_for_port(host, port)
+        yield (host, port)
+    finally:
+        call_devpi_in_dir(
+            master_dir,
+            ["devpi-server", "--stop"])
+
+
+@pytest.yield_fixture(scope="module")
+def replica_host_port(request, call_devpi_in_dir, server_directory):
+    host = 'localhost'
+    port = get_open_port(host)
+    replica_dir = server_directory.join("replica").strpath
+    call_devpi_in_dir(
+        replica_dir,
+        ["devpi-server", "--start", "--host", host, "--port", str(port)])
+    try:
+        wait_for_port(host, port)
+        yield (host, port)
+    finally:
+        call_devpi_in_dir(
+            replica_dir,
+            ["devpi-server", "--stop"])
+
+
+nginx_conf_content = """
+worker_processes  1;
+daemon off;
+pid nginx.pid;
+error_log nginx_error.log;
+
+events {
+    worker_connections  32;
+}
+
+http {
+    access_log off;
+    default_type  application/octet-stream;
+    sendfile        on;
+    keepalive_timeout 0;
+    include nginx-devpi.conf;
+}
+"""
+
+
+def _nginx_host_port(host, port, call_devpi_in_dir, server_directory):
+    # let xproc find the correct executable instead of py.test
+    nginx = py.path.local.sysfind("nginx")
+    if nginx is None:
+        pytest.skip("No nginx executable found.")
+    nginx = str(nginx)
+
+    orig_dir = server_directory.chdir()
+    try:
+        call_devpi_in_dir(
+            server_directory.strpath,
+            ["devpi-server", "--gen-config", "--host", host, "--port", str(port)])
+    finally:
+        orig_dir.chdir()
+    nginx_directory = server_directory.join("gen-config")
+    nginx_devpi_conf = nginx_directory.join("nginx-devpi.conf")
+    nginx_port = get_open_port(host)
+    nginx_devpi_conf_content = nginx_devpi_conf.read()
+    nginx_devpi_conf_content = nginx_devpi_conf_content.replace(
+        "listen 80;",
+        "listen %s;" % nginx_port)
+    nginx_devpi_conf.write(nginx_devpi_conf_content)
+    nginx_conf = nginx_directory.join("nginx.conf")
+    nginx_conf.write(nginx_conf_content)
+    try:
+        subprocess.check_output([
+            nginx, "-t",
+            "-c", nginx_conf.strpath,
+            "-p", nginx_directory.strpath], stderr=subprocess.STDOUT)
+    except subprocess.CalledProcessError as e:
+        print(e.output, file=sys.stderr)
+        raise
+    p = subprocess.Popen([
+        nginx, "-c", nginx_conf.strpath, "-p", nginx_directory.strpath])
+    wait_for_port(host, nginx_port)
+    return (p, nginx_port)
+
+
+@pytest.yield_fixture(scope="module")
+def nginx_host_port(request, call_devpi_in_dir, server_directory):
+    if sys.platform.startswith("win"):
+        pytest.skip("no nginx on windows")
+    # we need the skip above before master_host_port is called
+    (host, port) = request.getfuncargvalue("master_host_port")
+    (p, nginx_port) = _nginx_host_port(
+        host, port, call_devpi_in_dir, server_directory)
+    try:
+        yield (host, nginx_port)
+    finally:
+        p.terminate()
+        p.wait()
+
+
+@pytest.yield_fixture(scope="module")
+def nginx_replica_host_port(replica_host_port, call_devpi_in_dir, server_directory):
+    (host, port) = replica_host_port
+    (p, nginx_port) = _nginx_host_port(
+        host, port, call_devpi_in_dir, server_directory)
+    try:
+        yield (host, nginx_port)
+    finally:
+        p.terminate()
+        p.wait()
+
+
+@pytest.fixture(scope="session")
+def simpypiserver():
+    import threading
+    host = 'localhost'
+    port = get_open_port(host)
+    server = httpserver.HTTPServer((host, port), SimPyPIRequestHandler)
+    thread = threading.Thread(target=server.serve_forever)
+    thread.daemon = True
+    thread.start()
+    wait_for_port(host, port, 5)
+    print("Started simpypi server %s:%s" % server.server_address)
+    return server
+
+
+class SimPyPI:
+    def __init__(self, address):
+        self.baseurl = "http://%s:%s" % address
+        self.simpleurl = "%s/simple" % self.baseurl
+        self.projects = {}
+        self.files = {}
+        self.clear_log()
+
+    def clear_log(self):
+        self.log = []
+
+    def add_log(self, *args):
+        msg = ' '.join(str(x) for x in args)
+        print(msg, file=sys.stderr)
+        self.log.append(msg)
+
+    def add_project(self, name, title=None):
+        if title is None:
+            title = name
+        self.projects.setdefault(
+            name, dict(
+                title=title,
+                releases=set(),
+                pypiserial=None))
+        return self.projects[name]
+
+    def remove_project(self, name):
+        self.projects.pop(name)
+
+    def add_release(self, name, title=None, text="", pkgver=None, hash_type=None,
+                    pypiserial=None, **kw):
+        project = self.add_project(name, title=title)
+        ret, text = make_simple_pkg_info(
+            name, text=text, pkgver=pkgver, hash_type=hash_type,
+            pypiserial=pypiserial)
+        assert text
+        project['releases'].add(text.encode('utf-8'))
+
+    def add_file(self, relpath, content, stream=False, chunksize=1024,
+                 length=None, callback=None):
+        if length is None:
+            length = len(content)
+        info = dict(
+            content=content,
+            stream=stream,
+            chunksize=chunksize,
+            callback=callback)
+        if length is not False:
+            info['length'] = length
+        self.files[relpath] = info
+
+
+@pytest.fixture
+def simpypi(simpypiserver):
+    simpypiserver.simpypi = SimPyPI(simpypiserver.server_address)
+    return simpypiserver.simpypi
+
+
 ### incremental testing
 
 def pytest_runtest_makereport(item, call):
@@ -840,13 +1261,21 @@ def getsha256(s):
 
 
 @pytest.yield_fixture
-def dummyrequest():
-    from pyramid.testing import DummyRequest, setUp, tearDown
-    request = DummyRequest()
-    setUp(request=request)
-    yield request
+def pyramidconfig():
+    from pyramid.testing import setUp, tearDown
+    config = setUp()
+    yield config
     tearDown()
 
+
+@pytest.yield_fixture
+def dummyrequest(pyramidconfig):
+    from pyramid.testing import DummyRequest
+    request = DummyRequest()
+    pyramidconfig.begin(request=request)
+    yield request
+
+
 @pytest.fixture
 def blank_request():
     from pyramid.request import Request
