@@ -1,12 +1,14 @@
 from __future__ import unicode_literals
 
 import sys
+import pkg_resources
+import py
 import pytest
 import json
-import time
 from devpi_server.importexport import *
 from devpi_server.main import Fatal
 from devpi_common.archive import Archive, zip_dict
+from devpi_common.metadata import Version
 
 import devpi_server
 
@@ -37,6 +39,23 @@ def test_empty_export(tmpdir, xom):
     with pytest.raises(Fatal):
         make_export(tmpdir, xom)
 
+
+def test_empty_serverdir(tmpdir, capfd, monkeypatch):
+    from devpi_server.main import main
+    empty = tmpdir.join("empty").ensure(dir=True)
+    export = tmpdir.join("export")
+    monkeypatch.setattr("devpi_server.main.configure_logging", lambda a: None)
+    ret = main([
+        "devpi-server",
+        "--serverdir", empty.strpath,
+        "--export", export.strpath])
+    out, err = capfd.readouterr()
+    assert empty.listdir() == []
+    assert ret == 1
+    assert out == ''
+    assert ("The path '%s' contains no devpi-server data." % empty) in err
+
+
 def test_import_on_existing_server_data(tmpdir, xom):
     with xom.keyfs.transaction(write=True):
         xom.model.create_user("someuser", password="qwe")
@@ -71,6 +90,9 @@ class TestImportExport:
         class ImpExp:
             def __init__(self, options=()):
                 self.exportdir = gentmp()
+                self.testdatadir = py.path.local(
+                    pkg_resources.resource_filename(
+                        'test_devpi_server', 'importexportdata'))
                 self.mapp1 = makemapp(
                     options=("--export", self.exportdir) + options)
 
@@ -79,6 +101,16 @@ class TestImportExport:
                     self.mapp1.xom.config.init_nodeinfo()
                 assert self.mapp1.xom.main() == 0
 
+            def import_testdata(self, name, options=(), plugin=None):
+                path = self.testdatadir.join(name).strpath
+                mapp = makemapp(
+                    options=("--import", path) + options)
+                if plugin is not None:
+                    mapp.xom.config.pluginmanager.register(plugin)
+                mapp.xom.config.init_nodeinfo()
+                assert mapp.xom.main() == 0
+                return mapp
+
             def new_import(self, options=(), plugin=None):
                 mapp2 = makemapp(
                     options=("--import", str(self.exportdir)) + options)
@@ -93,6 +125,26 @@ class TestImportExport:
     def impexp(self, makeimpexp):
         return makeimpexp()
 
+    def test_importing_multiple_indexes_with_releases(self, impexp):
+        mapp1 = impexp.mapp1
+        api1 = mapp1.create_and_use()
+        content = b'content1'
+        mapp1.upload_file_pypi("hello-1.0.tar.gz", content, "hello", "1.0")
+        path, = mapp1.get_release_paths("hello")
+        path = path.strip("/")
+        stagename2 = api1.user + "/" + "dev6"
+        api2 = mapp1.create_index(stagename2)
+        content = b'content2'
+        mapp1.upload_file_pypi("pkg1-1.0.tar.gz", content, "pkg1", "1.0")
+        impexp.export()
+        mapp2 = impexp.new_import()
+        mapp2.use(api1.stagename)
+        assert mapp2.get_release_paths('hello') == [
+            '/user1/dev/+f/d0b/425e00e15a0d3/hello-1.0.tar.gz']
+        mapp2.use(api2.stagename)
+        assert mapp2.get_release_paths('pkg1') == [
+            '/user1/dev6/+f/dab/741b6289e7dcc/pkg1-1.0.tar.gz']
+
     def test_uuid(self, impexp):
         impexp.export()
         mapp2 = impexp.new_import()
@@ -122,6 +174,59 @@ class TestImportExport:
         indexlist = mapp2.getindexlist(api.user)
         assert indexlist[api.stagename]["custom_data"] == 42
 
+    def test_indexes_mirror_whitelist(self, impexp):
+        mapp1 = impexp.mapp1
+        api = mapp1.create_and_use()
+        mapp1.set_mirror_whitelist("*")
+        impexp.export()
+        mapp2 = impexp.new_import()
+        assert api.user in mapp2.getuserlist()
+        indexlist = mapp2.getindexlist(api.user)
+        assert indexlist[api.stagename]["mirror_whitelist"] == ["*"]
+
+    def test_bad_username(self, caplog, impexp):
+        with pytest.raises(SystemExit):
+            impexp.import_testdata('badusername')
+        (record,) = caplog.getrecords('contains characters')
+        assert 'root~foo.com' in record.message
+        (record,) = caplog.getrecords('You should edit')
+        assert 'dataindex.json' in record.message
+
+    def test_bad_indexname(self, caplog, impexp):
+        with pytest.raises(SystemExit):
+            impexp.import_testdata('badindexname')
+        (record,) = caplog.getrecords('contains characters')
+        assert 'root/pypi!Jo' in record.message
+        (record,) = caplog.getrecords('You should edit')
+        assert 'dataindex.json' in record.message
+
+    def test_normalization(self, caplog, impexp):
+        mapp = impexp.import_testdata('normalization')
+        with mapp.xom.keyfs.transaction(write=False):
+            stage = mapp.xom.model.getstage('root/dev')
+            links = stage.get_releaselinks("hello.pkg")
+            assert len(links) == 1
+            assert links[0].project == "hello-pkg"
+            link = stage.get_link_from_entrypath(links[0].entrypath)
+            assert link.entry.file_get_content() == b"content"
+
+    def test_normalization_merge(self, caplog, impexp):
+        mapp = impexp.import_testdata('normalization_merge')
+        with mapp.xom.keyfs.transaction(write=False):
+            stage = mapp.xom.model.getstage('root/dev')
+            links = sorted(
+                stage.get_releaselinks("hello.pkg"),
+                key=lambda x: Version(x.version))
+            assert len(links) == 2
+            assert links[0].project == "hello-pkg"
+            assert links[1].project == "hello-pkg"
+            assert links[0].version == "1.0"
+            assert links[1].version == "1.1"
+            link = stage.get_link_from_entrypath(links[0].entrypath)
+            assert link.entry.file_get_content() == b"content1"
+            link = stage.get_link_from_entrypath(links[1].entrypath)
+            assert link.entry.file_get_content() == b"content2"
+
     def test_upload_releasefile_with_toxresult(self, impexp):
         from test_devpi_server.example import tox_result_data
         mapp1 = impexp.mapp1
@@ -148,7 +253,7 @@ class TestImportExport:
             assert len(results) == 1
             assert results[0] == tox_result_data
             linkstore = stage.get_linkstore_perstage(
-                link.projectname, link.version)
+                link.project, link.version)
             tox_link, = linkstore.get_links(rel="toxresult", for_entrypath=link)
             history_log = tox_link.get_logs()
             assert len(history_log) == 1
@@ -233,7 +338,7 @@ class TestImportExport:
             assert len(results) == 1
             assert results[0] == tox_result_data
             linkstore = stage.get_linkstore_perstage(
-                link.projectname, link.version)
+                link.project, link.version)
             tox_link, = linkstore.get_links(rel="toxresult", for_entrypath=link)
             history_log = tox_link.get_logs()
             assert len(history_log) == 1
@@ -251,7 +356,7 @@ class TestImportExport:
         with mapp1.xom.keyfs.transaction(write=True):
             stage = mapp1.xom.model.getstage(api.stagename)
             key_projversion = stage.key_projversion("hello", "1.0")
-            verdata = key_projversion.get()
+            verdata = key_projversion.get(readonly=False)
             del verdata["version"]
             key_projversion.set(verdata)
         impexp.export()
@@ -267,6 +372,34 @@ class TestImportExport:
             assert len(links) == 1
             assert links[0].entry.file_get_content() == b"content"
 
+    def test_same_filename_in_different_versions(self, impexp):
+        # for some unknown reason, the same filename can be uploaded in two
+        # different versions. Seems to be related to PEP440
+        mapp1 = impexp.mapp1
+        api = mapp1.create_and_use()
+        content1 = b'content1'
+        content2 = b'content2'
+        mapp1.upload_file_pypi("hello-1.0.foo.tar.gz", content1, "hello", "1.0")
+        mapp1.upload_file_pypi("hello-1.0.foo.tar.gz", content2, "hello", "1.0.foo")
+
+        impexp.export()
+        mapp2 = impexp.new_import()
+
+        with mapp2.xom.keyfs.transaction():
+            stage = mapp2.xom.model.getstage(api.stagename)
+            # first
+            verdata = stage.get_versiondata_perstage("hello", "1.0")
+            assert verdata["version"] == "1.0"
+            links = stage.get_linkstore_perstage("hello", "1.0").get_links()
+            assert len(links) == 1
+            assert links[0].entry.file_get_content() == b"content1"
+            # second
+            verdata = stage.get_versiondata_perstage("hello", "1.0.foo")
+            assert verdata["version"] == "1.0.foo"
+            links = stage.get_linkstore_perstage("hello", "1.0.foo").get_links()
+            assert len(links) == 1
+            assert links[0].entry.file_get_content() == b"content2"
+
     def test_dashes_in_name_issue199(self, impexp):
         mapp1 = impexp.mapp1
         api = mapp1.create_and_use()
@@ -421,34 +554,6 @@ class TestImportExport:
             assert py.builtin._totext(
                 archive.read("index.html"), 'utf-8') == "<html/>"
 
-    def test_multiple_docs_on_same_version(self, impexp):
-        mapp1 = impexp.mapp1
-        api = mapp1.create_and_use()
-        mapp1.set_versiondata({"name": "hello", "version": "1.0"})
-        with mapp1.xom.keyfs.transaction(write=True):
-            # create entries with and without log
-            stage = mapp1.xom.model.getstage(mapp1.current_stage)
-            link = stage.store_doczip(
-                "Hello", "1.0",
-                content=zip_dict({
-                    "index.html": "<html><body>Hello"}))
-            link.add_log('upload', stage.user.name, dst=stage.name)
-            time.sleep(1.1)  # force different times in log entry
-            link = stage.store_doczip(
-                "hello", "1.0",
-                content=zip_dict({
-                    "index.html": "<html><body>hello"}))
-            link.add_log('upload', stage.user.name, dst=stage.name)
-        impexp.export()
-        mapp2 = impexp.new_import()
-        with mapp2.xom.keyfs.transaction(write=False):
-            stage = mapp2.xom.model.getstage(api.stagename)
-            doczip = stage.get_doczip("hello", "1.0")
-            archive = Archive(py.io.BytesIO(doczip))
-            assert 'index.html' in archive.namelist()
-            assert py.builtin._totext(
-                archive.read("index.html"), 'utf-8') == "<html><body>hello"
-
     def test_name_mangling_relates_to_issue132(self, impexp):
         mapp1 = impexp.mapp1
         api = mapp1.create_and_use()
@@ -469,10 +574,10 @@ class TestImportExport:
 
             links = stage.get_releaselinks("he_llo")
             assert len(links) == 2
-            projectname = stage.get_projectname("he-llo")
-            links = stage.get_releaselinks(projectname)
+            links = stage.get_releaselinks("he-llo")
             assert len(links) == 2
 
+    @pytest.mark.storage_with_filesystem
     @pytest.mark.skipif(not hasattr(os, 'link'),
                         reason="OS doesn't support hard links")
     def test_hard_links(self, makeimpexp):
@@ -491,7 +596,7 @@ class TestImportExport:
         assert impexp.exportdir.join(
           'user1', 'dev', 'he_llo-1.0.doc.zip').stat().nlink == 2
         assert impexp.exportdir.join(
-          'user1', 'dev', 'he_llo', 'he-llo-1.0.tar.gz').stat().nlink == 2
+          'user1', 'dev', 'he-llo', '1.0', 'he-llo-1.0.tar.gz').stat().nlink == 2
 
         mapp2 = impexp.new_import()
 
@@ -530,6 +635,7 @@ class TestImportExport:
             assert "uploadtrigger_jenkins" not in stage.ixconfig
 
     def test_import_fails_if_uploadtrigger_jenkins_set(self, impexp):
+        from devpi_server.model import InvalidIndexconfig
         mapp1 = impexp.mapp1
         api = mapp1.create_and_use()
         (user, index) = api.stagename.split('/')
@@ -545,9 +651,9 @@ class TestImportExport:
 
         impexp.export()
 
-        with pytest.raises(TypeError) as excinfo:
+        with pytest.raises(InvalidIndexconfig) as excinfo:
             impexp.new_import()
-        assert "uploadtrigger_jenkins" in excinfo.value.args[0]
+        assert "uploadtrigger_jenkins" in excinfo.value.args[0][0]
 
     def test_plugin_index_config(self, impexp):
         class Plugin:
@@ -573,3 +679,68 @@ class TestImportExport:
             stage = mapp2.xom.model.getstage(api.stagename)
             assert "foo_plugin" in stage.ixconfig
             assert stage.ixconfig["foo_plugin"] == "foo"
+
+    @pytest.mark.nomockprojectsremote
+    def test_mirror_settings_preserved(self, httpget, impexp, pypiurls):
+        mapp1 = impexp.mapp1
+        indexconfig = dict(
+            type="mirror",
+            mirror_url="http://localhost:6543/index/",
+            mirror_cache_expiry="600")
+        api = mapp1.create_and_use(indexconfig=indexconfig)
+
+        impexp.export()
+
+        httpget.mockresponse(pypiurls.simple, code=200, text="")
+        httpget.mockresponse(indexconfig["mirror_url"], code=200, text="")
+
+        mapp2 = impexp.new_import()
+        result = mapp2.getjson(api.index)
+        assert result["type"] == "indexconfig"
+        assert result["result"] == dict(
+            type="mirror",
+            volatile=True,
+            mirror_url="http://localhost:6543/index/",
+            mirror_cache_expiry=600,
+            # XXX backward compatibility with devpi-client <= 2.4.1
+            acl_upload=[],
+            bases=[],
+            pypi_whitelist=[],
+            # XXX end
+            projects=[])
+
+    @pytest.mark.nomockprojectsremote
+    def test_no_mirror_releases_touched(self, httpget, impexp, pypiurls):
+        mapp1 = impexp.mapp1
+        indexconfig = dict(
+            type="mirror",
+            mirror_url="http://localhost:6543/index/")
+        api = mapp1.create_and_use(indexconfig=indexconfig)
+
+        httpget.mockresponse(
+            pypiurls.simple, code=200,
+            text='<a href="pytest">pytest</a>')
+        httpget.mockresponse(
+            indexconfig["mirror_url"], code=200,
+            text='<a href="devpi">devpi</a>')
+
+        impexp.export()
+
+        assert os.listdir(impexp.exportdir.strpath) == ['dataindex.json']
+
+        httpget.mockresponse(pypiurls.simple, code=200, text="")
+        httpget.mockresponse(indexconfig["mirror_url"], code=200, text="")
+
+        mapp2 = impexp.new_import()
+        result = mapp2.getjson(api.index)
+        assert result["type"] == "indexconfig"
+        assert result["result"] == dict(
+            type="mirror",
+            volatile=True,
+            mirror_url="http://localhost:6543/index/",
+            # XXX backward compatibility with devpi-client <= 2.4.1
+            acl_upload=[],
+            bases=[],
+            pypi_whitelist=[],
+            # XXX end
+            projects=[])
