@@ -18,7 +18,6 @@ def udict(**kw):
         d[py.builtin._totext(name)] = val
     return d
 
-
 @pytest.fixture(params=[(), ("root/pypi",)])
 def bases(request):
     return request.param
@@ -56,22 +55,67 @@ def stage(request, user):
 def user(model):
     return model.create_user("hello", password="123")
 
-def test_has_pypi_base(model, pypistage):
+def test_has_mirror_base(model, pypistage):
     pypistage.mock_simple("pytest", "<a href='pytest-1.0.zip' /a>")
-    assert pypistage.has_pypi_base("pytest")
-    assert pypistage.get_projectname_perstage("pytest")
+    assert pypistage.has_mirror_base("pytest")
+    assert pypistage.has_project_perstage("pytest")
     user = model.create_user("user1", "pass")
     stage1 = user.create_stage("stage1", bases=())
-    assert not stage1.has_pypi_base("pytest")
+    assert not stage1.has_mirror_base("pytest")
 
     stage2 = user.create_stage("stage2", bases=("root/pypi",))
-    assert stage2.has_pypi_base("pytest")
+    assert stage2.has_mirror_base("pytest")
     register_and_store(stage2, "pytest-1.1.tar.gz")
-    assert not stage2.has_pypi_base("pytest")
+    assert not stage2.has_mirror_base("pytest")
     ixconfig = stage2.ixconfig.copy()
-    ixconfig["pypi_whitelist"] = ["pytest"]
+    ixconfig["mirror_whitelist"] = ["pytest"]
     stage2.modify(**ixconfig)
-    assert stage2.has_pypi_base("pytest")
+    assert stage2.has_mirror_base("pytest")
+
+
+def test_get_mirror_whitelist_info(model, pypistage):
+    pypistage.mock_simple("pytest", "<a href='pytest-1.0.zip' /a>")
+    assert pypistage.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=True,
+        blocked_by_mirror_whitelist=None)
+    user = model.create_user("user1", "pass")
+    stage1 = user.create_stage("stage1", bases=())
+    assert stage1.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=False,
+        blocked_by_mirror_whitelist=None)
+    register_and_store(stage1, "pytest-1.1.tar.gz")
+    assert stage1.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=False,
+        blocked_by_mirror_whitelist=None)
+    stage2 = user.create_stage("stage2", bases=("root/pypi",))
+    assert stage2.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=True,
+        blocked_by_mirror_whitelist=None)
+    register_and_store(stage2, "pytest-1.1.tar.gz")
+    assert stage2.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=False,
+        blocked_by_mirror_whitelist='root/pypi')
+    # now add to whitelist
+    ixconfig = stage2.ixconfig.copy()
+    ixconfig["mirror_whitelist"] = ["pytest"]
+    stage2.modify(**ixconfig)
+    assert stage2.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=True,
+        blocked_by_mirror_whitelist=None)
+    # now remove from whitelist
+    ixconfig = stage2.ixconfig.copy()
+    ixconfig["mirror_whitelist"] = []
+    stage2.modify(**ixconfig)
+    assert stage2.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=False,
+        blocked_by_mirror_whitelist='root/pypi')
+    # and try "*"
+    ixconfig = stage2.ixconfig.copy()
+    ixconfig["mirror_whitelist"] = ["*"]
+    stage2.modify(**ixconfig)
+    assert stage2.get_mirror_whitelist_info("pytest") == dict(
+        has_mirror_base=True,
+        blocked_by_mirror_whitelist=None)
 
 
 class TestStage:
@@ -84,6 +128,11 @@ class TestStage:
         assert model.getstage("hello", "world2") is None
         assert model.getstage("hello", "world") is not None
 
+    def test_store_and_retrieve_simple(self, stage):
+        register_and_store(stage, "someproject-1.1.tar.gz")
+        assert len(stage.get_simplelinks_perstage("someproject")) == 1
+        assert len(stage.get_releaselinks_perstage("someproject")) == 1
+
     @pytest.mark.with_notifier
     def test_delete_user_hooks_issue228(self, model, caplog):
         keyfs = model.xom.keyfs
@@ -127,33 +176,34 @@ class TestStage:
         assert stage.ixconfig["bases"] == ("root/pypi",)
 
     def test_empty(self, stage, bases):
-        assert not stage.get_releaselinks("someproject")
-        assert not stage.list_projectnames_perstage()
+        assert not stage.has_project("someproject")
+        assert not stage.list_projects_perstage()
 
     def test_inheritance_simple(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject", "<a href='someproject-1.0.zip' /a>")
-        assert stage.list_projectnames_perstage() == set()
+        assert stage.list_projects_perstage() == set()
         links = stage.get_releaselinks("someproject")
         assert len(links) == 1
         stage.set_versiondata(udict(name="someproject", version="1.1"))
-        assert stage.list_projectnames_perstage() == set(["someproject"])
+        assert stage.list_projects_perstage() == set(["someproject"])
 
     def test_inheritance_twice(self, pypistage, stage, user):
         user.create_stage(index="dev2", bases=("root/pypi",))
         stage_dev2 = user.getstage("dev2")
-        stage.modify(bases=(stage_dev2.name,), pypi_whitelist=['someproject'])
+        stage.modify(bases=(stage_dev2.name,), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject",
                               "<a href='someproject-1.0.zip' /a>")
         register_and_store(stage_dev2, "someproject-1.1.tar.gz")
         register_and_store(stage_dev2, "someproject-1.2.tar.gz")
+        assert len(stage.get_simplelinks("someproject")) == 3
         links = stage.get_releaselinks("someproject")
         assert len(links) == 3
         assert links[0].basename == "someproject-1.2.tar.gz"
         assert links[1].basename == "someproject-1.1.tar.gz"
         assert links[2].basename == "someproject-1.0.zip"
-        assert stage.list_projectnames_perstage() == set()
-        assert stage_dev2.list_projectnames_perstage() == set(["someproject"])
+        assert stage.list_projects_perstage() == set()
+        assert stage_dev2.list_projects_perstage() == set(["someproject"])
 
     def test_inheritance_complex_issue_214(self, pypistage, model):
         prov_user = model.create_user('provider', password="123")
@@ -182,7 +232,7 @@ class TestStage:
         assert extagg_index2.list_versions('pkg') == set(['1.0', '2.0', '3.0'])
 
     def test_inheritance_normalize_multipackage(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['some-project'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['some-project'])
         pypistage.mock_simple("some-project", """
             <a href='some_project-1.0.zip' /a>
             <a href='some_project-1.0.tar.gz' /a>
@@ -196,7 +246,7 @@ class TestStage:
         assert links[0].basename == "some_project-1.2.tar.gz"
         assert links[1].basename == "some_project-1.0.zip"
         assert links[2].basename == "some_project-1.0.tar.gz"
-        assert stage.list_projectnames_perstage() == set(["some-project"])
+        assert stage.list_projects_perstage() == set(["some-project"])
 
     def test_inheritance_tolerance_on_different_names(self, stage, user):
         register_and_store(stage, "some_project-1.2.tar.gz",
@@ -213,7 +263,7 @@ class TestStage:
         assert len(links) == 1
 
     def test_get_releaselinks_inheritance_shadow(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.0.zip' /a>")
         register_and_store(stage, "someproject-1.0.zip", b"123")
@@ -224,7 +274,7 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
 
     def test_get_releaselinks_inheritance_shadow_egg(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['py'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['py'])
         pypistage.mock_simple("py",
         """<a href="http://bb.org/download/py.zip#egg=py-dev" />
            <a href="http://bb.org/download/master#egg=py-dev2" />
@@ -237,16 +287,14 @@ class TestStage:
         assert e1.basename == "py.zip"
         assert e2.basename == "master"
 
-    def test_inheritance_error(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+    def test_inheritance_error_are_nop(self, pypistage, stage):
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject", status_code = -1)
-        with pytest.raises(stage.UpstreamError):
-            stage.get_releaselinks("someproject")
-        with pytest.raises(stage.UpstreamError):
-            stage.list_versions("someproject")
+        assert stage.get_releaselinks("someproject") == []
+        assert stage.list_versions("someproject") == set([])
 
     def test_get_versiondata_inherited(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.0.zip' /a>")
         verdata = stage.get_versiondata("someproject", "1.0")
@@ -254,7 +302,7 @@ class TestStage:
         assert "someproject-1.0.zip" in str(verdata)
 
     def test_get_versiondata_inherit_not_exist_version(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.0.zip' /a>")
         assert not stage.get_versiondata("someproject", "2.0")
@@ -268,7 +316,7 @@ class TestStage:
         entries = stage.get_releaselinks("some")
         assert len(entries) == 1
         assert entries[0].hash_spec == entry.hash_spec
-        assert stage.list_projectnames_perstage() == set(["some"])
+        assert stage.list_projects_perstage() == set(["some"])
         verdata = stage.get_versiondata("some", "1.0")
         links = verdata["+elinks"]
         assert len(links) == 1
@@ -288,7 +336,7 @@ class TestStage:
         assert len(ls.get_links()) == 2
 
     def test_project_versiondata_shadowed(self, pypistage, stage):
-        stage.modify(bases=("root/pypi",), pypi_whitelist=['someproject'])
+        stage.modify(bases=("root/pypi",), mirror_whitelist=['someproject'])
         pypistage.mock_simple("someproject",
             "<a href='someproject-1.0.zip' /a>")
         content = b"123"
@@ -312,12 +360,32 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
         # if we add the project to the whitelist, we also get the release
         # from pypi
-        stage.modify(pypi_whitelist=['someproject'])
+        stage.modify(mirror_whitelist=['someproject'])
         links = stage.get_releaselinks("someproject")
         assert len(links) == 2
         assert links[0].entrypath.endswith("someproject-1.1.zip")
         assert links[1].entrypath.endswith("someproject-1.0.zip")
 
+    def test_project_whitelist_empty_project(self, pypistage, stage):
+        stage.modify(bases=("root/pypi",))
+        pypistage.mock_simple("someproject",
+            "<a href='someproject-1.1.zip' /a>")
+        stage.set_versiondata(udict(name="someproject", version="1.0"))
+        links = stage.get_releaselinks("someproject")
+        # because the whitelist doesn't include "someproject" we get
+        # no releases, because we only registered, but didn't upload
+        assert len(links) == 0
+
+    def test_project_whitelist_nothing_in_stage(self, pypistage, stage):
+        stage.modify(bases=("root/pypi",))
+        pypistage.mock_simple("someproject",
+            "<a href='someproject-1.1.zip' /a>")
+        links = stage.get_releaselinks("someproject")
+        # because the whitelist doesn't include "someproject" we get
+        # no releases, because we only registered, but didn't upload
+        assert len(links) == 1
+        assert links[0].entrypath.endswith("someproject-1.1.zip")
+
     def test_project_whitelist_inheritance(self, pypistage, stage, user):
         user.create_stage(index="dev2", bases=("root/pypi",))
         stage_dev2 = user.getstage("dev2")
@@ -334,7 +402,7 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
         # if we add the project to the whitelist of the inherited index, we
         # also get the release from pypi
-        stage_dev2.modify(pypi_whitelist=['someproject'])
+        stage_dev2.modify(mirror_whitelist=['someproject'])
         links = stage.get_releaselinks("someproject")
         assert len(links) == 2
         assert links[0].entrypath.endswith("someproject-1.1.zip")
@@ -354,7 +422,7 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
         # if we allow all projects in the whitelist, we also get the release
         # from pypi
-        stage.modify(pypi_whitelist=['*'])
+        stage.modify(mirror_whitelist=['*'])
         links = stage.get_releaselinks("someproject")
         assert len(links) == 2
         assert links[0].entrypath.endswith("someproject-1.1.zip")
@@ -376,7 +444,7 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
         # if we add all projects to the whitelist of the inherited index, we
         # also get the release from pypi
-        stage_dev2.modify(pypi_whitelist=['*'])
+        stage_dev2.modify(mirror_whitelist=['*'])
         links = stage.get_releaselinks("someproject")
         assert len(links) == 2
         assert links[0].entrypath.endswith("someproject-1.1.zip")
@@ -398,29 +466,60 @@ class TestStage:
         assert links[0].entrypath.endswith("someproject-1.0.zip")
         # if we add all projects to the whitelist of the inheriting index, we
         # also get the release from pypi
-        stage.modify(pypi_whitelist=['*'])
+        stage.modify(mirror_whitelist=['*'])
         links = stage.get_releaselinks("someproject")
         assert len(links) == 2
         assert links[0].entrypath.endswith("someproject-1.1.zip")
         assert links[1].entrypath.endswith("someproject-1.0.zip")
 
-    def test_store_and_delete_project(self, stage, bases):
-        content = b"123"
-        register_and_store(stage, "some-1.0.zip", content)
-        assert stage.get_versiondata_perstage("some", "1.0")
-        stage.del_project("some")
-        assert not stage.list_versions_perstage("some")
+    @pytest.mark.parametrize("setting, expected", [
+        ('someproject', ['someproject']),
+        ('he_llo', ['he-llo']),
+        ('he_llo,Django', ['he-llo', 'django']),
+        ('foo,bar', ['foo', 'bar']),
+        ('*', ['*'])])
+    def test_whitelist_setting(self, pypistage, stage, setting, expected):
+        from devpi_server.model import InvalidIndexconfig
+        stage.modify(mirror_whitelist=setting)
+        ixconfig = stage.get()
+        assert ixconfig['pypi_whitelist'] == []
+        assert ixconfig['mirror_whitelist'] == expected
+        stage.modify(pypi_whitelist=setting)
+        ixconfig = stage.get()
+        assert ixconfig['pypi_whitelist'] == []
+        assert ixconfig['mirror_whitelist'] == expected
+        stage.modify(pypi_whitelist=[], mirror_whitelist=setting)
+        ixconfig = stage.get()
+        assert ixconfig['pypi_whitelist'] == []
+        assert ixconfig['mirror_whitelist'] == expected
+        with pytest.raises(InvalidIndexconfig):
+            stage.modify(pypi_whitelist=setting, mirror_whitelist=[])
+        with pytest.raises(InvalidIndexconfig):
+            stage.modify(pypi_whitelist=setting, mirror_whitelist=setting)
 
-    def test_store_and_delete_release(self, stage, bases):
-        register_and_store(stage, "some-1.0.zip")
-        register_and_store(stage, "some-1.1.zip")
-        assert stage.get_versiondata_perstage("some", "1.0")
-        assert stage.get_versiondata_perstage("some", "1.1")
-        stage.del_versiondata("some", "1.0")
-        assert not stage.get_versiondata_perstage("some", "1.0")
-        assert stage.get_versiondata_perstage("some", "1.1")
-        stage.del_versiondata("some", "1.1")
-        assert stage.get_projectname("some") is None
+    def test_store_and_delete_project(self, stage):
+        register_and_store(stage, "some_xyz-1.0.zip", b"123")
+        assert stage.get_versiondata_perstage("Some_xyz", "1.0")
+        assert stage.get_versiondata_perstage("SOME_XYZ", "1.0")
+        assert stage.get_versiondata_perstage("some-xyz", "1.0")
+        stage.del_project("SoMe-XYZ")
+        assert not stage.list_versions_perstage("Some-xyz")
+        assert not stage.list_versions_perstage("Some_xyz")
+
+    def test_store_and_delete_release(self, stage):
+        register_and_store(stage, "Some_xyz-1.0.zip")
+        register_and_store(stage, "Some_xyz-1.1.zip")
+        # the name in versiondata is the "display" name, the originally
+        # registered name.
+        assert stage.get_versiondata_perstage("SOME_xYz", "1.0")["name"] == "Some_xyz"
+        assert stage.get_versiondata_perstage("some-xyz", "1.1")["name"] == "Some_xyz"
+        stage.del_versiondata("SOME-XYZ", "1.0")
+        assert not stage.get_versiondata_perstage("SOME_xyz", "1.0")
+        assert stage.get_versiondata_perstage("SOME-xyz", "1.1") == \
+               stage.get_versiondata_perstage("some-xyz", "1.1")
+        stage.del_versiondata("SomE_xyz", "1.1")
+        assert not stage.has_project_perstage("SOME-xyz")
+        assert not stage.has_project_perstage("some-xyz")
 
     def test_delete_not_existing(self, stage, bases):
         with pytest.raises(stage.NotFound) as excinfo:
@@ -463,6 +562,55 @@ class TestStage:
         assert tmpdir.join("_static").check(dir=1)
         assert tmpdir.join("_templ", "x.css").check(file=1)
 
+    def test_multiple_store_doczip_uses_project(self, stage, bases, tmpdir):
+        # check that two store_doczip calls with slightly
+        # different names will not lead to two doczip entries
+        stage.set_versiondata(udict(name="pkg1", version="1.0"))
+        stage.store_doczip("pkg1", "1.0", zip_dict({}))
+        content2 = zip_dict({"index.html": "<html/>"})
+        stage.store_doczip("Pkg1", "1.0", content2)
+
+        # check we have only have one doczip link
+        linkstore = stage.get_linkstore_perstage("pkg1", "1.0")
+        links = linkstore.get_links(rel="doczip")
+        assert len(links) == 1
+
+        # get doczip and check it's really the latest one
+        doczip2 = stage.get_doczip("pkg1", "1.0")
+        with Archive(BytesIO(doczip2)) as archive:
+            archive.extract(tmpdir)
+        assert tmpdir.join("index.html").read() == "<html/>"
+
+
+    def test_simulate_multiple_doczip_entries(self, stage, bases, tmpdir):
+        stage.set_versiondata(udict(name="pkg1", version="1.0"))
+        stage.store_doczip("pkg1", "1.0", zip_dict({}))
+
+        # simulate a second entry with a slightly different name
+        # (XXX not clear if this test is really neccessary. hpk thinks for
+        # exporting state from server<2.1.5 with such a double-entry one
+        # needs to install 2.1.5 and export from there anyway, clearing
+        # the problem. Then again server<2.3.2 allowed the store_doczip
+        # method to construct doczip filenames which differ only in
+        # casing)
+        linkstore = stage.get_linkstore_perstage("Pkg1", "1.0", readonly=False)
+        content = zip_dict({"index.html": "<html/>"})
+        linkstore.create_linked_entry(
+                rel="doczip",
+                basename="Pkg1-1.0.doc.zip",
+                file_content=content,
+        )
+
+        # check we have two doczip links
+        linkstore = stage.get_linkstore_perstage("pkg1", "1.0")
+        links = linkstore.get_links(rel="doczip")
+        assert len(links) == 2
+
+        # get doczip and check it's really the latest one
+        doczip = stage.get_doczip("pkg1", "1.0")
+        assert doczip == content
+
+
     def test_storedoczipfile(self, stage, bases):
         from devpi_common.archive import Archive
         stage.set_versiondata(udict(name="pkg1", version="1.0"))
@@ -485,7 +633,7 @@ class TestStage:
         content = b'123'
         link = register_and_store(stage, "pkg1-1.0.tar.gz", content=content)
         entry = link.entry
-        assert entry.projectname == "pkg1"
+        assert entry.project == "pkg1"
         assert entry.version == "1.0"
         toxresultdata = {'hello': 'world'}
         tlink = stage.store_toxresult(link, toxresultdata)
@@ -501,7 +649,7 @@ class TestStage:
         back_data = json.loads(tentry.file_get_content().decode("utf8"))
         assert back_data == toxresultdata
 
-        assert tentry.projectname == entry.projectname
+        assert tentry.project == entry.project
         assert tentry.version == entry.version
 
         results = stage.get_toxresults(link)
@@ -561,7 +709,7 @@ class TestStage:
     def test_get_versiondata_latest_inheritance(self, user, model, stage):
         stage_base_name = stage.index + "base"
         user.create_stage(index=stage_base_name, bases=(stage.name,))
-        stage_sub = model.getstage(stage.user.name, stage_base_name)
+        stage_sub = model.getstage(stage.username, stage_base_name)
         stage_sub.set_versiondata(udict(name="hello", version="1.0"))
         stage.set_versiondata(udict(name="hello", version="1.1"))
         assert stage_sub.get_latest_version_perstage("hello") == "1.0"
@@ -576,17 +724,16 @@ class TestStage:
         stage.set_versiondata(udict(name="hello-World", version="1.0"))
         for name in ("Hello-World", "hello_world"):
             caplog.handler.records = []
-            caplog.setLevel(logging.WARNING)
+            caplog.set_level(logging.WARNING)
             stage.set_versiondata(udict(name=name, version="1.0"))
             rec = caplog.getrecords()
-            assert len(rec) == 1, [str(x) for x in rec]
-            assert "using" in rec[0].msg and name in rec[0].msg
+            assert not rec
 
     @pytest.mark.start_threads
     def test_set_versiondata_hook(self, stage, queue):
         class Plugin:
             def devpiserver_on_changed_versiondata(self,
-                    stage, projectname, version, metadata):
+                    stage, project, version, metadata):
                 queue.put((stage, metadata))
         stage.xom.config.pluginmanager.register(Plugin())
         orig_metadata = udict(name="hello", version="1.0")
@@ -601,11 +748,33 @@ class TestStage:
         assert stage2.name == stage.name
         assert not metadata
 
+    @pytest.mark.start_threads
+    @pytest.mark.notransaction
+    def test_stage_created_hook(self, xom, queue):
+        class Plugin:
+            def devpiserver_stage_created(self, stage):
+                queue.put(stage)
+        xom.config.pluginmanager.register(Plugin())
+        with xom.keyfs.transaction(write=True):
+            model = xom.model
+            user = model.create_user("user", "password", email="some@email.com")
+            user.create_stage("hello")
+        while 1:
+            stage = queue.get(timeout=10)
+            if stage.name != "root/pypi":
+                break
+        assert stage.name == "user/hello"
+
+        with xom.keyfs.transaction(write=True):
+            user.create_stage("hello2")
+        assert queue.get(timeout=10).name == "user/hello2"
+
+
     @pytest.mark.start_threads
     def test_doczip_uploaded_hook(self, stage, queue):
         class Plugin:
-            def devpiserver_on_upload(self, stage, projectname, version, link):
-                queue.put((stage, projectname, version, link))
+            def devpiserver_on_upload(self, stage, project, version, link):
+                queue.put((stage, project, version, link))
         stage.xom.config.pluginmanager.register(Plugin())
         stage.set_versiondata(udict(name="pkg1", version="1.0"))
         content = zip_dict({"index.html": "<html/>",
@@ -619,7 +788,7 @@ class TestStage:
             assert link.entry.file_get_content() == content
         # delete, which shouldnt trigger devpiserver_on_upload
         with stage.xom.keyfs.transaction(write=True):
-            linkstore = stage.get_linkstore_perstage("pkg1", "1.0")
+            linkstore = stage.get_linkstore_perstage("pkg1", "1.0", readonly=False)
             linkstore.remove_links()
 
         # now write again and check that we get something from the queue
@@ -632,17 +801,19 @@ class TestStage:
 
 
     def test_get_existing_project(self, stage):
+        assert not stage.get_versiondata("hello", "1.0")
+        assert not stage.get_versiondata("This", "1.0")
         stage.set_versiondata(udict(name="Hello", version="1.0"))
         stage.set_versiondata(udict(name="this", version="1.0"))
-        name = stage.get_projectname("hello")
-        assert name == "Hello"
+        assert stage.get_versiondata("hello", "1.0")
+        assert stage.get_versiondata("This", "1.0")
 
 
 class TestLinkStore:
     @pytest.fixture
     def linkstore(self, stage):
         stage.set_versiondata(udict(name="proj1", version="1.0"))
-        return stage.get_linkstore_perstage("proj1", "1.0")
+        return stage.get_linkstore_perstage("proj1", "1.0", readonly=False)
 
     def test_store_file(self, linkstore):
         linkstore.create_linked_entry(
@@ -749,3 +920,40 @@ def test_setdefault_indexes(xom, model):
         userconfig = model.get_user("root").get()
         for key in userconfig["indexes"]["pypi"]:
             assert py.builtin._istext(key)
+
+
+@pytest.mark.parametrize("key", ("acl_upload", "bases", "mirror_whitelist", "pypi_whitelist"))
+@pytest.mark.parametrize("value, result", (
+    ("", []), ("x,y", ["x", "y"]), ("x,,y", ["x", "y"])))
+def test_get_indexconfig_lists(key, value, result):
+    class hooks:
+        def devpiserver_indexconfig_defaults(self, index_type):
+            return {}
+    kvdict = get_indexconfig(hooks(), type="stage", **{key: value})
+    if key == "pypi_whitelist":
+        # check behaviour of older devpi-client
+        key = "mirror_whitelist"
+    assert kvdict[key] == result
+
+
+@pytest.mark.parametrize(["input", "expected"], [
+    ({},
+     dict(type="stage")),
+    ({"volatile": "foo"},
+     dict(type="stage", volatile=True)),
+    ({"volatile": "False"},
+     dict(type="stage", volatile=False)),
+    ({"volatile": "False", "bases": "root/pypi"},
+     dict(type="stage", volatile=False, bases=["root/pypi"])),
+    ({"volatile": "False", "bases": ["root/pypi"]},
+     dict(type="stage", volatile=False, bases=["root/pypi"])),
+    ({"volatile": "False", "bases": ["root/pypi"], "acl_upload": ["hello"]},
+     dict(type="stage", volatile=False, bases=["root/pypi"],
+          acl_upload=["hello"])),
+])
+def test_get_indexconfig_values(xom, input, expected):
+    class hooks:
+        def devpiserver_indexconfig_defaults(self, index_type):
+            return {}
+    result = get_indexconfig(hooks(), type="stage", **input)
+    assert result == expected
