@@ -1,22 +1,16 @@
+# -*- coding: utf-8 -*-
 import hashlib
 import pytest
 import py
+from devpi_server.log import thread_pop_log
+from devpi_server.fileutil import load
 from devpi_server.replica import *  # noqa
-from devpi_common.url import URL
 
 def loads(bytestring):
     return load(py.io.BytesIO(bytestring))
 
 pytestmark = [pytest.mark.notransaction]
 
-def test_view_name2serials(pypistage, testapp):
-    pypistage.mock_simple("package", '<a href="/package-1.0.zip" />',
-                          pypiserial=15)
-    r = testapp.get("/root/pypi/+name2serials", expect_errors=False)
-    io = py.io.BytesIO(r.body)
-    entries = load(io)
-    assert entries["package"] == 15
-
 @pytest.fixture
 def testapp(testapp):
     master_uuid = testapp.xom.config.get_master_uuid()
@@ -56,26 +50,26 @@ class TestChangelog:
         data = loads(body)
         assert "this" in str(data)
 
-    def test_get_wait(self, reqchangelog, testapp, mapp, noiter, monkeypatch):
-        mapp.create_user("this", password="p")
-        latest_serial = self.get_latest_serial(testapp)
-        monkeypatch.setattr(testapp.xom.keyfs.notifier.cv_new_transaction,
-                            "wait", lambda *x: 0/0)
-        with pytest.raises(ZeroDivisionError):
-            reqchangelog(latest_serial+1)
-
-    def test_get_wait_blocking_ends(self, testapp, mapp, noiter, monkeypatch,
+    def test_wait_entry_fails(self, testapp, mapp, noiter, monkeypatch,
                                     reqchangelog):
         mapp.create_user("this", password="p")
         latest_serial = self.get_latest_serial(testapp)
-        monkeypatch.setattr(MasterChangelogRequest, "MAX_REPLICA_BLOCK_TIME",
-                            0.01)
-        monkeypatch.setattr(MasterChangelogRequest, "WAKEUP_INTERVAL",
-                            0.001)
+        monkeypatch.setattr(MasterChangelogRequest, "MAX_REPLICA_BLOCK_TIME", 0.01)
         r = reqchangelog(latest_serial+1)
         assert r.status_code == 202
         assert int(r.headers["X-DEVPI-SERIAL"]) == latest_serial
 
+    def test_wait_entry_succeeds(self, blank_request, xom, mapp):
+        mapp.create_user("this", password="p")
+        req = blank_request()
+        req.registry = {"xom": xom}
+        mcr = MasterChangelogRequest(req)
+        with xom.keyfs.transaction():
+            with pytest.raises(HTTPNotFound):
+                mcr._wait_for_entry(xom.keyfs.get_current_serial() + 10)
+            entry = mcr._wait_for_entry(xom.keyfs.get_current_serial())
+        assert entry
+
     def test_master_id_mismatch(self, testapp):
         testapp.xget(400, "/+changelog/0", headers={H_EXPECTED_MASTER_ID:str("123")})
         r = testapp.xget(200, "/+changelog/0", headers={H_EXPECTED_MASTER_ID: ''})
@@ -84,55 +78,9 @@ class TestChangelog:
         testapp.xget(400, "/+changelog/0")
 
 
-class TestPyPIProxy:
-    def test_pypi_proxy(self, xom, reqmock):
-        from devpi_server.keyfs import dump
-        url = "http://localhost:3141/root/pypi/+name2serials"
-        master_url = URL("http://localhost:3141")
-        proxy = PyPIProxy(xom._httpsession, master_url)
-        io = py.io.BytesIO()
-        dump({"hello": 42}, io)
-        data = io.getvalue()
-        reqmock.mockresponse(url=url, code=200, method="GET", data=data)
-        name2serials = proxy.list_packages_with_serial()
-        assert name2serials == {"hello": 42}
-
-    def test_replica_startup(self, replica_xom):
-        assert isinstance(replica_xom.proxy, PyPIProxy)
-
-
-@pytest.mark.parametrize("normname,realname",
-                         [("proj", "proj"), ("proj-x", "proj_x")])
-def test_pypi_project_changed(replica_xom, normname, realname):
-    handler = PypiProjectChanged(replica_xom)
-    class Ev:
-        value = dict(projectname=realname, serial=12)
-        typedkey = replica_xom.keyfs.get_key("PYPILINKS")(name=normname)
-    handler(Ev())
-    assert replica_xom.pypimirror.name2serials[realname] == 12
-    if normname != realname:
-        assert replica_xom.pypimirror.normname2name[normname] == realname
-
-    class Ev2:
-        value = dict(projectname=realname, serial=15)
-        typedkey = replica_xom.keyfs.get_key("PYPILINKS")(name=normname)
-    handler(Ev2())
-    assert replica_xom.pypimirror.name2serials[realname] == 15
-
-    class Ev3:
-        value = dict(projectname=realname, serial=12)
-        typedkey = replica_xom.keyfs.get_key("PYPILINKS")(name=normname)
-    handler(Ev3())
-    assert replica_xom.pypimirror.name2serials[realname] == 15
-
-    class Ev4:
-        typedkey = replica_xom.keyfs.get_key("PYPILINKS")(name=normname)
-        value = None
-    handler(Ev4())
-    assert realname not in replica_xom.pypimirror.name2serials
-    if normname != realname:
-        assert normname not in replica_xom.pypimirror.normname2name
-
+def get_raw_changelog_entry(xom, serial):
+    with xom.keyfs._storage.get_connection() as conn:
+        return conn.get_raw_changelog_entry(serial)
 
 class TestReplicaThread:
     @pytest.fixture
@@ -145,7 +93,15 @@ class TestReplicaThread:
     @pytest.fixture
     def mockchangelog(self, reqmock):
         def mockchangelog(num, code, data=b'',
-                          headers={H_MASTER_UUID: "123"}):
+                          uuid="123", headers=None):
+            if headers is None:
+                headers = {}
+            headers = dict((k.lower(), v) for k, v in headers.items())
+            if uuid is not None:
+                headers.setdefault(H_MASTER_UUID.lower(), "123")
+            headers.setdefault("x-devpi-serial", str(2))
+            if headers["x-devpi-serial"] is None:
+                del headers["x-devpi-serial"]
             reqmock.mockresponse("http://localhost/+changelog/%s" % num,
                                  code=code, data=data, headers=headers)
         return mockchangelog
@@ -164,10 +120,36 @@ class TestReplicaThread:
             rt.thread_run()
         assert caplog.getrecords("could not process")
 
+    def test_thread_run_recovers_from_error(self, mock, rt, reqmock, mockchangelog, caplog, xom):
+        import socket
+        # setup a regular request
+        data = get_raw_changelog_entry(xom, 0)
+        mockchangelog(0, code=200, data=data)
+        # get the result
+        orig_req = reqmock.url2reply[("http://localhost/+changelog/0", None)]
+        # setup so the first attempt fails, then the second succeeds
+        reqmock.url2reply = mock.Mock()
+        reqmock.url2reply.get.side_effect = [socket.error(), orig_req]
+        rt.thread.sleep = mock.Mock()
+        rt.thread.sleep.side_effect = [
+            # 1. sleep
+            0,
+            # 2. raise exception to get into exception part of while loop
+            ZeroDivisionError(),
+            # 3. kill the thread
+            ZeroDivisionError()]
+        # run
+        with pytest.raises(ZeroDivisionError):
+            rt.thread_run()
+        msgs = [x.msg for x in caplog.getrecords(".*http://localhost/\+changelog/0")]
+        assert msgs == [
+            '[REP] fetching %s',
+            '[REP] error fetching %s: %s',
+            '[REP] fetching %s']
+
     def test_thread_run_ok(self, rt, mockchangelog, caplog, xom):
         rt.thread.sleep = lambda *x: 0/0
-        data = xom.keyfs._fs.get_raw_changelog_entry(0)
-        assert data
+        data = get_raw_changelog_entry(xom, 0)
         mockchangelog(0, code=200, data=data)
         mockchangelog(1, code=404, data=data)
         with pytest.raises(ZeroDivisionError):
@@ -176,7 +158,7 @@ class TestReplicaThread:
 
     def test_thread_run_no_uuid(self, rt, mockchangelog, caplog, xom):
         rt.thread.sleep = lambda x: 0/0
-        mockchangelog(0, code=200, data=b'123', headers={})
+        mockchangelog(0, code=200, data=b'123', uuid=None)
         with pytest.raises(ZeroDivisionError):
             rt.thread_run()
         assert caplog.getrecords("remote.*no.*UUID")
@@ -184,8 +166,7 @@ class TestReplicaThread:
     def test_thread_run_ok_uuid_change(self, rt, mockchangelog, caplog, xom, monkeypatch):
         monkeypatch.setattr("os._exit", lambda n: 0/0)
         rt.thread.sleep = lambda *x: 0/0
-        data = xom.keyfs._fs.get_raw_changelog_entry(0)
-        assert data
+        data = get_raw_changelog_entry(xom, 0)
         mockchangelog(0, code=200, data=data)
         mockchangelog(1, code=200, data=data,
                       headers={"x-devpi-master-uuid": "001"})
@@ -193,6 +174,49 @@ class TestReplicaThread:
             rt.thread_run()
         assert caplog.getrecords("master UUID.*001.*does not match")
 
+    def test_thread_run_serial_mismatch(self, rt, mockchangelog, caplog, xom, monkeypatch):
+        monkeypatch.setattr("os._exit", lambda n: 0/0)
+        rt.thread.sleep = lambda *x: 0/0
+
+        # we need to have at least two commits
+        with xom.keyfs.transaction(write=True):
+            xom.model.create_user("qlwkej", "qwe")
+
+        data = get_raw_changelog_entry(xom, 0)
+        mockchangelog(0, code=200, data=data)
+        data = get_raw_changelog_entry(xom, 1)
+        mockchangelog(1, code=200, data=data,
+                      headers={"x-devpi-serial": "0"})
+        with pytest.raises(ZeroDivisionError):
+            rt.thread_run()
+        assert caplog.getrecords("Got serial 0 from master which is smaller than last "
+                    "recorded serial")
+
+    def test_thread_run_invalid_serial(self, rt, mockchangelog, caplog, xom, monkeypatch):
+        monkeypatch.setattr("os._exit", lambda n: 0/0)
+        rt.thread.sleep = lambda *x: 0/0
+        data = get_raw_changelog_entry(xom, 0)
+        assert data
+        mockchangelog(0, code=200, data=data)
+        data = get_raw_changelog_entry(xom, 1)
+        mockchangelog(1, code=200, data=data,
+                      headers={"x-devpi-serial": "foo"})
+        with pytest.raises(ZeroDivisionError):
+            rt.thread_run()
+        assert caplog.getrecords("error fetching.*invalid literal for int")
+
+    def test_thread_run_missing_serial(self, rt, mockchangelog, caplog, xom, monkeypatch):
+        monkeypatch.setattr("os._exit", lambda n: 0/0)
+        rt.thread.sleep = lambda *x: 0/0
+        data = get_raw_changelog_entry(xom, 0)
+        mockchangelog(0, code=200, data=data)
+        data = get_raw_changelog_entry(xom, 1)
+        mockchangelog(1, code=200, data=data,
+                      headers={"x-devpi-serial": None})
+        with pytest.raises(ZeroDivisionError):
+            rt.thread_run()
+        assert caplog.getrecords("error fetching.*x-devpi-serial")
+
     def test_thread_run_try_again(self, rt, mockchangelog, caplog):
         l = [1]
         def exit_if_shutdown():
@@ -244,7 +268,7 @@ class TestTweenReplica:
         reqmock.mock("http://localhost/blankpath",
                      code=200, headers={"X-DEVPI-SERIAL": "10"})
         l = []
-        monkeypatch.setattr(xom.keyfs.notifier, "wait_tx_serial",
+        monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: l.append(x))
         handler = tween_replica_proxy(None, {"xom": xom})
         response = handler(blank_request(method="PUT"))
@@ -256,7 +280,7 @@ class TestTweenReplica:
         reqmock.mock("http://localhost/blankpath",
                      code=200, reason="GOOD", headers={"X-DEVPI-SERIAL": "10"})
         l = []
-        monkeypatch.setattr(xom.keyfs.notifier, "wait_tx_serial",
+        monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: l.append(x))
         handler = tween_replica_proxy(None, {"xom": xom})
         response = handler(blank_request(method="PUT"))
@@ -270,7 +294,7 @@ class TestTweenReplica:
                                         "location": "http://localhost/hello"}
         )
         l = []
-        monkeypatch.setattr(xom.keyfs.notifier, "wait_tx_serial",
+        monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: l.append(x))
         handler = tween_replica_proxy(None, {"xom": xom})
         # normally the app is wrapped by OutsideURLMiddleware, since this is
@@ -289,7 +313,7 @@ class TestTweenReplica:
                         "Foo": "abc",
                         "Keep-Alive": "timeout=30",
                         "X-DEVPI-SERIAL": "0"})
-        monkeypatch.setattr(xom.keyfs.notifier, "wait_tx_serial",
+        monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: x)
         handler = tween_replica_proxy(None, {"xom": xom})
         response = handler(blank_request(method="PUT"))
@@ -297,15 +321,29 @@ class TestTweenReplica:
         assert 'foo' not in response.headers
         assert 'keep-alive' not in response.headers
 
-def replay(xom, replica_xom):
+def replay(xom, replica_xom, events=True):
+    threadlog.info("test: replaying replica")
     for serial in range(replica_xom.keyfs.get_next_serial(),
                         xom.keyfs.get_next_serial()):
         if serial == -1:
             continue
-        change_entry = xom.keyfs._fs.get_changes(serial)
-        threadlog.info("importing to replica %s", serial)
+        with xom.keyfs._storage.get_connection() as conn:
+            change_entry = conn.get_changes(serial)
+        threadlog.info("test: importing to replica %s", serial)
         replica_xom.keyfs.import_changes(serial, change_entry)
 
+    # replay notifications
+    if events:
+        noti_thread = replica_xom.keyfs.notifier
+        event_serial = noti_thread.read_event_serial()
+        thread_push_log("NOTI")
+        while event_serial < replica_xom.keyfs.get_current_serial():
+            event_serial += 1
+            noti_thread._execute_hooks(event_serial, threadlog, raising=True)
+            noti_thread.write_event_serial(event_serial)
+        thread_pop_log("NOTI")
+
+
 class TestFileReplication:
     @pytest.fixture
     def replica_xom(self, makexom):
@@ -343,14 +381,13 @@ class TestFileReplication:
             assert not replica_xom.model.get_user("world")
             assert replica_xom.model.get_user("hello")
 
-
     def test_fetch(self, gen, reqmock, xom, replica_xom):
         replay(xom, replica_xom)
         content1 = b'hello'
         md5 = hashlib.md5(content1).hexdigest()
         link = gen.pypi_package_link("pytest-1.8.zip#md5=%s" % md5, md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link)
+            entry = xom.filestore.maplink(link, "root", "pypi")
             assert not entry.file_exists()
 
         replay(xom, replica_xom)
@@ -374,7 +411,7 @@ class TestFileReplication:
         assert persisted_errors == replica_xom.errors.errors
         with replica_xom.keyfs.transaction():
             assert not r_entry.file_exists()
-            assert not os.path.exists(r_entry._filepath)
+            assert not replica_xom.filestore.storedir.join(r_entry.relpath).exists()
 
         # then we try to return the correct thing
         with xom.keyfs.transaction(write=True):
@@ -402,7 +439,7 @@ class TestFileReplication:
         md5 = hashlib.md5(content1).hexdigest()
         link = gen.pypi_package_link("pytest-1.8.zip#md5=%s" % md5, md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link)
+            entry = xom.filestore.maplink(link, "root", "pypi")
             assert not entry.file_exists()
 
         master_url = replica_xom.config.master_url
@@ -416,7 +453,7 @@ class TestFileReplication:
         with xom.keyfs.transaction(write=True):
             entry.file_delete()
             entry.delete()
-        assert not os.path.exists(entry._filepath)
+        assert not xom.filestore.storedir.join(entry.relpath).exists()
 
         # and simulate what the master will respond
         xom.httpget.mockresponse(master_file_path, status_code=410)
@@ -433,7 +470,7 @@ class TestFileReplication:
         content1 = b'hello'
         link = gen.pypi_package_link("some-1.8.zip", md5=False)
         with xom.keyfs.transaction(write=True):
-            entry = xom.filestore.maplink(link)
+            entry = xom.filestore.maplink(link, "root", "pypi")
             assert not entry.file_exists()
             assert not entry.hash_spec
 
@@ -452,8 +489,9 @@ class TestFileReplication:
         # simulate some 500 master server error
         replica_xom.httpget.mockresponse(master_file_path, status_code=500,
                                          content=b'')
-        with pytest.raises(FileReplicationError):
+        with pytest.raises(FileReplicationError) as e:
             replay(xom, replica_xom)
+        assert str(e.value) == 'FileReplicationError with http://localhost/root/pypi/+e/https_pypi.python.org_package_some/some-1.8.zip, code=500, relpath=root/pypi/+e/https_pypi.python.org_package_some/some-1.8.zip, message=failed'
 
         # now get the real thing
         replica_xom.httpget.mockresponse(master_file_path, status_code=200,
@@ -464,15 +502,14 @@ class TestFileReplication:
             assert r_entry.file_get_content() == content1
 
 
-    @pytest.mark.parametrize("code", [200, 500])
     def test_cache_remote_file_fails(self, xom, replica_xom, gen,
-                                     monkeypatch, reqmock, code):
+                                     monkeypatch, reqmock):
         l = []
-        monkeypatch.setattr(xom.keyfs.notifier, "wait_tx_serial",
+        monkeypatch.setattr(xom.keyfs, "wait_tx_serial",
                             lambda x: l.append(x))
         with xom.keyfs.transaction(write=True):
             link = gen.pypi_package_link("pytest-1.8.zip", md5=True)
-            entry = xom.filestore.maplink(link)
+            entry = xom.filestore.maplink(link, "root", "pypi")
             assert entry.hash_spec and not entry.file_exists()
         replay(xom, replica_xom)
         with replica_xom.keyfs.transaction():
@@ -482,17 +519,48 @@ class TestFileReplication:
                      "X-DEVPI-SERIAL": "10"}
             entry = replica_xom.filestore.get_file_entry(entry.relpath)
             url = replica_xom.config.master_url.joinpath(entry.relpath).url
-            reqmock.mockresponse(url, code=code,
+            reqmock.mockresponse(url, code=500,
                                  headers=headers, data=b'123')
-            if code == 200:
-                l = []
-                monkeypatch.setattr(replica_xom.keyfs.notifier, "wait_tx_serial",
-                    lambda serial: l.append(serial))
             with pytest.raises(entry.BadGateway):
-                entry.cache_remote_file_replica()
-            if code == 200:
-                assert len(l) == 1
-                assert l[0] == 10
+                for part in entry.iter_remote_file_replica():
+                    pass
+
+
+    def test_checksum_mismatch(self, xom, replica_xom, gen, maketestapp,
+                               makemapp, reqmock):
+        # this test might seem to be doing the same as test_fetch above, but
+        # test_fetch creates a new transaction for the same file, which doesn't
+        # happen 'in real life'™
+        from devpi_server.replica import ReplicationErrors
+        app = maketestapp(xom)
+        mapp = makemapp(app)
+        api = mapp.create_and_use()
+        content1 = mapp.makepkg("hello-1.0.zip", b"content1", "hello", "1.0")
+        mapp.upload_file_pypi("hello-1.0.zip", content1, "hello", "1.0")
+        r_app = maketestapp(replica_xom)
+        # first we try to return something wrong
+        master_url = replica_xom.config.master_url
+        (path,) = mapp.get_release_paths('hello')
+        master_file_url = master_url.joinpath(path).url
+        replica_xom.httpget.mockresponse(master_file_url, code=200, content=b'13')
+        replay(xom, replica_xom)
+        assert xom.keyfs.get_current_serial() == replica_xom.keyfs.get_current_serial()
+        replication_errors = ReplicationErrors(replica_xom.config.serverdir)
+        assert list(replication_errors.errors.keys()) == [
+            '%s/+f/d0b/425e00e15a0d3/hello-1.0.zip' % api.stagename]
+        # the master and replica are in sync, so getting the file on the
+        # replica needs to fetch it again
+        headers = {"content-length": "8",
+                   "last-modified": "Thu, 25 Nov 2010 20:00:27 GMT",
+                   "content-type": "application/zip",
+                   "X-DEVPI-SERIAL": str(xom.keyfs.get_current_serial())}
+        reqmock.mockresponse(master_file_url, code=200, headers=headers)
+        replica_xom.httpget.mockresponse(master_file_url, code=200, content=content1, headers=headers)
+        r = r_app.get(path)
+        assert r.status_code == 200
+        assert r.body == content1
+        replication_errors = ReplicationErrors(replica_xom.config.serverdir)
+        assert list(replication_errors.errors.keys()) == []
 
 
 def test_should_fetch_remote_file():
@@ -507,3 +575,79 @@ def test_should_fetch_remote_file():
            should_fetch_remote_file(Entry(), {H_REPLICA_FILEREPL: str("YES")})
 
 
+def test_simplelinks_update_updates_projectname(httpget, monkeypatch,
+    pypistage, replica_pypistage, pypiurls, replica_xom, xom):
+
+    pypistage.mock_simple_projects([])
+    pypistage.mock_simple("pytest", pkgver="pytest-1.0.zip")
+    with xom.keyfs.transaction():
+        assert not pypistage.list_projects_perstage()
+
+    with xom.keyfs.transaction():
+        pypistage.get_simplelinks("pytest")
+
+    # replicate including executing events
+    replay(xom, replica_xom)
+
+    with replica_xom.keyfs.transaction():
+        st = replica_xom.model.getstage(pypistage.name)
+        assert st.list_projects_perstage() == set(["pytest"])
+
+
+def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypistage,
+                                  pypiurls, replica_xom, xom):
+    orig_simple = pypiurls.simple
+
+    # prepare the data on master
+    pypistage.mock_simple("pytest", pkgver="pytest-1.0.zip")
+    with xom.keyfs.transaction(write=True):
+        pypistage.get_releaselinks("pytest")
+
+    # replicate the state
+    replay(xom, replica_xom)
+
+    # now check
+    pypiurls.simple = 'http://localhost:3111/root/pypi/+simple/'
+    serial = xom.keyfs.get_current_serial()
+    httpget.mock_simple(
+        'pytest',
+        text='<a href="https://pypi.python.org/pytest/pytest-1.0.zip">pytest-1.0.zip</a>',
+        headers={'X-DEVPI-SERIAL': str(serial)})
+    with replica_xom.keyfs.transaction():
+        ret = replica_pypistage.get_releaselinks("pytest")
+    assert len(ret) == 1
+    assert ret[0].relpath == 'root/pypi/+e/https_pypi.python.org_pytest/pytest-1.0.zip'
+
+    # now we change the links and expire the cache
+    pypiurls.simple = orig_simple
+    pypistage.mock_simple("pytest", pkgver="pytest-1.1.zip", pypiserial=10001)
+    pypistage.cache_link_updates.expire('pytest')
+    with xom.keyfs.transaction(write=True):
+        pypistage.get_releaselinks("pytest")
+    assert xom.keyfs.get_current_serial() > serial
+
+    # we patch wait_tx_serial so we can check and replay
+    called = []
+    def wait_tx_serial(serial):
+        called.append(True)
+        assert xom.keyfs.get_current_serial() == serial
+        assert replica_xom.keyfs.get_current_serial() < serial
+        replay(xom, replica_xom, events=False)
+        assert replica_xom.keyfs.get_current_serial() == serial
+
+    monkeypatch.setattr(replica_xom.keyfs, 'wait_tx_serial', wait_tx_serial)
+    pypiurls.simple = 'http://localhost:3111/root/pypi/+simple/'
+    httpget.mock_simple(
+        'pytest',
+        text='<a href="https://pypi.python.org/pkg/pytest-1.1.zip">pytest-1.1.zip</a>',
+        pypiserial=10001,
+        headers={'X-DEVPI-SERIAL': str(xom.keyfs.get_current_serial())})
+    with replica_xom.keyfs.transaction():
+        # make the replica believe it hasn't updated for a longer time
+        r_pypistage = replica_xom.model.getstage("root/pypi")
+        r_pypistage.cache_link_updates.expire("pytest")
+        ret = replica_pypistage.get_releaselinks("pytest")
+    assert called == [True]
+    replay(xom, replica_xom)
+    assert len(ret) == 1
+    assert ret[0].relpath == 'root/pypi/+e/https_pypi.python.org_pytest/pytest-1.1.zip'
