@@ -16,9 +16,12 @@ from devpi_common.viewhelp import ViewLinkStore
 
 import devpi_server.views
 from devpi_server.views import tween_keyfs_transaction, make_uuid_headers
+from devpi_server.extpypi import parse_index
 
 from .functional import TestUserThings, TestIndexThings  # noqa
+from .functional import TestMirrorIndexThings  # noqa
 
+import devpi_server.filestore
 from devpi_server.filestore import get_default_hash_spec, make_splitdir
 
 proj = pytest.mark.parametrize("proj", [True, False])
@@ -33,30 +36,61 @@ def hash_spec_matches(hash_spec, content):
     return digest == hash_value
 
 
-@pytest.mark.parametrize("user,status", [
+@pytest.mark.parametrize("kind", ["user", "index"])
+@pytest.mark.parametrize("name,status", [
     ("foo_bar", 'ok'),
     ("foo-bar", 'ok'),
     ("foo.bar", 'ok'),
     ("foo.bar42", 'ok'),
-    (":foobar", 'warn'),
+    ("foo@bar42", 'ok'),
+    ("foo:bar42", 'fatal'),
+    ("foo!bar42", 'fatal'),
+    ("foo~bar42", 'fatal'),
+    (":foobar", 'fatal'),
     (":foobar:", 'fatal')])
-def test_invalid_username(caplog, testapp, user, status):
+def test_invalid_name(caplog, testapp, name, status, kind):
     reqdict = dict(password="123")
-    r = testapp.put_json("/%s" % user, reqdict, expect_errors=True)
+    if kind == "user":
+        r = testapp.put_json("/%s" % name, reqdict, expect_errors=True)
+    else:
+        r = testapp.put_json("/foo", reqdict)
+        testapp.set_auth("foo", "123")
+        r = testapp.put_json("/foo/%s" % name, {}, expect_errors=True)
     if status in ('ok', 'warn'):
-        code = 201
+        if kind == "user":
+            code = 201
+        else:
+            code = 200
     else:
         code = 400
     assert r.status_code == code
-    if status == 'warn':
-        msg = "username '%s' will be invalid with next release, use characters, numbers, underscore, dash and dots only" % user
-        logmsg = caplog.getrecords('invalid')[-1]
-        assert logmsg.message.endswith(msg)
     if status == 'fatal':
-        msg = "username '%s' is invalid, use characters, numbers, underscore, dash and dots only" % user
+        msg = (
+            "%sname '%s' contains characters that aren't allowed. "
+            "Any ascii symbol besides -.@_ is blocked." % (kind, name))
         assert r.json['message'] == msg
 
 
+def test_user_patch_keeps_missing_keys(testapp):
+    # needed for devpi-client < 2.5.0
+    testapp.put_json("/foo", dict(password="123"))
+    testapp.set_auth('foo', '123')
+    r = testapp.get("/foo")
+    assert r.json['result'] == {'username': 'foo', 'indexes': {}}
+    testapp.patch_json("/foo", dict(title="foo"))
+    r = testapp.get("/foo")
+    assert r.json['result'] == {
+        'username': 'foo', 'title': 'foo', 'indexes': {}}
+    testapp.patch_json("/foo", dict(description="bar"))
+    r = testapp.get("/foo")
+    assert r.json['result'] == {
+        'username': 'foo', 'title': 'foo', 'description': 'bar', 'indexes': {}}
+    testapp.patch_json("/foo", dict(description=""))
+    r = testapp.get("/foo")
+    assert r.json['result'] == {
+        'username': 'foo', 'title': 'foo', 'indexes': {}}
+
+
 @pytest.mark.parametrize("nodeinfo,expected", [
     ({}, (None, None)),
     ({"uuid": "123", "role":"master"}, ("123", "123")),
@@ -70,11 +104,12 @@ def test_make_uuid_headers(nodeinfo, expected):
 def test_simple_project(pypistage, testapp):
     name = "qpwoei"
     r = testapp.get("/root/pypi/+simple/" + name)
-    assert r.status_code == 200
+    assert r.status_code == 404
     assert r.headers["X-DEVPI-SERIAL"]
     # easy_install fails if the result isn't html
     assert "html" in r.headers['content-type']
-    assert not BeautifulSoup(r.text).findAll("a")
+    assert not parse_index("http://localhost", r.text, scrape=False).releaselinks
+
     path = "/%s-1.0.zip" % name
     pypistage.mock_simple(name, text='<a href="%s"/>' % path)
     r = testapp.get("/root/pypi/+simple/%s" % name)
@@ -85,7 +120,7 @@ def test_simple_project(pypistage, testapp):
 
 @pytest.mark.parametrize("outside_url", ['', 'http://localhost/devpi'])
 def test_simple_project_outside_url_subpath(mapp, outside_url, pypistage, testapp):
-    api = mapp.create_and_use()
+    api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
     mapp.upload_file_pypi(
         "qpwoei-1.0.tar.gz", b'123', "qpwoei", "1.0", indexname=api.stagename)
     pypistage.mock_simple("qpwoei", text='<a href="/qpwoei-1.0.zip"/>')
@@ -109,17 +144,20 @@ def test_simple_project_outside_url_subpath(mapp, outside_url, pypistage, testap
     [
         'pip/1.4.1',
         'setuptools/6.1',
+        'Python-urllib/3.5 setuptools/6.1',
+        'setuptools/6.1 Python-urllib/3.5',
+        'pex/1.0.1',
         'pip/6.0.dev1 {"cpu":"x86_64","distro":{"name":"OS X","version":"10.9.5"},"implementation":{"name":"CPython","version":"2.7.8"},"installer":{"name":"pip","version":"6.0.dev1"},"python":"2.7.8","system":{"name":"Darwin","release":"13.4.0"}}'],
-    ids=['pip', 'setuptools', 'pip6'])
+    ids=['pip', 'setuptools', 'urllib-setuptools', 'setuptools-urllib', 'pex', 'pip6'])
 def test_project_redirect(pypistage, testapp, user_agent):
     name = "qpwoei"
     headers = {'User-Agent': str(user_agent), "Accept": str("text/html")}
 
-    r = testapp.get("/root/pypi/%s" % name, headers=headers)
+    r = testapp.get("/root/pypi/%s" % name, headers=headers, follow=False)
     assert r.status_code == 302
     assert r.headers["location"].endswith("/root/pypi/+simple/%s" % name)
     # trailing slash will redirect to non trailing slash first
-    r = testapp.get("/root/pypi/%s/" % name, headers=headers)
+    r = testapp.get("/root/pypi/%s/" % name, headers=headers, follow=False)
     assert r.status_code == 302
     assert r.headers["location"].endswith("/root/pypi/+simple/%s" % name)
 
@@ -131,8 +169,8 @@ def test_simple_project_unicode_rejected(pypistage, testapp, dummyrequest):
     dummyrequest.log = pypistage.xom.log
     dummyrequest.context = RootFactory(dummyrequest)
     view = PyPIView(dummyrequest)
-    name = py.builtin._totext(b"qpw\xc3\xb6", "utf-8")
-    dummyrequest.matchdict.update(user="x", index="y", name=name)
+    project = py.builtin._totext(b"qpw\xc3\xb6", "utf-8")
+    dummyrequest.matchdict.update(user="x", index="y", project=project)
     with pytest.raises(HTTPClientError):
         view.simple_list_project()
 
@@ -150,7 +188,9 @@ def test_simple_project_pypi_egg(pypistage, testapp):
     r = testapp.get("/root/pypi")
     assert r.status_code == 200
 
+@pytest.mark.nomockprojectsremote
 def test_simple_list(pypistage, testapp):
+    pypistage.mock_simple_projects(["hello1", "hello2"])
     pypistage.mock_simple("hello1", "<html/>")
     pypistage.mock_simple("hello2", "<html/>")
     r = testapp.get("/root/pypi/+simple/hello1", expect_errors=False)
@@ -159,17 +199,41 @@ def test_simple_list(pypistage, testapp):
     assert int(r2.headers["X-DEVPI-SERIAL"]) == serial + 1
 
     r = testapp.get("/root/pypi/+simple/hello3")
-    assert r.status_code == 200
-    assert "no such project" in r.text
+    assert r.status_code == 404
     # easy_install fails if the result isn't html
     assert "html" in r.headers['content-type']
+
+    # get the full projects page and see what is in
     r = testapp.get("/root/pypi/+simple/")
     assert r.status_code == 200
-    links = BeautifulSoup(r.text).findAll("a")
+    links = BeautifulSoup(r.text, "html.parser").findAll("a")
     assert len(links) == 2
     hrefs = [a.get("href") for a in links]
     assert hrefs == ["hello1", "hello2"]
 
+def test_correct_resolution_order(pypistage, mapp, testapp):
+    pypistage.mock_simple("hello", pkgver="hello-1.0.tar.gz")
+    index1 = mapp.create_and_use()
+    index2 = mapp.create_and_use(indexconfig=dict(bases=[index1.stagename]))
+    index3 = mapp.create_and_use(
+        indexconfig=dict(bases=[index2.stagename, 'root/pypi']))
+    mapp.use(index1.stagename)
+    mapp.login(index1.user, index1.password)
+    mapp.upload_file_pypi("hello-1.0.tar.gz", b'123', "hello", "1.0",
+                          indexname=index1.stagename)
+    # the package should be found in our internal index, not on root/pypi
+    r = testapp.get("/%s/+simple/hello" % index3.stagename)
+    assert index1.stagename in r.text
+
+def test_simple_page_pypi_serial(pypistage, testapp):
+    pypistage.mock_simple("hello1", text="qwe", pypiserial=None)
+    r = testapp.get("/root/pypi/+simple/hello1", expect_errors=False)
+    assert "X-PYPI-LAST-SERIAL" not in r.headers
+    pypistage.mock_simple("hello2", pkgver="hello2-1.0.zip")
+    r = testapp.get("/root/pypi/+simple/hello2", expect_errors=False)
+    assert r.headers.get("X-PYPI-LAST-SERIAL") == '10000'
+    assert '/hello2-1.0.zip">hello2-1.0.zip</a>' in r.unicode_body
+
 def test_simple_refresh(mapp, model, pypistage, testapp):
     pypistage.mock_simple("hello", "<html/>")
     r = testapp.xget(200, "/root/pypi/+simple/hello")
@@ -177,15 +241,14 @@ def test_simple_refresh(mapp, model, pypistage, testapp):
     assert input.attrs['name'] == 'refresh'
     assert input.attrs['value'] == 'Refresh'
     with model.keyfs.transaction(write=False):
-        info = pypistage._load_project_cache("hello")
+        info = pypistage.key_projsimplelinks("hello").get()
     assert info != {}
-    assert info['projectname'] == 'hello'
     r = testapp.post("/root/pypi/+simple/hello/refresh")
     assert r.status_code == 302
     assert r.location.endswith("/root/pypi/+simple/hello")
     with model.keyfs.transaction(write=False):
-        info = pypistage._load_project_cache("hello")
-    assert info["entrylist"] == []
+        info = pypistage.key_projsimplelinks("hello").get()
+    assert info["links"] == []
 
 def test_inheritance_versiondata(mapp, model):
     api1 = mapp.create_and_use()
@@ -196,35 +259,33 @@ def test_inheritance_versiondata(mapp, model):
     assert len(r["result"]) == 1
 
 
-@pytest.mark.parametrize("projectname", ["pkg", "pkg_some"])
+@pytest.mark.parametrize("project", ["pkg", "pkg-some"])
 @pytest.mark.parametrize("stagename", [None, "root/pypi"])
-def test_simple_refresh_inherited(mapp, model, pypistage, testapp, projectname,
+def test_simple_refresh_inherited(mapp, model, pypistage, testapp, project,
                                   stagename):
-    pypistage.mock_simple(projectname, '<a href="/%s-1.0.zip" />' % projectname,
+    pypistage.mock_simple(project, '<a href="/%s-1.0.zip" />' % project,
                           serial=100)
     if stagename is None:
-        api = mapp.create_and_use()
+        api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
     else:
         api = mapp.use(stagename)
     stagename = api.stagename
 
-    r = testapp.xget(200, "/%s/+simple/%s" % (stagename, projectname))
+    r = testapp.xget(200, "/%s/+simple/%s" % (stagename, project))
     input, = r.html.select('form input')
     assert input.attrs['name'] == 'refresh'
     #assert input.attrs['value'] == 'Refresh PyPI links'
     with model.keyfs.transaction(write=False):
-        info = pypistage._load_project_cache(projectname)
+        info = pypistage.key_projsimplelinks(project).get()
     assert info != {}
-    assert info['projectname'] == projectname
-    pypistage.mock_simple(projectname, '<a href="/%s-2.0.zip" />' % projectname,
+    pypistage.mock_simple(project, '<a href="/%s-2.0.zip" />' % project,
                           serial=200)
-    r = testapp.post("/%s/+simple/%s/refresh" % (stagename, projectname))
+    r = testapp.post("/%s/+simple/%s/refresh" % (stagename, project))
     assert r.status_code == 302
-    assert r.location.endswith("/%s/+simple/%s" % (stagename, projectname))
+    assert r.location.endswith("/%s/+simple/%s" % (stagename, project))
     with model.keyfs.transaction(write=False):
-        info = pypistage._load_project_cache(projectname)
-    assert info["projectname"] == projectname
-    elist = info["entrylist"]
+        info = pypistage.key_projsimplelinks(project).get()
+    elist = info["links"]
     assert len(elist) == 1
     assert elist[0][0].endswith("-2.0.zip")
 
@@ -235,6 +296,19 @@ def test_simple_refresh_inherited_not_whitelisted(mapp, testapp):
     r = testapp.xget(200, "/%s/+simple/pkg" % api.stagename)
     assert len(r.html.select('form')) == 0
 
+
+def test_simple_blocked_warning(mapp, pypistage, testapp):
+    pypistage.mock_simple('pkg', '<a href="/pkg-1.0.zip" />', serial=100)
+    api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
+    mapp.set_versiondata(dict(name="pkg", version="1.0"), set_whitelist=False)
+    r = testapp.xget(200, "/%s/+simple/pkg" % api.stagename)
+    (paragraph,) = r.html.select('p')
+    assert paragraph.text == "INFO: Because this project isn't in the mirror_whitelist, no releases from root/pypi are included."
+    mapp.set_versiondata(dict(name="pkg", version="1.1"), set_whitelist=True)
+    r = testapp.xget(200, "/%s/+simple/pkg" % api.stagename)
+    assert r.html.select('p') == []
+
+
 def test_indexroot(testapp, model):
     with model.keyfs.transaction(write=True):
         user = model.create_user("user", "123")
@@ -252,7 +326,7 @@ def test_indexroot_root_pypi(testapp, xom):
     '/root/pypi/{name}/2.6',
     '/root/pypi/+simple/{name}',
 ])
-@pytest.mark.parametrize("code", [-1, 404, 500, 501, 502, 503])
+@pytest.mark.parametrize("code", [-1, 500, 501, 502, 503])
 def test_upstream_not_reachable(reqmock, pypistage, testapp, code, url):
     name = "whatever{code}".format(code=code+100)
     pypistage.mock_simple(name, '', status_code=code)
@@ -260,9 +334,38 @@ def test_upstream_not_reachable(reqmock, pypistage, testapp, code, url):
     assert r.status_code == 502
 
 
+def test_upstream_not_reachable_but_cache_still_returned(pypistage, mapp, testapp, monkeypatch):
+    index_name = 'user/name'
+    name = 'pkg1'
+    version = '1.0'
+    mapp.create_and_use(index_name, indexconfig=dict(bases=["root/pypi"]))
+    mapp.upload_file_pypi(
+        '{name}-{version}.tgz'.format(name=name, version=version),
+        b'123',
+        name=name,
+        version=version,
+        indexname=index_name,
+        register=True)
+    # first we check what happens if the cache is empty
+    pypistage.mock_simple(name, '', status_code=502)
+    r = testapp.get('/{index_name}/{name}'.format(index_name=index_name, name=name), accept="application/json")
+    assert r.status_code == 200
+    assert set(r.json['result']) == set(['1.0'])
+    # then we simulate that the mirror is available to fill the cache
+    pypistage.mock_simple(name, '<a href="/%s-1.1.zip" />' % name)
+    r = testapp.get('/{index_name}/{name}'.format(index_name=index_name, name=name), accept="application/json")
+    assert r.status_code == 200
+    assert set(r.json['result']) == set(['1.0', '1.1'])
+    # and check once more with the filled cache
+    pypistage.mock_simple(name, '', status_code=502)
+    r = testapp.get('/{index_name}/{name}'.format(index_name=index_name, name=name), accept="application/json")
+    assert r.status_code == 200
+    assert set(r.json['result']) == set(['1.0', '1.1'])
+
+
 def test_pkgserv(httpget, pypistage, testapp):
     pypistage.mock_simple("package", '<a href="/package-1.0.zip" />')
-    httpget.setextfile("/package-1.0.zip", b"123")
+    pypistage.mock_extfile("/package-1.0.zip", b"123")
     r = testapp.get("/root/pypi/+simple/package")
     assert r.status_code == 200
     href = getfirstlink(r.text).get("href")
@@ -277,7 +380,7 @@ def test_pkgserv_remote_failure(httpget, pypistage, testapp):
     assert r.status_code == 200
     href = getfirstlink(r.text).get("href")
     url = URL(r.request.url).joinpath(href).url
-    httpget.setextfile("/package-1.0.zip", b"123", status_code=500)
+    pypistage.mock_extfile("/package-1.0.zip", b"123", status_code=500)
     r = testapp.get(url)
     assert r.status_code == 502
 
@@ -305,6 +408,189 @@ class TestStatus:
         assert data["replication-errors"] == {}
 
 
+class TestStatusInfoPlugin:
+    @pytest.fixture
+    def plugin(self):
+        from devpi_server.views import devpiweb_get_status_info
+        return devpiweb_get_status_info
+
+    def _xomrequest(self, xom):
+        from pyramid.request import Request
+        request = Request.blank("/blankpath")
+        request.registry = dict(
+            xom=xom,
+            devpi_version_info=[])
+        return request
+
+    @pytest.mark.with_notifier
+    def test_no_issue(self, plugin, xom):
+        request = self._xomrequest(xom)
+        serial = xom.keyfs.get_current_serial()
+        xom.keyfs.notifier.wait_event_serial(serial)
+        # if devpi-web is installed make sure we look again
+        # at the serial in case a hook created a new serial
+        serial = xom.keyfs.get_current_serial()
+        xom.keyfs.notifier.wait_event_serial(serial)
+        result = plugin(request)
+        assert not xom.is_replica()
+        assert result == []
+
+    @pytest.mark.xfail(reason="sometimes fail due to race condition in db table creation")
+    @pytest.mark.with_replica_thread
+    @pytest.mark.with_notifier
+    def test_no_issue_replica(self, plugin, xom):
+        request = self._xomrequest(xom)
+        serial = xom.keyfs.get_current_serial()
+        xom.keyfs.notifier.wait_event_serial(serial)
+        result = plugin(request)
+        assert hasattr(xom, 'replica_thread')
+        assert xom.is_replica()
+        assert result == []
+
+    def test_events_lagging(self, plugin, xom, monkeypatch):
+        # write transaction so event processing can lag behind
+        with xom.keyfs.transaction(write=True):
+            xom.model.create_user("hello", "pass")
+
+        import time
+        now = time.time()
+        request = self._xomrequest(xom)
+        # nothing if events never processed directly after startup
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 30)
+        result = plugin(request)
+        assert result == []
+        # fatal after 5 minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
+        result = plugin(request)
+        assert result == [dict(
+            status='fatal',
+            msg="The event processing doesn't seem to start")]
+        # fake first event processed
+        xom.keyfs.notifier.write_event_serial(0)
+        xom.keyfs.notifier.event_serial_in_sync_at = now
+        # no report in the first minute
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 30)
+        result = plugin(request)
+        assert result == []
+        # warning after 5 minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
+        result = plugin(request)
+        assert result == [dict(
+            status='warn',
+            msg='No changes processed by plugins for more than 5 minutes')]
+        # fatal after 30 minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 1810)
+        result = plugin(request)
+        assert result == [dict(
+            status='fatal',
+            msg='No changes processed by plugins for more than 30 minutes')]
+        # warning about sync after one hour
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 3610)
+        result = plugin(request)
+        assert result == [
+            dict(
+                status='warn',
+                msg="The event processing hasn't been in sync for more than 1 hour"),
+            dict(
+                status='fatal',
+                msg='No changes processed by plugins for more than 30 minutes')]
+        # fatal sync state after 6 hours
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 21610)
+        result = plugin(request)
+        assert result == [
+            dict(
+                status='fatal',
+                msg="The event processing hasn't been in sync for more than 6 hours"),
+            dict(
+                status='fatal',
+                msg='No changes processed by plugins for more than 30 minutes')]
+
+    def test_replica_lagging(self, plugin, makexom, monkeypatch):
+        from devpi_server.replica import ReplicaThread
+        import time
+        now = time.time()
+        xom = makexom(["--master=http://localhost"])
+        request = self._xomrequest(xom)
+        xom.replica_thread = ReplicaThread(xom)
+        assert xom.is_replica()
+        # fake first serial processed
+        xom.replica_thread.update_master_serial(0)
+        xom.replica_thread.replica_in_sync_at = now
+        # no report in the first minute
+        result = plugin(request)
+        assert result == []
+        # warning after one minute
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
+        result = plugin(request)
+        assert result == [dict(
+            status='warn',
+            msg='Replica is behind master for more than 1 minute')]
+        # fatal after five minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
+        result = plugin(request)
+        assert result == [dict(
+            status='fatal',
+            msg='Replica is behind master for more than 5 minutes')]
+
+    def test_initial_master_connection(self, plugin, makexom, monkeypatch):
+        from devpi_server.replica import ReplicaThread
+        import time
+        now = time.time()
+        xom = makexom(["--master=http://localhost"])
+        request = self._xomrequest(xom)
+        xom.replica_thread = ReplicaThread(xom)
+        assert xom.is_replica()
+        assert xom.replica_thread.started_at is None
+        # fake replica start
+        xom.replica_thread.started_at = now
+        # no report in the first minute
+        result = plugin(request)
+        assert result == []
+        # warning after one minute
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
+        result = plugin(request)
+        assert result == [dict(
+            status='warn',
+            msg='No contact to master for more than 1 minute')]
+        # fatal after five minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
+        result = plugin(request)
+        assert result == [dict(
+            status='fatal',
+            msg='No contact to master for more than 5 minutes')]
+
+    @pytest.mark.xfail(reason="sometimes fail due to race condition in db table creation")
+    @pytest.mark.with_replica_thread
+    @pytest.mark.with_notifier
+    def test_no_master_update(self, plugin, xom, monkeypatch):
+        import time
+        now = time.time()
+        request = self._xomrequest(xom)
+        serial = xom.keyfs.get_current_serial()
+        xom.keyfs.notifier.wait_event_serial(serial)
+        serial = xom.keyfs.get_current_serial()
+        xom.keyfs.notifier.wait_event_serial(serial)
+        assert hasattr(xom, 'replica_thread')
+        assert xom.is_replica()
+        # fake last update
+        xom.replica_thread.update_from_master_at = now
+        # no report in the first minute
+        result = plugin(request)
+        assert result == []
+        # warning after one minute
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 70)
+        result = plugin(request)
+        assert result == [dict(
+            status='warn',
+            msg='No update from master for more than 1 minute')]
+        # fatal after five minutes
+        monkeypatch.setattr(devpi_server.views, "time", lambda: now + 310)
+        result = plugin(request)
+        assert result == [dict(
+            status='fatal',
+            msg='No update from master for more than 5 minutes')]
+
+
 def test_apiconfig_with_outside_url(testapp):
     testapp.xom.config.args.outside_url = u = "http://outside.com"
     r = testapp.get_json("/root/pypi/+api")
@@ -347,7 +633,8 @@ class TestSubmitValidation:
             def __init__(self, stagename="user/dev"):
                 self.stagename = stagename
                 self.username = stagename.split("/")[0]
-                self.api = mapp.create_and_use(stagename)
+                self.api = mapp.create_and_use(
+                    stagename, indexconfig=dict(bases=["root/pypi"]))
 
             def metadata(self, metadata, code):
                 return testapp.post(self.api.pypisubmit, metadata, code=code)
@@ -411,8 +698,14 @@ class TestSubmitValidation:
         r = submit.file("pkg5-2.6.qwe", b"123", {"name": "Pkg5"}, code=400)
         assert "not a valid" in r.status
         r = submit.file("pkg5-2.7.tgz", b"123", {"name": "pkg5"}, code=200)
-        paths = mapp.get_release_paths("Pkg5")
-        assert paths[0].endswith("pkg5-2.7.tgz")
+        mapp.get_release_paths("Pkg5")
+
+    def test_upload_file_version_not_in_filename(self, submit, mapp):
+        metadata = {"name": "Pkg5", "version": "1.0", ":action": "submit"}
+        submit.metadata(metadata, code=200)
+        r = submit.file("pkg5-0.0.0.tgz", b"123", {"name": "Pkg5", "version": "1.0"},
+                        code=400)
+        assert "does not contain version" in r.status
 
     def test_upload_use_registered_name_issue84(self, submit, mapp):
         metadata = {"name": "pkg_hello", "version":"1.0", ":action": "submit"}
@@ -438,12 +731,15 @@ class TestSubmitValidation:
         r = testapp.delete(submit.api.index + "/pkg-hello")
         assert r.status_code == 200
 
-    def test_upload_and_simple_index(self, submit, testapp):
+    def test_upload_and_simple_index_with_redirect(self, submit, testapp):
         metadata = {"name": "Pkg5", "version": "2.6", ":action": "submit"}
         submit.metadata(metadata, code=200)
         submit.file("pkg5-2.6.tgz", b"123", {"name": "Pkg5"}, code=200)
-        r = testapp.get("/%s/+simple/pkg5" % submit.stagename)
+        r = testapp.get("/%s/+simple/Pkg5" % submit.stagename, follow=False)
         assert r.status_code == 302
+        assert r.location.endswith("pkg5")
+        r = testapp.get(r.location)
+        assert r.status_code == 200
 
     def test_upload_and_delete_index(self, submit, testapp, mapp):
         metadata = {"name": "Pkg5", "version": "2.6", ":action": "submit"}
@@ -480,7 +776,7 @@ class TestSubmitValidation:
         mapp.delete_user(submit.username)
         # recreate user and index
         submit = submit.__class__(submit.stagename)
-        assert not mapp.get_release_paths("pkg5")
+        mapp.get_simple("pkg5", code=404)
 
     def test_upload_twice_to_nonvolatile(self, submit, testapp, mapp):
         mapp.modify_index(submit.stagename, indexconfig=dict(volatile=False))
@@ -506,7 +802,8 @@ class TestSubmitValidation:
         assert path1 == path2
         r = testapp.xget(200, path2)
         assert r.body == b'123'
-        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index)
+        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index,
+                         accept="application/json")
         links = r.json['result']['+links']
         assert len(links) == 2
         for link in links:
@@ -528,7 +825,8 @@ class TestSubmitValidation:
         path2, = mapp.get_release_paths("Pkg5")
         testapp.xget(410, path1)  # existed once but deleted during overwrite
         testapp.xget(200, path2)
-        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index)
+        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index,
+                         accept="application/json")
         links = r.json['result']['+links']
         assert len(links) == 2
         for link in links:
@@ -548,7 +846,8 @@ class TestSubmitValidation:
         submit.file("pkg5-2.6.tgz", b"123", {"name": "Pkg5"}, code=200)
         submit.file("pkg5-2.6.tgz", b"1234", {"name": "Pkg5"}, code=200)
         submit.file("pkg5-2.6.tgz", b"12345", {"name": "Pkg5"}, code=200)
-        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index)
+        r = testapp.xget(200, "%s/Pkg5/2.6" % mapp.api.index,
+                         accept="application/json")
         link, = r.json['result']['+links']
         log1, log2 = link['log']
         assert sorted(log1.keys()) == ['count', 'what', 'when', 'who']
@@ -565,7 +864,8 @@ class TestSubmitValidation:
         mapp.use(old_stage)
         req = dict(name="Pkg5", version="2.6", targetindex=new_stage)
         r = testapp.push("/%s" % old_stage, json.dumps(req))
-        r = testapp.xget(200, "/%s/Pkg5/2.6" % new_stage)
+        r = testapp.xget(200, "/%s/Pkg5/2.6" % new_stage,
+                         accept="application/json")
         link, = r.json['result']['+links']
         # the overwrite info should be gone
         log1, log2 = link['log']
@@ -625,8 +925,8 @@ class TestSubmitValidation:
         metadata = {"name": "Pkg1", "version": "1.0", ":action": "submit",
                     "description": "hello world"}
         submit.metadata(metadata, code=200)
-        location = mapp.getjson("/%s/pkg1" % submit.stagename, code=302)
-        assert location.endswith("/Pkg1")
+        mapp.getjson("/%s/pkg1" % submit.stagename, code=200)
+        #assert location.endswith("/Pkg1")
 
 
 def test_submit_authorization(mapp, testapp):
@@ -687,6 +987,36 @@ def test_push_from_base_error(mapp, testapp, monkeypatch, pypistage):
     assert r.status_code == 400
     assert "no files for" in r.json["message"]
 
+
+def test_push_from_pypi(httpget, mapp, pypistage, testapp):
+    pypistage.mock_simple("hello", text='<a href="hello-1.0.tar.gz"/>')
+    pypistage.mock_extfile("/simple/hello/hello-1.0.tar.gz", b"123")
+    mapp.create_and_login_user("foo")
+    mapp.create_index("newindex1", indexconfig=dict(bases=["root/pypi"]))
+    mapp.use("root/pypi")
+    req = dict(name="hello", version="1.0", targetindex="foo/newindex1")
+    r = testapp.push("/root/pypi", json.dumps(req))
+    assert r.status_code == 200
+    assert r.json == {
+        'result': [
+            [200, 'register', 'hello', '1.0', '->', 'foo/newindex1'],
+            [200, 'store_releasefile',
+             'foo/newindex1/+f/a66/5a45920422f9d/hello-1.0.tar.gz']],
+        'type': 'actionlog'}
+
+
+def test_push_from_pypi_fail(httpget, mapp, pypistage, testapp):
+    pypistage.mock_simple("hello", text='<a href="hello-1.0.tar.gz"/>')
+    pypistage.mock_extfile("/simple/hello/hello-1.0.tar.gz", b"123", status_code=502)
+    mapp.create_and_login_user("foo")
+    mapp.create_index("newindex1", indexconfig=dict(bases=["root/pypi"]))
+    mapp.use("root/pypi")
+    req = dict(name="hello", version="1.0", targetindex="foo/newindex1")
+    r = testapp.push("/root/pypi", json.dumps(req))
+    assert r.status_code == 502
+    assert r.json["message"] == "error 502 getting https://pypi.python.org/simple/hello/hello-1.0.tar.gz"
+
+
 def test_upload_docs_without_registration(mapp, testapp, monkeypatch):
     mapp.create_and_use()
     mapp.upload_file_pypi("pkg1-2.6.tgz", b"123", "pkg1", "2.6")
@@ -945,16 +1275,17 @@ class TestPluginPermissions:
     def test_plugin_user_create_index(self, mapp):
         mapp.login("pluginuser")
         assert "pluginuser" not in mapp.getuserlist()
-        # user must already exist to succeed, since it doesn't, it should fail
-        mapp.create_index("pluginuser/dev", code=404)
+        mapp.create_index("pluginuser/dev")
+        assert "pluginuser" in mapp.getuserlist()
+        assert sorted(mapp.getindexlist("pluginuser")) == ['pluginuser/dev']
 
 
 def test_upload_trigger(mapp):
     class Plugin:
         def devpiserver_on_upload_sync(self, log, application_url,
-                                       stage, projectname, version):
+                                       stage, project, version):
             self.results.append(
-                (application_url, stage.name, projectname, version))
+                (application_url, stage.name, project, version))
     plugin = Plugin()
     plugin.results = []
     mapp.xom.config.pluginmanager.register(plugin)
@@ -1011,12 +1342,12 @@ def test_delete_version_fails_on_non_volatile(mapp):
     mapp.delete_project("pkg1/2.6", code=403)
 
 
-def test_upload_pypi_fails(mapp):
+def test_upload_to_mirror_fails(mapp):
     mapp.upload_file_pypi(
             "pkg1-2.6.tgz", b"123", "pkg1", "2.6", code=404,
             indexname="root/pypi")
 
-def test_delete_pypi_fails(mapp):
+def test_delete_from_mirror_fails(mapp):
     mapp.login_root()
     mapp.use("root/pypi")
     mapp.delete_project("pytest/2.3.5", code=405)
@@ -1029,6 +1360,27 @@ def test_delete_volatile_fails(mapp):
     mapp.upload_file_pypi("pkg5-2.6.tgz", b"123", "pkg5", "2.6")
     mapp.delete_project("pkg5", code=403)
 
+
+@pytest.mark.parametrize("volatile", [True, False])
+@pytest.mark.parametrize("restrict_modify", [None, "root"])
+def test_delete_with_acl_upload(mapp, restrict_modify, volatile, xom):
+    xom.config.args.restrict_modify = restrict_modify
+    mapp.login_root()
+    mapp.create_user("user1", "1")
+    mapp.create_index("user1/dev", indexconfig=dict(
+        acl_upload=["user2"],
+        volatile=volatile))
+    mapp.create_and_login_user("user2")
+    mapp.use("user1/dev")
+    mapp.upload_file_pypi(
+        "pkg5-2.6.tgz", b"123", "pkg5", "2.6", set_whitelist=False)
+    mapp.upload_file_pypi(
+        "pkg5-2.7.tgz", b"123", "pkg5", "2.7", set_whitelist=False)
+    result_code = 200 if volatile else 403
+    mapp.delete_project('pkg5/2.6', code=result_code)
+    mapp.delete_project('pkg5', code=result_code)
+
+
 @proj
 def test_upload_docs_no_version(mapp, testapp, proj):
     api = mapp.create_and_use()
@@ -1037,7 +1389,7 @@ def test_upload_docs_no_version(mapp, testapp, proj):
     mapp.upload_doc("pkg1.zip", content, "Pkg1", "")
     vv = get_view_version_links(testapp, api.index, "Pkg1", "1.0", proj=proj)
     link = vv.get_link("doczip")
-    assert link.href.endswith("/Pkg1-1.0.doc.zip")
+    assert link.href.endswith("/pkg1-1.0.doc.zip")
     r = testapp.get(link.href)
     archive = Archive(py.io.BytesIO(r.body))
     assert 'index.html' in archive.namelist()
@@ -1085,26 +1437,6 @@ def test_wrong_login_format(testapp, mapp):
     assert r.status_code == 400
 
 
-
-@pytest.mark.parametrize(["input", "expected"], [
-    ({},
-      dict(type="stage", volatile=True, bases=["root/pypi"])),
-    ({"volatile": "False"},
-      dict(type="stage", volatile=False, bases=["root/pypi"])),
-    ({"volatile": "False", "bases": "root/pypi"},
-      dict(type="stage", volatile=False, bases=["root/pypi"])),
-    ({"volatile": "False", "bases": ["root/pypi"]},
-      dict(type="stage", volatile=False, bases=["root/pypi"])),
-    ({"volatile": "False", "bases": ["root/pypi"], "acl_upload": ["hello"]},
-      dict(type="stage", volatile=False, bases=["root/pypi"],
-           acl_upload=["hello"])),
-])
-def test_kvdict(xom, input, expected):
-    from devpi_server.views import getkvdict_index
-    result = getkvdict_index(xom.config.hook, input)
-    assert result == expected
-
-
 @pytest.mark.parametrize("headers, environ, outsideurl, expected", [
     (
         {"X-outside-url": "http://outside.com"}, {},
@@ -1145,6 +1477,38 @@ def test_outside_url_middleware(headers, environ, outsideurl, expected, testapp)
     assert r.json['result']['login'] == "%s/+login" % expected
 
 
+@pytest.mark.parametrize("stagename", [None, "root/pypi"])
+class TestOfflineMode:
+    @pytest.fixture
+    def xom(self, makexom):
+        return makexom(["--offline-mode"])
+
+    def _prepare(self, mapp, pypistage, stagename):
+        pypistage.mock_simple("package", '<a href="/package-1.0.zip" />', serial=100)
+        if stagename is None:
+            api = mapp.create_and_use(indexconfig=dict(bases=["root/pypi"]))
+        else:
+            api = mapp.use(stagename)
+        return api.stagename
+
+    def test_file_not_available(self, mapp, model, testapp, pypistage, stagename):
+        stagename = self._prepare(mapp, pypistage, stagename)
+        testapp.xget(200, "/%s/+simple/package" % stagename)
+        with model.keyfs.transaction(write=False):
+            is_fresh, links, serial = pypistage._load_cache_links("package")
+
+        assert len(links) == 0
+
+    def test_file_available(self, mapp, model, testapp, pypistage, monkeypatch, stagename):
+        stagename = self._prepare(mapp, pypistage, stagename)
+        monkeypatch.setattr(devpi_server.filestore.FileEntry, "file_exists", lambda a: True)
+        testapp.xget(200, "/%s/+simple/package" % stagename)
+        with model.keyfs.transaction(write=False):
+            is_fresh, links, serial = pypistage._load_cache_links("package")
+
+        assert links[0][0] == "package-1.0.zip"
+
+
 class Test_getjson:
     @pytest.fixture
     def abort_calls(self, monkeypatch):
